Package: GeoTox
Title: Spatiotemporal Mixture Risk Assessment
Version: 1.0.0
Authors@R: c(
    person("Skylar", "Marvel", , "skylar.marvel@nih.gov", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-2971-9743")),
    person("David", "Reif", , "david.reif@nih.gov", role = "aut",
           comment = c(ORCID = "0000-0001-7815-6767")),
    person("Kyle", "Messier", , "kyle.messier@nih.gov", role = "aut",
           comment = c(ORCID = "0000-0001-9508-9623")),
    person("Spatiotemporal Exposures and Toxicology Group", role = "cph")
  )
Description: Connecting spatiotemporal exposure to individual and
    population-level risk via source-to-outcome continuum modeling. The
    package, methods, and case-studies are described in Messier, Reif, and
    Marvel (2025) <doi:10.1186/s40246-024-00711-8> and Eccles et al.
    (2023) <doi:10.1016/j.scitotenv.2022.158905>.
License: MIT + file LICENSE
URL: https://niehs.github.io/GeoTox/, https://github.com/NIEHS/GeoTox
BugReports: https://github.com/NIEHS/GeoTox/issues
Depends: R (>= 4.4.0)
Imports: DBI, dbplyr, dplyr, duckdb, duckplyr, purrr, rlang, tibble,
        tidyr, tidyselect, truncnorm
Suggests: blob, ggplot2, ggridges, httk, httr2, knitr, quarto, readr,
        readxl, rmarkdown, sf, stringr, testthat (>= 3.0.0), withr, zip
VignetteBuilder: knitr, quarto
Config/Needs/website: rmarkdown, quarto
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
NeedsCompilation: no
Packaged: 2026-05-19 17:30:01 UTC; marvelsw
Author: Skylar Marvel [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2971-9743>),
  David Reif [aut] (ORCID: <https://orcid.org/0000-0001-7815-6767>),
  Kyle Messier [aut] (ORCID: <https://orcid.org/0000-0001-9508-9623>),
  Spatiotemporal Exposures and Toxicology Group [cph]
Maintainer: Skylar Marvel <skylar.marvel@nih.gov>
Repository: CRAN
Date/Publication: 2026-05-19 21:20:18 UTC
Built: R 4.6.0; ; 2026-05-19 22:23:15 UTC; unix

Package: GeoTox
Title: Spatiotemporal Mixture Risk Assessment
Version: 1.0.1
Authors@R: c(
    person("Skylar", "Marvel", , "skylar.marvel@nih.gov", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-2971-9743")),
    person("David", "Reif", , "david.reif@nih.gov", role = "aut",
           comment = c(ORCID = "0000-0001-7815-6767")),
    person("Kyle", "Messier", , "kyle.messier@nih.gov", role = "aut",
           comment = c(ORCID = "0000-0001-9508-9623")),
    person("Spatiotemporal Exposures and Toxicology Group", role = "cph")
  )
Description: Connecting spatiotemporal exposure to individual and
    population-level risk via source-to-outcome continuum modeling. The
    package, methods, and case-studies are described in Messier, Reif, and
    Marvel (2025) <doi:10.1186/s40246-024-00711-8> and Eccles et al.
    (2023) <doi:10.1016/j.scitotenv.2022.158905>.
License: MIT + file LICENSE
URL: https://niehs.github.io/GeoTox/, https://github.com/NIEHS/GeoTox
BugReports: https://github.com/NIEHS/GeoTox/issues
Depends: R (>= 4.4.0)
Imports: DBI, dbplyr, dplyr, duckdb, duckplyr, purrr, rlang, tibble,
        tidyr, tidyselect, truncnorm
Suggests: blob, ggplot2, ggridges, httk, httr2, knitr, quarto, readr,
        readxl, rmarkdown, sf, stringr, testthat (>= 3.0.0), withr, zip
VignetteBuilder: knitr, quarto
Config/Needs/website: rmarkdown, quarto
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
NeedsCompilation: no
Packaged: 2026-07-24 20:08:16 UTC; marvelsw
Author: Skylar Marvel [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2971-9743>),
  David Reif [aut] (ORCID: <https://orcid.org/0000-0001-7815-6767>),
  Kyle Messier [aut] (ORCID: <https://orcid.org/0000-0001-9508-9623>),
  Spatiotemporal Exposures and Toxicology Group [cph]
Maintainer: Skylar Marvel <skylar.marvel@nih.gov>
Repository: CRAN
Date/Publication: 2026-07-24 20:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,GeoTox)
export(GeoTox)
export(add_age)
export(add_exposure)
export(add_exposure_rate_params)
export(add_hill_params)
export(add_obesity)
export(calc_internal_dose)
export(calc_invitro_concentration)
export(calc_response)
export(calc_risk)
export(calc_sensitivity)
export(fit_hill)
export(get_assay_table)
export(get_boundary)
export(get_con)
export(get_concentration_mean)
export(get_risk_quantiles)
export(get_risk_sensitivity)
export(get_risk_values)
export(sample_simulated_css)
export(sensitivity_analysis)
export(set_boundary)
export(set_fixed_css)
export(set_sample)
export(set_simulated_css)
export(simulate_age)
export(simulate_exposure)
export(simulate_exposure_rate)
export(simulate_obesity)
export(simulate_population)
importFrom(rlang,.data)
importFrom(rlang,.env)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# GeoTox <a href="https://github.com/NIEHS/GeoTox"><img src="man/figures/logo.svg" align="right" height="139" alt="GeoTox website" /></a>

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![R-CMD-check](https://github.com/NIEHS/GeoTox/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/NIEHS/GeoTox/actions/workflows/R-CMD-check.yaml)
[![pkgdown](https://github.com/NIEHS/GeoTox/actions/workflows/pkgdown.yaml/badge.svg)](https://github.com/NIEHS/GeoTox/actions/workflows/pkgdown.yaml)
[![cov](https://NIEHS.github.io/GeoTox/badges/coverage.svg)](https://github.com/NIEHS/GeoTox/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/GeoTox)](https://CRAN.R-project.org/package=GeoTox)

<!-- badges: end -->

`GeoTox` is an open-source R software package for characterizing the
risk of perturbing molecular targets involved in adverse human health
outcomes based on exposure to spatially-referenced stressor mixtures via
the GeoTox framework - otherwise known as source-to-outcome-continuum
modeling. The package, methods, and case-studies are described in
[Messier, Reif, and Marvel, 2025, Human
Genomics](https://doi.org/10.1186/s40246-024-00711-8).

The GeoTox framework was first described in [Eccles et al. A geospatial
modeling approach to quantifying the risk of exposure to environmental
chemical mixtures via a common molecular target. Sci Total Environ. 2023
Jan 10;855:158905.](https://pubmed.ncbi.nlm.nih.gov/36152849/)

## Installation

The current stable version of the package can be installed from CRAN.

``` r
install.packages("GeoTox")
```

The previous implementation of the package, which uses in-memory storage
instead of DuckDB, can be installed by specifying the version.

``` r
#install.packages("pak")
pak::pak("GeoTox@0.3.0")
```

## Development Version

The development version can be installed from GitHub.

``` r
#install.packages("pak")
pak::pak("NIEHS/GeoTox@dev")
```

## Overview

Figure 1 below shows the steps in the source-to-outcome-continuum or
GeoTox modeling framework that is capable in the `GeoTox` package. The
`GeoTox` package uses R S3 object-oriented programming to create the
**GeoTox** object. The **GeoTox** object contains all of the information
required to run the analysis including tracking exposure, assays,
dose-response, and geospatial referencing. Each component of the object
and pipeline are stored in a relational DuckDB​​ database.

<p align="center">

<img src="man/figures/GeoTox.png" alt="GeoTox"> <br> <em>Figure 1:
Overview of the steps in the GeoTox framework</em>
</p>

## Getting Started

Please refer to the [package website](https://niehs.github.io/GeoTox/)
for a detailed description of how to use `GeoTox`. The development
version can be accessed [here](https://niehs.github.io/GeoTox/dev/).

## Citation

For citation information, please refer to our [CITATION
file](inst/CITATION).

## Issues and Contributions

To add or edit functionality, open a pull request into the main branch
with a detailed description of the proposed changes. Pull requests must
pass all status checks, and then will be approved or rejected by the
GeoTox maintainers.

Utilize [GitHub issues](https://github.com/NIEHS/GeoTox/issues) to
notify the authors of bugs, questions, or recommendations. Identify each
issue with the appropriate label to help ensure a timely response.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# GeoTox <a href="https://github.com/NIEHS/GeoTox"><img src="man/figures/logo.svg" align="right" height="139" alt="GeoTox website" /></a>

## Maintainer

Messier, Kyle (NIH/NIEHS) <kyle.messier@nih.gov>

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![R-CMD-check](https://github.com/NIEHS/GeoTox/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/NIEHS/GeoTox/actions/workflows/R-CMD-check.yaml)
[![pkgdown](https://github.com/NIEHS/GeoTox/actions/workflows/pkgdown.yaml/badge.svg)](https://github.com/NIEHS/GeoTox/actions/workflows/pkgdown.yaml)
[![cov](https://NIEHS.github.io/GeoTox/badges/coverage.svg)](https://github.com/NIEHS/GeoTox/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/GeoTox)](https://CRAN.R-project.org/package=GeoTox)

<!-- badges: end -->

`GeoTox` is an open-source R software package for characterizing the
risk of perturbing molecular targets involved in adverse human health
outcomes based on exposure to spatially-referenced stressor mixtures via
the GeoTox framework - otherwise known as source-to-outcome-continuum
modeling. The package, methods, and case-studies are described in
[Messier, Reif, and Marvel, 2025, Human
Genomics](https://doi.org/10.1186/s40246-024-00711-8).

The GeoTox framework was first described in [Eccles et al. A geospatial
modeling approach to quantifying the risk of exposure to environmental
chemical mixtures via a common molecular target. Sci Total Environ. 2023
Jan 10;855:158905.](https://pubmed.ncbi.nlm.nih.gov/36152849/)

## Installation

The current stable version of the package can be installed from CRAN.

``` r
install.packages("GeoTox")
```

The previous implementation of the package, which uses in-memory storage
instead of DuckDB, can be installed by specifying the version.

``` r
#install.packages("pak")
pak::pak("GeoTox@0.3.0")
```

## Development Version

The development version can be installed from GitHub.

``` r
#install.packages("pak")
pak::pak("NIEHS/GeoTox@dev")
```

## Overview

Figure 1 below shows the steps in the source-to-outcome-continuum or
GeoTox modeling framework that is capable in the `GeoTox` package. The
`GeoTox` package uses R S3 object-oriented programming to create the
**GeoTox** object. The **GeoTox** object contains all of the information
required to run the analysis including tracking exposure, assays,
dose-response, and geospatial referencing. Each component of the object
and pipeline are stored in a relational DuckDB​​ database.

<p align="center">

<img src="man/figures/GeoTox.png" alt="GeoTox"> <br> <em>Figure 1:
Overview of the steps in the GeoTox framework</em>
</p>

## Getting Started

Please refer to the [package website](https://niehs.github.io/GeoTox/)
for a detailed description of how to use `GeoTox`. The development
version can be accessed [here](https://niehs.github.io/GeoTox/dev/).

## Citation

For citation information, please refer to our [CITATION
file](inst/CITATION).

## Issues and Contributions

To add or edit functionality, open a pull request into the main branch
with a detailed description of the proposed changes. Pull requests must
pass all status checks, and then will be approved or rejected by the
GeoTox maintainers.

Utilize [GitHub issues](https://github.com/NIEHS/GeoTox/issues) to
notify the authors of bugs, questions, or recommendations. Identify each
issue with the appropriate label to help ensure a timely response.

## -----------------------------------------------------------------------------
#| label: setup
library(dplyr)
library(httk)
library(httr2)
library(purrr)
library(readr)
library(readxl)
library(sf)
library(stringr)
library(tibble)
library(tidyr)

geo_tox_data <- list()


## -----------------------------------------------------------------------------
#| label: exposure
filename <- "Region4b_2020ATS_Exposure_Concentrations.xlsx"
tmp <- tempfile(filename)
download.file(
  paste0(
    "https://gaftp.epa.gov/rtrmodeling_public/AirToxScreen/2020/",
    "Exposure%20Concentrations/",
    filename
  ),
  tmp
)
exposure <- read_xlsx(tmp)

# Normalization function
min_max_norm = function(x) {
  min_x <- min(x, na.rm = TRUE)
  max_x <- max(x, na.rm = TRUE)
  if (min_x == max_x) {
    rep(0, length(x))
  } else {
    (x - min_x) / (max_x - min_x)
  }
}

geo_tox_data$exposure <- exposure |>
  # North Carolina counties
  filter(State == "NC", !grepl("0$", FIPS)) |>
  # Aggregate chemicals by county
  summarize(
    across(
      -c(State:Population), # Note: use "-c(State:Tract)" for 2019 data
      list(mean = mean, sd = sd),
      .names = "{col}|||{fn}"
    ),
    .by = FIPS
  ) |>
  pivot_longer(-FIPS) |>
  separate_wider_delim(name, "|||", names = c("chemical", "stat")) |>
  pivot_wider(names_from = stat) |>
  # Normalize concentrations
  mutate(norm = min_max_norm(mean), .by = chemical)


## -----------------------------------------------------------------------------
#| label: comptox
# Copy/paste the chemical names into the batch search field
# https://comptox.epa.gov/dashboard/batch-search
cat(geo_tox_data$exposure |> distinct(chemical) |> pull(), sep = "\n")
# Export results with "CAS-RN" identifiers as a csv file, then process in R

# Remove rows without clear chemical matches, investigate manually if desired
exposure_casrn <- read_csv("CCD-Batch-Search.csv") |>
  filter(DTXSID != "N/A", !grepl("WARNING", FOUND_BY))

# Update exposure data with CompTox Dashboard data
geo_tox_data$exposure <- geo_tox_data$exposure |>
  inner_join(exposure_casrn, by = join_by(chemical == INPUT)) |>
  select(FIPS, casn = CASRN, chnm = PREFERRED_NAME, mean, sd, norm)


## -----------------------------------------------------------------------------
#| label: available-assays
# Get all supported assays
help_text <- request("https://ice.ntp.niehs.nih.gov/api/v1/search/help") |>
  req_perform() |>
  resp_body_string()
supported_assays <- str_split_i(help_text, "Supported assay\\(s\\):", 2) |>
  str_split_1("\",\"") |>
  str_replace_all(c("\"" = "")) |>
  str_trim()

# Search for assays available for given chemids
chemids <- unique(geo_tox_data$exposure$casn)
resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/search") |>
  req_body_json(list(chemids = chemids, assays = supported_assays)) |>
  req_perform()
result <- resp |> resp_body_json() |> pluck("endPoints")
fields <- c("assay", "casrn", "dtxsid", "substanceName", "endpoint", "value")
df <- map(fields, \(x) map_chr(result, x)) |>
  set_names(fields) |>
  bind_cols()
available_assays <- df |> distinct(assay) |> pull() |> sort()


## -----------------------------------------------------------------------------
#| label: cHTS-hits
get_cHTS_hits <- function(assays = NULL, chemids = NULL) {
  if (is.null(assays) & is.null(chemids)) {
    stop("Must provide at least one of 'assays' or 'chemids'")
  }

  # Format query parameters
  req_params <- list()

  if (!is.null(assays)) {
    if (!is.list(assays)) assays <- as.list(assays)
    req_params$assays <- assays
  }

  if (!is.null(chemids)) {
    if (!is.list(chemids)) chemids <- as.list(chemids)
    req_params$chemids <- chemids
  }

  # Query ICE API
  resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/search") |>
    req_body_json(req_params) |>
    req_perform()

  if (resp$status_code != 200) {
    stop("Failed to retrieve data from ICE API")
  }

  # Return active chemicals
  result <- resp |> resp_body_json() |> pluck("endPoints")

  fields <- c("assay", "casrn", "dtxsid", "substanceName", "endpoint", "value")

  map(fields, \(x) map_chr(result, x)) |>
    set_names(fields) |>
    bind_cols() |>
    filter(endpoint == "Call", value == "Active") |>
    select(-c(endpoint, value)) |>
    distinct()
}

assays <- c(
  "APR_HepG2_p53Act_1hr",
  "APR_HepG2_p53Act_24hr",
  "APR_HepG2_p53Act_72hr",
  "ATG_p53_CIS",
  "TOX21_DT40_LUC",
  "TOX21_DT40_100_LUC",
  "TOX21_DT40_657_LUC",
  "TOX21_ELG1_LUC_Agonist",
  "TOX21_H2AX_HTRF_CHO_Agonist_ratio",
  "TOX21_p53_BLA_p1_ratio",
  "TOX21_p53_BLA_p2_ratio",
  "TOX21_p53_BLA_p3_ratio",
  "TOX21_p53_BLA_p4_ratio",
  "TOX21_p53_BLA_p5_ratio"
)

chemids <- unique(geo_tox_data$exposure$casn)

cHTS_hits <- get_cHTS_hits(assays = assays, chemids = chemids)


## -----------------------------------------------------------------------------
#| label: dose-response
get_ICE_dose_resp <- function(assays = NULL, chemids = NULL) {
  if (is.null(assays) & is.null(chemids)) {
    stop("Must provide at least one of 'assays' or 'chemids'")
  }

  # Format query parameters
  req_params <- list()

  if (!is.null(assays)) {
    if (!is.list(assays)) assays <- as.list(assays)
    req_params$assays <- assays
  }

  if (!is.null(chemids)) {
    if (!is.list(chemids)) chemids <- as.list(chemids)
    req_params$chemids <- chemids
  }

  # Query ICE API
  resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/curves") |>
    req_body_json(req_params) |>
    req_perform()

  if (resp$status_code != 200) {
    stop("Failed to retrieve data from ICE API")
  }

  # Return dose-response data
  result <- resp |> resp_body_json() |> pluck("curves")

  map(result, function(x) {
    tibble(
      endp = x[["endpoint"]],
      casn = x[["casrn"]],
      chnm = x[["substance"]],
      call = x[["call"]],
      logc = map_dbl(x[["concentrationResponses"]], "concentration") |> log10(),
      resp = map_dbl(x[["concentrationResponses"]], "response")
    )
  }) |>
    bind_rows()
}

assays <- unique(cHTS_hits$assay)

chemids <- intersect(cHTS_hits$casrn, geo_tox_data$exposure$casn)

dose_response <- get_ICE_dose_resp(assays = assays, chemids = chemids)

# Only keep active calls for assay/chemical combinations
geo_tox_data$dose_response <- dose_response |>
  filter(call == "Active") |>
  select(-call)

# Update dose-response chemical names using CompTox Dashboard data
geo_tox_data$dose_response <- geo_tox_data$dose_response |>
  inner_join(exposure_casrn, by = join_by(casn == CASRN)) |>
  select(endp, casn, chnm = PREFERRED_NAME, logc, resp)


## -----------------------------------------------------------------------------
#| label: age
# Data for North Carolina
url <- paste0(
  "https://www2.census.gov/programs-surveys/popest/datasets/",
  "2020-2024/counties/asrh/cc-est2024-alldata-37.csv"
)
age <- read_csv(url)

geo_tox_data$age <- age |>
  # 7/1/2020 population estimate
  filter(YEAR == 2) |>
  # Create FIPS
  mutate(FIPS = str_c(STATE, COUNTY)) |>
  # Keep selected columns
  select(FIPS, AGEGRP, TOT_POP)


## -----------------------------------------------------------------------------
#| label: obesity-status
places <- read_csv(
  "PLACES__County_Data_(GIS_Friendly_Format),_2020_release.csv"
)

# Convert confidence interval to standard deviation
extract_SD <- function(x) {
  range <- as.numeric(str_split_1(str_sub(x, 2, -2), ","))
  diff(range) / 3.92
}

geo_tox_data$obesity <- places |>
  # North Carolina Counties
  filter(StateAbbr == "NC") |>
  # Select obesity data
  select(FIPS = CountyFIPS, OBESITY_CrudePrev, OBESITY_Crude95CI) |>
  # Change confidence interval to standard deviation
  rowwise() |>
  mutate(OBESITY_SD = extract_SD(OBESITY_Crude95CI)) |>
  ungroup() |>
  select(-OBESITY_Crude95CI)


## -----------------------------------------------------------------------------
#| label: C_ss
set.seed(2345)
n_samples <- 500

# Get CASN for which httk simulation is possible. Try using load_dawson2021,
# load_sipes2017, or load_pradeep2020 to increase availability.
load_sipes2017()
casn <- intersect(
  unique(geo_tox_data$dose_response$casn),
  get_cheminfo(suppress.messages = TRUE)
)

# Define population demographics for httk simulation
pop_demo <- cross_join(
  tibble(age_group = list(
    c(0, 2), c(3, 5), c(6, 10), c(11, 15), c(16, 20), c(21, 30), c(31, 40),
    c(41, 50), c(51, 60), c(61, 70), c(71, 100)
  )),
  tibble(weight = c("Normal", "Obese"))
)

# Create wrapper function around httk steps
simulate_css <- function(
  chem.cas,
  agelim_years,
  weight_category,
  samples,
  verbose = TRUE
) {
  if (verbose) {
    cat(
      chem.cas,
      paste0("(", paste(agelim_years, collapse = ", "), ")"),
      weight_category,
      "\n"
    )
  }

  httkpop <- list(
    method = "vi",
    gendernum = NULL,
    agelim_years = agelim_years,
    agelim_months = NULL,
    weight_category = weight_category,
    reths = c(
      "Mexican American",
      "Other Hispanic",
      "Non-Hispanic White",
      "Non-Hispanic Black",
      "Other"
    )
  )

  css <- try(
    suppressWarnings({
      mcs <- create_mc_samples(
        chem.cas = chem.cas,
        samples = samples,
        httkpop.generate.arg.list = httkpop,
        suppress.messages = TRUE
      )

      calc_analytic_css(
        chem.cas = chem.cas,
        parameters = mcs,
        model = "3compartmentss",
        suppress.messages = TRUE
      )
    }),
    silent = TRUE
  )

  # Return
  if (is(css, "try-error")) {
    warning("simulate_css() failed to generate data for CASN ", chem.cas)
    list(NA)
  } else {
    list(css)
  }
}

# Simulate C_ss values
simulated_css <- map(casn, function(chem.cas) {
  pop_demo |>
    rowwise() |>
    mutate(
      css = simulate_css(.env$chem.cas, age_group, weight, .env$n_samples)
    ) |>
    ungroup()
})
simulated_css <- setNames(simulated_css, casn)

# Remove CASN that failed simulate_css
casn_keep <- map_lgl(simulated_css, function(df) {
  !(length(df$css[[1]]) == 1 && is.na(df$css[[1]]))
})
simulated_css <- simulated_css[casn_keep]

geo_tox_data$simulated_css <- simulated_css |>
  map(\(x) {
    x |>
      mutate(
        age_lb = map_int(age_group, first),
        age_ub = map_int(age_group, last)
      ) |>
      select(age_lb, age_ub, weight, css) |>
      unnest(css)
  }) |>
  enframe(name = "casn") |>
  unnest(value)


## -----------------------------------------------------------------------------
#| label: retain
casn_keep <- unique(geo_tox_data$simulated_css$casn)

geo_tox_data$exposure <- geo_tox_data$exposure |>
  filter(casn %in% casn_keep)

geo_tox_data$dose_response <- geo_tox_data$dose_response |>
  filter(casn %in% casn_keep)


## -----------------------------------------------------------------------------
#| label: boundaries
county <- st_read("cb_2020_us_county_5m/cb_2020_us_county_5m.shp")
state <- st_read("cb_2020_us_state_5m/cb_2020_us_state_5m.shp")

geo_tox_data$boundaries <- list(
  county = county |>
    filter(STATEFP == 37) |>
    select(FIPS = GEOID, geometry),
  state = state |>
    filter(STATEFP == 37) |>
    select(geometry)
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, message=FALSE-----------------------------------------------------
library(dplyr)
library(sf)
library(tidyr)
library(readr)
library(stringr)
library(purrr)
library(readxl)
library(httk)
library(httr2)

geo_tox_data <- list()

## ----eval=FALSE---------------------------------------------------------------
# filename <- "2019_Toxics_Exposure_Concentrations.xlsx"
# tmp <- tempfile(filename)
# download.file(
#   paste0("https://www.epa.gov/system/files/documents/2022-12/", filename),
#   tmp
# )
# exposure <- read_xlsx(tmp)
# 
# # Normalization function
# min_max_norm = function(x) {
#   min_x <- min(x, na.rm = TRUE)
#   max_x <- max(x, na.rm = TRUE)
#   if (min_x == max_x) {
#     rep(0, length(x))
#   } else {
#     (x - min_x) / (max_x - min_x)
#   }
# }
# 
# geo_tox_data$exposure <- exposure |>
#   # North Carolina counties
#   filter(State == "NC", !grepl("0$", FIPS)) |>
#   # Aggregate chemicals by county
#   summarize(across(-c(State:Tract), c(mean, sd)), .by = FIPS) |>
#   pivot_longer(-FIPS, names_to = "chemical") |>
#   mutate(stat = if_else(grepl("_1$", chemical), "mean", "sd"),
#          chemical = gsub('.{2}$', '', chemical)) |>
#   pivot_wider(names_from = stat) |>
#   # Normalize concentrations
#   mutate(norm = min_max_norm(mean), .by = chemical)

## ----eval=FALSE---------------------------------------------------------------
# # Copy/paste the chemical names into the batch search field
# # https://comptox.epa.gov/dashboard/batch-search
# cat(geo_tox_data$exposure |> distinct(chemical) |> pull(), sep = "\n")
# # Export results with "CAS-RN" identifiers as a csv file, then process in R
# 
# exposure_casrn <- read_csv("CCD-Batch-Search.csv",
#                            show_col_types = FALSE) |>
#   filter(DTXSID != "N/A",
#          !grepl("WARNING", FOUND_BY))
# 
# # Update exposure data with CompTox Dashboard data
# geo_tox_data$exposure <- geo_tox_data$exposure |>
#   inner_join(exposure_casrn, by = join_by(chemical == INPUT)) |>
#   select(FIPS, casn = CASRN, chnm = PREFERRED_NAME, mean, sd, norm)

## ----eval=FALSE---------------------------------------------------------------
# get_cHTS_hits <- function(assays = NULL, chemids = NULL) {
# 
#   if (is.null(assays) & is.null(chemids)) {
#     stop("Must provide at least one of 'assays' or 'chemids'")
#   }
# 
#   # Format query parameters
#   req_params <- list()
# 
#   if (!is.null(assays)) {
#     if (!is.list(assays)) assays <- as.list(assays)
#     req_params$assays <- assays
#   }
# 
#   if (!is.null(chemids)) {
#     if (!is.list(chemids)) chemids <- as.list(chemids)
#     req_params$chemids <- chemids
#   }
# 
#   # Query ICE API
#   resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/search") |>
#     req_body_json(req_params) |>
#     req_perform()
# 
#   if (resp$status_code != 200) {
#     stop("Failed to retrieve data from ICE API")
#   }
# 
#   # Return active chemicals
#   result <- resp |> resp_body_json() |> pluck("endPoints")
# 
#   fields <- c("assay", "casrn", "dtxsid", "substanceName",
#               "endpoint", "value")
# 
#   map(fields, \(x) map_chr(result, x)) |>
#     set_names(fields) |>
#     bind_cols() |>
#     filter(endpoint == "Call", value == "Active") |>
#     select(-c(endpoint, value)) |>
#     distinct()
# }
# 
# assays <- c("APR_HepG2_p53Act_1hr",
#             "APR_HepG2_p53Act_24hr",
#             "APR_HepG2_p53Act_72hr",
#             "ATG_p53_CIS",
#             "TOX21_DT40_LUC",
#             "TOX21_DT40_100_LUC",
#             "TOX21_DT40_657_LUC",
#             "TOX21_ELG1_LUC_Agonist",
#             "TOX21_H2AX_HTRF_CHO_Agonist_ratio",
#             "TOX21_p53_BLA_p1_ratio",
#             "TOX21_p53_BLA_p2_ratio",
#             "TOX21_p53_BLA_p3_ratio",
#             "TOX21_p53_BLA_p4_ratio",
#             "TOX21_p53_BLA_p5_ratio")
# 
# chemids <- unique(geo_tox_data$exposure$casn)
# 
# cHTS_hits_API <- get_cHTS_hits(assays = assays, chemids = chemids)

## ----eval=FALSE---------------------------------------------------------------
# get_ICE_dose_resp <- function(assays = NULL, chemids = NULL) {
# 
#   if (is.null(assays) & is.null(chemids)) {
#     stop("Must provide at least one of 'assays' or 'chemids'")
#   }
# 
#   # Format query parameters
#   req_params <- list()
# 
#   if (!is.null(assays)) {
#     if (!is.list(assays)) assays <- as.list(assays)
#     req_params$assays <- assays
#   }
# 
#   if (!is.null(chemids)) {
#     if (!is.list(chemids)) chemids <- as.list(chemids)
#     req_params$chemids <- chemids
#   }
# 
#   # Query ICE API
#   resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/curves") |>
#     req_body_json(req_params) |>
#     req_perform()
# 
#   if (resp$status_code != 200) {
#     stop("Failed to retrieve data from ICE API")
#   }
# 
#   # Return dose-response data
#   result <- resp |> resp_body_json() |> pluck("curves")
# 
#   map(result, function(x) {
#     tibble(
#       endp = x[["assay"]],
#       casn = x[["casrn"]],
#       call = x[["dsstoxsid"]],
#       chnm = x[["substance"]],
#       call = x[["call"]],
#       logc = map_dbl(x[["concentrationResponses"]], "concentration") |> log10(),
#       resp = map_dbl(x[["concentrationResponses"]], "response")
#     )
#   }) |>
#     bind_rows()
# }
# 
# assays <- unique(cHTS_hits_API$assay)
# 
# chemids <- intersect(cHTS_hits_API$casrn, geo_tox_data$exposure$casn)
# 
# dose_response <- get_ICE_dose_resp(assays = assays, chemids = chemids)
# 
# # Only keep active calls for assay/chemical combinations
# geo_tox_data$dose_response <- dose_response |>
#   filter(call == "Active") |>
#   select(-call)
# 
# # Update dose-response data with CompTox Dashboard data
# geo_tox_data$dose_response <- geo_tox_data$dose_response |>
#   inner_join(exposure_casrn, by = join_by(casn == CASRN)) |>
#   select(endp, casn, chnm = PREFERRED_NAME, logc, resp)

## ----eval=FALSE---------------------------------------------------------------
# # Data for North Carolina
# url <- paste0("https://www2.census.gov/programs-surveys/popest/datasets/",
#               "2010-2020/counties/asrh/CC-EST2020-ALLDATA-37.csv")
# age <- read_csv(url, show_col_types = FALSE)
# 
# geo_tox_data$age <- age |>
#   # 7/1/2019 population estimate
#   filter(YEAR == 12) |>
#   # Create FIPS
#   mutate(FIPS = str_c(STATE, COUNTY)) |>
#   # Keep selected columns
#   select(FIPS, AGEGRP, TOT_POP)

## ----eval=FALSE---------------------------------------------------------------
# places <- read_csv("PLACES__County_Data_(GIS_Friendly_Format),_2020_release.csv",
#                    show_col_types = FALSE)
# 
# # Convert confidence interval to standard deviation
# extract_SD <- function(x) {
#   range <- as.numeric(str_split_1(str_sub(x, 2, -2), ","))
#   diff(range) / 3.92
# }
# 
# geo_tox_data$obesity <- places |>
#   # North Carolina Counties
#   filter(StateAbbr == "NC") |>
#   # Select obesity data
#   select(FIPS = CountyFIPS, OBESITY_CrudePrev, OBESITY_Crude95CI) |>
#   # Change confidence interval to standard deviation
#   rowwise() |>
#   mutate(OBESITY_SD = extract_SD(OBESITY_Crude95CI)) |>
#   ungroup() |>
#   select(-OBESITY_Crude95CI)

## ----eval=FALSE---------------------------------------------------------------
# set.seed(2345)
# n_samples <- 500
# 
# # Get CASN for which httk simulation is possible. Try using load_dawson2021,
# # load_sipes2017, or load_pradeep2020 to increase availability.
# load_sipes2017()
# casn <- intersect(unique(geo_tox_data$dose_response$casn),
#                   get_cheminfo(suppress.messages = TRUE))
# 
# # Define population demographics for httk simulation
# pop_demo <- cross_join(
#   tibble(age_group = list(c(0, 2), c(3, 5), c(6, 10), c(11, 15),
#                           c(16, 20), c(21, 30), c(31, 40), c(41, 50),
#                           c(51, 60), c(61, 70), c(71, 100))),
#   tibble(weight = c("Normal", "Obese"))) |>
#   # Create column of lower age_group values
#   rowwise() |>
#   mutate(age_min = age_group[1]) |>
#   ungroup()
# 
# # Create wrapper function around httk steps
# simulate_css <- function(chem.cas, agelim_years, weight_category,
#                          samples, verbose = TRUE) {
# 
#   if (verbose) {
#     cat(chem.cas,
#         paste0("(", paste(agelim_years, collapse = ", "), ")"),
#         weight_category,
#         "\n")
#   }
# 
#   httkpop <- list(method = "vi",
#                   gendernum = NULL,
#                   agelim_years = agelim_years,
#                   agelim_months = NULL,
#                   weight_category = weight_category,
#                   reths = c(
#                     "Mexican American",
#                     "Other Hispanic",
#                     "Non-Hispanic White",
#                     "Non-Hispanic Black",
#                     "Other"
#                   ))
# 
#   css <- try(
#     suppressWarnings({
#       mcs <- create_mc_samples(chem.cas = chem.cas,
#                                samples = samples,
#                                httkpop.generate.arg.list = httkpop,
#                                suppress.messages = TRUE)
# 
#       calc_analytic_css(chem.cas = chem.cas,
#                         parameters = mcs,
#                         model = "3compartmentss",
#                         suppress.messages = TRUE)
#     }),
#     silent = TRUE
#   )
# 
#   # Return
#   if (is(css, "try-error")) {
#     warning(paste0("simulate_css failed to generate data for CASN ", chem.cas))
#     list(NA)
#   } else {
#     list(css)
#   }
# }
# 
# # Simulate `C_ss` values
# simulated_css <- map(casn, function(chem.cas) {
#   pop_demo |>
#     rowwise() |>
#     mutate(
#       css = simulate_css(.env$chem.cas, age_group, weight, .env$n_samples)
#     ) |>
#     ungroup()
# })
# simulated_css <- setNames(simulated_css, casn)
# 
# # Remove CASN that failed simulate_css
# casn_keep <- map_lgl(simulated_css, function(df) {
#   !(length(df$css[[1]]) == 1 && is.na(df$css[[1]]))
# })
# simulated_css <- simulated_css[casn_keep]
# 
# # Get median `C_ss` values for each age_group
# simulated_css <- map(
#   simulated_css,
#   function(cas_df) {
#     cas_df |>
#       nest(.by = age_group) |>
#       mutate(
#         age_median_css = map_dbl(data, function(df) median(unlist(df$css),
#                                                            na.rm = TRUE))
#       ) |>
#       unnest(data)
#   }
# )
# 
# # Get median `C_ss` values for each weight
# simulated_css <- map(
#   simulated_css,
#   function(cas_df) {
#     cas_df |>
#       nest(.by = weight) |>
#       mutate(
#         weight_median_css = map_dbl(data, function(df) median(unlist(df$css),
#                                                               na.rm = TRUE))
#       ) |>
#       unnest(data) |>
#       arrange(age_min, weight)
#   }
# )
# 
# geo_tox_data$simulated_css <- simulated_css

## ----eval=FALSE---------------------------------------------------------------
# casn_keep <- names(geo_tox_data$simulated_css)
# 
# geo_tox_data$exposure <- geo_tox_data$exposure |>
#   filter(casn %in% casn_keep)
# 
# geo_tox_data$dose_response <- geo_tox_data$dose_response |>
#   filter(casn %in% casn_keep)

## ----eval=FALSE---------------------------------------------------------------
# county <- st_read("cb_2019_us_county_5m/cb_2019_us_county_5m.shp")
# state <- st_read("cb_2019_us_state_5m/cb_2019_us_state_5m.shp")
# 
# geo_tox_data$boundaries <- list(
#   county = county |>
#     filter(STATEFP == 37) |>
#     select(FIPS = GEOID, geometry),
#   state = state |>
#     filter(STATEFP == 37) |>
#     select(geometry)
# )

#' @export
#' @rdname get_risk_values
get_assay_table <- function(GT) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (!DBI::dbExistsTable(con, "assay")) {
    stop("Assay table not found in database.", call. = FALSE)
  }

  dplyr::tbl(con, "assay") |> dplyr::collect()
}

test_that("get assay table", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  expect_error(
    get_assay_table(GT),
    "Assay table not found in database."
  )
  assay_tbl <- GT |>
    add_assay(data.frame(name = "assay1", casn = "50-00-0")) |>
    get_assay_table()
  expect_equal(assay_tbl$name, "assay1")
})

#' Get risk metric results
#'
#' Several functions used to fetch risk metric results from risk tables in a
#' GeoTox database. The outputs of these functions are useful for plotting or
#' further analysis.
#'
#' Normally an 'assay' table is created when adding the Hill model
#' concentration-response fit parameters with [add_hill_params()]. If no `assay`
#' input is specified in [fit_hill()], then an 'assay' table will not be
#' created. For [get_risk_values()], the `assay` parameter can be used to filter
#' results based on assay details stored in the 'assay' table. This is useful
#' when multiple assays are available in the database. The `assay` input should
#' be a named vector specifying the column name and value to filter by, e.g.
#' `assay = c(endp = "mortality")`. For [get_risk_quantiles()], if there is no
#' assay data in the GeoTox database, then the output "assay_id" column will be
#' filled with NA values. If there is assay data, use [get_assay_table()] to
#' retrieve assay information and link the "assay_id" to assay details.
#'
#' Use the `table_name` parameter to specify which risk table to query. There
#' can be several risk tables in a GeoTox database. The default table is named
#' "risk" and is created by either [calc_risk()] or the wrapper function
#' [calc_response()]. Sensitivity analysis results created by either
#' [calc_sensitivity()] or the wrapper function [sensitivity_analysis()] are
#' stored in other tables with names like "risk_sensitivity_age", etc. Refer to
#' [calc_sensitivity()] to see which tables may be available.
#'
#' [get_risk_sensitivity()] is a wrapper function for [get_risk_values()] around
#' all risk tables created by the original risk computation and subsequent
#' sensitivity analysis. The column names are "baseline" for the original risk
#' table, while the other column names correspond to the parameter varied in the
#' sensitivity analysis.
#'
#' @param GT GeoTox object.
#' @param metric Risk metric to retrieve. One of "GCA.Eff", "IA.Eff",
#'   "GCA.HQ.10", or "IA.HQ.10".
#' @param assay Named vector specifying an assay filter (default NULL).
#' @param table_name Name of the risk table to query (default "risk").
#'
#' @returns A data frame or vector.
#' @export
#' @seealso [calc_risk()], [calc_response()], [calc_sensitivity()],
#'   [sensitivity_analysis()], [fit_hill()], [add_hill_params()]
#'
#' @examples
#' # Example setup is shown below in \dontrun().
#' # Pre-generated results will be loaded instead to avoid long example runtime.
#' 
#' \dontrun{
#' # Setup required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal", 21,
#'   "00-00-1", 50, 99, "Normal", 22,
#'   "00-00-1",  0, 49,  "Obese", 61,
#'   "00-00-1", 50, 99,  "Obese", 62,
#'   "00-00-2",  0, 49, "Normal", 11,
#'   "00-00-2", 50, 99, "Normal", 12,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' hill_df <- tibble::tribble(
#'   ~assay, ~model, ~casn, ~logc, ~resp,
#'   "a1", "human", "00-00-1",    0,  10,
#'   "a1", "human", "00-00-1",    1,  20,
#'   "a1", "human", "00-00-1",    2,  80,
#'   "a1", "human", "00-00-1",    3, 100,
#'   "a1", "human", "00-00-2", -0.5,   5,
#'   "a1", "human", "00-00-2",  0.5,  20,
#'   "a1", "human", "00-00-2",  1.5,  55,
#'   "a1", "human", "00-00-2",  2.5,  60,
#'   "a2",   "rat", "00-00-1",   -1,   0,
#'   "a2",   "rat", "00-00-1",    0,  10,
#'   "a2",   "rat", "00-00-1",    1,  30,
#'   "a2",   "rat", "00-00-1",    2,  40
#' )
#' set.seed(1234)
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   set_simulated_css(css_df) |>
#'   add_exposure_rate_params() |>
#'   add_hill_params(fit_hill(
#'     hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
#'   )) |>
#'   simulate_population(exposure = exposure_df) |>
#'   calc_response() |>
#'   sensitivity_analysis()
#' 
#' }
#'
#' # Load results from pre-generated database for this example
#' temp_dir <- tempdir()
#' zip::unzip(
#'   system.file("extdata", "sensitivity.duckdb.zip", package = "GeoTox"),
#'   junkpaths = TRUE,
#'   exdir = temp_dir
#' )
#' GT <- GeoTox(paste0(temp_dir, "/sensitivity.duckdb"))
#' 
#' # Look at 'assay' table contents
#' get_assay_table(GT)
#'
#' # Get "GCA.HQ.10" values from 'risk' table for the "a1" assay
#' get_risk_values(GT, metric = "GCA.HQ.10", assay = c(name = "a1"))
#'
#' # Get "IA.Eff" values from all risk tables for the "a2" assay
#' get_risk_sensitivity(GT, metric = "IA.Eff", assay = c(name = "a2"))
#'
#' # Get "GCA.Eff" quantiles from 'risk' table
#' get_risk_quantiles(GT, metric = "GCA.Eff", quantiles = c(0.25, 0.5, 0.75))
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at the 'risk' table contents
#' dplyr::tbl(con, "risk") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
get_risk_values <- function(
    GT,
    metric = c("GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"),
    assay = NULL,
    table_name = "risk"
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  metric <- match.arg(metric)

  if (!DBI::dbExistsTable(con, table_name)) {
    stop("No '", table_name, "' table found in the GeoTox connection.",
         call. = FALSE)
  }

  # Process assay input
  assay_tbl_exists <- DBI::dbExistsTable(con, "assay")
  if (!assay_tbl_exists & !is.null(assay)) {
    warning(
      "Assay '", assay, "' specified but 'assay' table not found in database. ",
      "Ignoring assay parameter.", call. = FALSE
    )
    assay <- NULL
  }
  if (assay_tbl_exists) {
    assay_tbl <- dplyr::tbl(con, "assay")
    assay_col <- names(assay)
    # Convert to char in case input was numeric, e.g. assay = c(id = 1)
    assay <- as.character(assay)
    if (is.null(assay_col) || (length(assay_col) != 1 | length(assay) != 1)) {
      stop(
        "A single assay must be specified with the 'assay' input as a named ",
        "vector, i.e. 'assay = c(column = value)'. Use 'get_assay_table()' ",
        "to see available columns and values.", call. = FALSE
      )
    }
    if (!assay_col %in% colnames(assay_tbl)) {
      stop(
        "Column '", assay_col, "' not found in assay table. Use ",
        "'get_assay_table()' to see available columns.", call. = FALSE
      )
    }
    # Rename "id" for after joining tables
    if (assay_col == "id") assay_col <- "assay_id"
  }

  out <- dplyr::tbl(con, table_name)
  if (assay_tbl_exists) {
    out <- out |>
      dplyr::left_join(
        dplyr::tbl(con, "assay"),
        by = dplyr::join_by("assay_id" == "id")
      ) |>
      dplyr::filter(.data[[assay_col]] == !!assay)
    if (dplyr::count(out) |> dplyr::pull() == 0) {
      stop(
        "No results found for assay '", assay, "'. Use 'get_assay_table()' ",
        "to see available assays.", call. = FALSE
      )
    }
  }
  out |>
    dplyr::arrange(.data$sample_id) |>
    dplyr::select(tidyselect::all_of(metric)) |>
    dplyr::pull()
}

#' Perform sensitivity analysis
#'
#' Calculate risk sensitivity to all available `vary` parameters in
#' [calc_sensitivity()].
#'
#' Sensitivity is calculated in the order: age, weight, css_params, fit_params,
#' C_ext. The `max_mult` vector allows specifying different upper bound
#' multipliers for each parameter.
#'
#' @param GT GeoTox object.
#' @param max_mult Vector of length 5 containing upper bound multipliers for max
#'   response (default 1.5).
#' @param ... Additional arguments passed to each call of [calc_sensitivity()].
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [calc_sensitivity()]
#'
#' @examples
#' # Example setup is shown below in \dontrun().
#' # Pre-generated results will be loaded instead to avoid long example runtime.
#' 
#' \dontrun{
#' # Setup required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal", 21,
#'   "00-00-1", 50, 99, "Normal", 22,
#'   "00-00-1",  0, 49,  "Obese", 61,
#'   "00-00-1", 50, 99,  "Obese", 62,
#'   "00-00-2",  0, 49, "Normal", 11,
#'   "00-00-2", 50, 99, "Normal", 12,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' hill_df <- tibble::tribble(
#'   ~assay, ~model, ~casn, ~logc, ~resp,
#'   "a1", "human", "00-00-1",    0,  10,
#'   "a1", "human", "00-00-1",    1,  20,
#'   "a1", "human", "00-00-1",    2,  80,
#'   "a1", "human", "00-00-1",    3, 100,
#'   "a1", "human", "00-00-2", -0.5,   5,
#'   "a1", "human", "00-00-2",  0.5,  20,
#'   "a1", "human", "00-00-2",  1.5,  55,
#'   "a1", "human", "00-00-2",  2.5,  60,
#'   "a2",   "rat", "00-00-1",   -1,   0,
#'   "a2",   "rat", "00-00-1",    0,  10,
#'   "a2",   "rat", "00-00-1",    1,  30,
#'   "a2",   "rat", "00-00-1",    2,  40
#' )
#' set.seed(1234)
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   set_simulated_css(css_df) |>
#'   add_exposure_rate_params() |>
#'   add_hill_params(fit_hill(
#'     hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
#'   )) |>
#'   simulate_population(exposure = exposure_df) |>
#'   calc_response()
#'
#' # Perform sensitivity analysis
#' GT <- GT |> sensitivity_analysis()
#' 
#' }
#' 
#' # Load results from pre-generated database for this example
#' temp_dir <- tempdir()
#' zip::unzip(
#'   system.file("extdata", "sensitivity.duckdb.zip", package = "GeoTox"),
#'   junkpaths = TRUE,
#'   exdir = temp_dir
#' )
#' GT <- GeoTox(paste0(temp_dir, "/sensitivity.duckdb"))
#' 
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant table
#'
#' dplyr::tbl(con, "risk_sensitivity_age") |> dplyr::collect()
#'
#' dplyr::tbl(con, "risk_sensitivity_weight") |> dplyr::collect()
#'
#' dplyr::tbl(con, "risk_sensitivity_css_params") |> dplyr::collect()
#'
#' dplyr::tbl(con, "risk_sensitivity_fit_params") |> dplyr::collect()
#'
#' dplyr::tbl(con, "risk_sensitivity_C_ext") |> dplyr::collect()
#'
#' # Compared to baseline risk table
#' dplyr::tbl(con, "risk") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
sensitivity_analysis <- function(
    GT, max_mult = c(1.5, 1.5, 1.5, 1.5, 1.5), ...
) {
  GT |>
    calc_sensitivity("age",        max_mult = max_mult[1], ...) |>
    calc_sensitivity("weight",     max_mult = max_mult[2], ...) |>
    calc_sensitivity("css_params", max_mult = max_mult[3], ...) |>
    calc_sensitivity("fit_params", max_mult = max_mult[4], ...) |>
    calc_sensitivity("C_ext",      max_mult = max_mult[5], ...)
  invisible(GT)
}

#' Simulate age values
#'
#' Simulate 'age' values to be stored in the 'sample' table of a GeoTox
#' database.
#'
#' An 'age' table containing simulation data must already exist in the GeoTox
#' database, which is added using [add_age()].
#'
#' @param GT GeoTox object.
#' @param n Number of individuals to simulate per location (default 1000).
#'   Ignored if 'sample' table already exists.
#' @param overwrite Logical indicating whether to overwrite existing 'age'
#'   values in the 'sample' table (default FALSE).
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [add_age()], [simulate_population()]
#'
#' @examples
#' # Example age simulation data
#' age_df <- data.frame(
#'   FIPS = rep(c(10000, 20000), each = 19),
#'   AGEGRP = rep(0:18, times = 2),
#'   TOT_POP = 0
#' )
#' # FIPS 10000, populate age group 40-44
#' age_df$TOT_POP[c(1, 10)] = 100
#' # FIPS 20000, populate age groups 50-59
#' age_df$TOT_POP[c(1, 12, 13) + 19] = c(200, 100, 100)
#'
#' # Simulate age values
#' GT <- GeoTox() |>
#'   add_age(age_df) |>
#'   simulate_age(n = 5)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' # Overwrite existing age values
#' GT <- GT |> simulate_age(overwrite = TRUE)
#'
#' # Look at updated 'sample' table
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
simulate_age <- function(GT, n = 1e3, overwrite = FALSE) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (!DBI::dbExistsTable(con, "age")) {
    stop("No 'age' table found in the GeoTox connection.", call. = FALSE)
  }

  # Collect all the age simulation input data
  age_cols <- c("location_id", "AGEGRP", "TOT_POP")
  age_df <- dplyr::tbl(con, "age") |>
    dplyr::select(tidyselect::all_of(age_cols)) |>
    dplyr::collect() |>
    tidyr::nest(.by = "location_id") |>
    dplyr::arrange(.data$location_id) |>
    dplyr::mutate(n_row = purrr::map_int(.data$data, nrow))
  if (any(age_df$n_row != 19)) {
    stop("The 'age' table must contain 19 rows for each location.",
         call. = FALSE)
  } else {
    age_df <- age_df |> dplyr::select(-c("n_row"))
  }

  # Simulate age data for each location
  if (!DBI::dbExistsTable(con, "sample")) {
    # Create table
    # Initialize IDs so update can be used instead of append for each location.
    # Appending to grow the table is much slower for large datasets.
    # if (inherits(con, "SQLiteConnection")) {
    #   RSQLite::initExtension(con, extension = "series")
    #   sql <- paste0(
    #     "CREATE TABLE sample AS ",
    #     "SELECT value AS id ",
    #     "FROM generate_series(1, ", n * nrow(age_df), ", 1)"
    #   )
    # } else {
    sql <- paste0(
      "CREATE TABLE sample AS ",
      "SELECT CAST(id AS INTEGER) AS id ",
      "FROM range(1, ", n * nrow(age_df) + 1, ") AS t(id)"
    )
    # }
    DBI::dbExecute(con, sql)
    # DBI::dbExecute(con, "ALTER TABLE sample ADD PRIMARY KEY (id)")
    DBI::dbExecute(con, "ALTER TABLE sample ADD COLUMN location_id INTEGER")
    DBI::dbExecute(con, "ALTER TABLE sample ADD COLUMN age INTEGER")
    # Generate ages by location
    age_df |>
      dplyr::mutate(idx = dplyr::row_number()) |>
      purrr::pwalk(\(location_id, data, idx) {
        update_df <- dplyr::bind_cols(
          id = ((idx - 1) * n + 1):(idx * n),
          location_id = location_id,
          age = .simulate_age(data, n)
        )
        update_table(con, "sample", update_df, reset_seed = GT$par$reset_seed)
      })
  } else {
    sample_tbl <- dplyr::tbl(con, "sample")
    if (!"age" %in% colnames(sample_tbl)) {
      DBI::dbExecute(con, "ALTER TABLE sample ADD COLUMN age INTEGER")
    } else if (!overwrite) {
      stop("GeoTox connection already has a 'sample' table with an 'age' ",
           "column. Set `overwrite = TRUE` to replace the existing values.\n",
           "Warning: Overwriting will not alter downstream values.",
           call. = FALSE)
    }
    # Generate ages by location
    age_df |>
      purrr::pwalk(\(location_id, data) {
        update_df <- sample_tbl |>
          dplyr::filter(.data$location_id == .env$location_id) |>
          dplyr::select("id") |>
          dplyr::arrange(.data$id) |>
          dplyr::collect()
        if (nrow(update_df) == 0) return()
        update_df <- update_df |>
          dplyr::mutate(age = .simulate_age(data, dplyr::n()))
        update_table(con, "sample", update_df, reset_seed = GT$par$reset_seed)
      })
  }
  invisible(GT)
}

.simulate_age <- function(x, n) {

  # Ensure order for AGEGRP
  x <- x[order(x$AGEGRP), ]

  # Probability of each age group
  prob <- x$TOT_POP[-1]

  # Assume equal probability within group
  prob <- rep(prob / 5, each = 5)

  # Sample ages
  sample(0:89, size = n, prob = prob, replace = TRUE)
}

#' Store and retrieve boundary geometries
#'
#' Boundary information is stored as serialized 'BLOB' objects in the 'boundary'
#' table.
#'
#' NOTE: This function requires the `sf` package to be installed.
#'
#' This function takes a named list of sf objects and stores them in the
#' database. If a location boundary (default "county") is provided, then the
#' non-geometry fields will be added to the 'location' table and replaced with a
#' 'location_id' value so they can be linked to data in the 'sample' table.
#'
#' @param GT GeoTox object.
#' @param df_list Named list of data frames containing boundary geometries as sf
#'   objects.
#' @param location Name of element in `df_list` that contains location boundary
#'   information (default "county").
#' @param overwrite Logical indicating whether to overwrite existing 'boundary'
#'   table.
#'
#' @returns For `set_boundary()`, the same GeoTox object, invisibly. For
#'   `get_boundary()`, a data frame with columns 'id' (boundary name) and 'data'
#'   (sf object).
#' @export
#'
#' @examples
#' # Setup sf objects
#' county <- sf::st_sf(
#'   FIPS = c(10000, 20000),
#'   geometry = sf::st_sfc(sf::st_point(1:2), sf::st_point(3:4))
#' )
#' state <- sf::st_sf(
#'   STATE = "XYZ",
#'   geometry = sf::st_sfc(sf::st_point(5:6))
#' )
#' df_list <- list(county = county, state = state)
#'
#' # Add boundary to GeoTox database
#' GT <- GeoTox() |> set_boundary(df_list)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "boundary") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' # Retrieve boundary from GeoTox database
#' boundary <- get_boundary(GT)
#' boundary
#' boundary |> tibble::deframe()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
set_boundary <- function(GT, df_list, location = "county", overwrite = FALSE) {
  if (!rlang::is_installed("sf")) { # nocov start
    stop(
      "Package \"sf\" must be installed to use this function.",
      call. = FALSE
    )
  } # nocov end
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  if (!inherits(df_list, "list") ||
      is.null(names(df_list)) ||
      !all(sapply(df_list, inherits, "sf"))) {
    stop("`df_list` must be a named list of sf objects.", call. = FALSE)
  }
  # Link df_list[[location]] to location table
  if (!is.null(location) && !is.null(df_list[[location]])) {
    # Isolate geometry field
    data <- sf::st_drop_geometry(df_list[[location]])
    geom <- sf::st_geometry(df_list[[location]])
    # Add non-geometry fields to the location table
    add_table(con, "location", data, reset_seed = GT$par$reset_seed)
    # Replace non-geometry fields with location foreign key
    loc <- dplyr::tbl(con, "location") |> dplyr::collect()
    loc_id <- data |>
      dplyr::left_join(loc, by = intersect(names(data), names(loc))) |>
      dplyr::pull("id")
    df_list[[location]] <- sf::st_as_sf(data.frame(location_id = loc_id, geom))
  }
  # Serialize the sf objects to BLOBs for storage in the database
  df <- data.frame(
    id = names(df_list),
    data = blob::new_blob(purrr::map(df_list, \(x) serialize(x, NULL)))
  )
  if (DBI::dbExistsTable(con, "boundary") & !overwrite) {
    stop("GeoTox connection already has a 'boundary' table. ",
         "Set `overwrite = TRUE` to replace it.", call. = FALSE)
  }
  write_table(
    con,
    "boundary",
    df,
    field.types = c(data = "BLOB"),
    overwrite = overwrite,
    reset_seed = GT$par$reset_seed
  )
  invisible(GT)
}

#' @param GT GeoTox object.
#' @export
#' @rdname set_boundary
get_boundary <- function(GT) {
  if (!rlang::is_installed("sf")) { # nocov start
    stop(
      "Package \"sf\" must be installed to use this function.",
      call. = FALSE
    )
  } # nocov end
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  if (!DBI::dbExistsTable(con, "boundary")) {
    stop("No 'boundary' table found in the GeoTox connection.", call. = FALSE)
  }
  dplyr::tbl(con, "boundary") |>
    dplyr::collect() |>
    dplyr::mutate(data = purrr::map(.data$data, unserialize))
}

test_that("boundary table", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup sf objects
  county <- sf::st_sf(
    FIPS = c("00001", "00002"),
    geometry = sf::st_sfc(sf::st_point(1:2), sf::st_point(3:4))
  )
  state <- sf::st_sf(
    STATE = "XYZ",
    geometry = sf::st_sfc(sf::st_point(5:6))
  )
  df_list <- list(county = county, state = state)

  # Bad df_list
  expect_error(
    set_boundary(GT, list(county = data.frame(a = 1))),
    "`df_list` must be a named list of sf objects."
  )

  # No 'boundary' table
  expect_error(
    get_boundary(GT),
    "No 'boundary' table found in the GeoTox connection."
  )

  # Set boundary
  set_boundary(GT, df_list)

  # Overwrite
  expect_error(
    set_boundary(GT, df_list),
    "GeoTox connection already has a 'boundary' table."
  )
  set_boundary(GT, df_list, overwrite = TRUE)

  # Get boundary
  res <- get_boundary(GT) |> tibble::deframe()
  expect_equal(res$county$geometry, county$geometry)
  expect_equal(res$county$location_id, 1:nrow(county))
  expect_equal(res$state$geometry, state$geometry)
  expect_equal(res$state$STATE, state$STATE)
})

#' Sample from pre-simulated steady-state plasma concentrations
#'
#' Sample steady-state plasma concentrations (C\eqn{_{ss}}) for individuals from
#' pre-simulated values stored in the 'simulated_css' table in a GeoTox
#' database.
#'
#' The C\eqn{_{ss}} values are sampled from the 'simulated_css' table, which
#' must already exist in the GeoTox database; it can be created using
#' [set_simulated_css()]. The minimum characteristics that are used to match
#' individuals to sets of C\eqn{_{ss}} values are age and weight category
#' ("Normal" or "Obese"). Additional columns (which must exist in both the
#' 'simulated_css' and 'sample' tables) can be specified using the
#' `css_extra_cols` argument. If `css_exta_cols` is provided it will be added to
#' the GeoTox object parameter list, `GT$par`.
#'
#' In addition to the 'simulated_css' table, both 'sample' and 'concentration'
#' tables must also exist in the GeoTox database. The 'sample' table must
#' contain the characteristics of individuals (age, weight category, and any
#' additional columns specified in `css_extra_cols`). One way to create this
#' 'sample' table is by using [set_sample()]. The 'concentration' table will
#' typically be created using [simulate_exposure()] and will be populated with
#' rows for each individual and substance combination where exposure data is
#' available. The sampled C\eqn{_{ss}} values will be stored in a new "C_ss"
#' column in the 'concentration' table.
#'
#' The `substance_order` argument can be used to specify the order in which
#' substances are evaluated when sampling C\eqn{_{ss}} values. The default is to
#' evaluate substances in the order they appear in the 'substance' table.
#' However, if a different order is needed for some reason (e.g., replication of
#' results from a previous GeoTox implementation), the user can provide a named
#' list where the name is the column in the 'substance' table to use for
#' ordering (e.g., "casn") and the value is a vector of substance identifiers in
#' the desired order. If provided, the `substance_order` will be added to the
#' GeoTox object parameter list, `GT$par`.
#'
#' @param GT GeoTox object.
#' @param css_extra_cols Additional columns to match from the 'simulated_css'
#'   table (default NULL).
#' @param substance_order Named list specifying order of substance evaluation
#'   (default NULL).
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [set_simulated_css()], [simulate_population()]
#'
#' @examples
#' # Create required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' # Note: normally the css_df would have many more rows for each combination of
#' # the non-'css' columns to allow for sampling.
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal",  1,
#'   "00-00-1", 50, 99, "Normal",  2,
#'   "00-00-1",  0, 49,  "Obese", 11,
#'   "00-00-1", 50, 99,  "Obese", 12,
#'   "00-00-2",  0, 49, "Normal", 21,
#'   "00-00-2", 50, 99, "Normal", 22,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   add_exposure(exposure_df) |>
#'   simulate_exposure() |>
#'   set_simulated_css(css_df)
#'
#' # Sample simulated C_ss values
#' GT <- GT |> sample_simulated_css()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables. sample_simulated_css() generated the 'C_ss' column
#' # of the 'concentration' table.
#'
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#'
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Replace sample and css tables with new data including an extra column
#' # Limit to a single substance for simplicity
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight, ~sign,
#'   10000, 25, "Normal", "+",
#'   10000, 35,  "Obese", "-",
#'   20000, 50, "Normal", "+"
#' )
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css, ~sign,
#'   "00-00-1",  0, 49, "Normal",   1, "+",
#'   "00-00-1", 50, 99, "Normal",   2, "+",
#'   "00-00-1",  0, 49,  "Obese",  11, "+",
#'   "00-00-1", 50, 99,  "Obese",  12, "+",
#'   "00-00-1",  0, 49, "Normal",  -1, "-",
#'   "00-00-1", 50, 99, "Normal",  -2, "-",
#'   "00-00-1",  0, 49,  "Obese", -11, "-",
#'   "00-00-1", 50, 99,  "Obese", -12, "-"
#' )
#' GT <- GT |>
#'  set_sample(sample_df, overwrite = TRUE) |>
#'  simulate_exposure() |>
#'  set_simulated_css(css_df, overwrite = TRUE)
#'
#' # Sample simulated C_ss values with extra column
#' # Notice how the extra column name is added to GT$par
#' str(GT$par)
#' GT <- GT |> sample_simulated_css(css_extra_cols = "sign")
#' str(GT$par)
#'
#' # Look at new 'concentration' table. Values will be missing for substance 2
#' # since it is not in the new css_df.
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
sample_simulated_css <- function(
    GT, css_extra_cols = NULL, substance_order = NULL
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  tables <- DBI::dbListTables(con)
  if (!"simulated_css" %in% tables) {
    stop("No 'simulated_css' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!"concentration" %in% tables) {
    stop("No 'concentration' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!"sample" %in% tables) {
    stop("No 'sample' table found in the GeoTox connection.", call. = FALSE)
  }

  # Update parameters
  if (!is.null(css_extra_cols)) {
    set_par(con, "css_extra_cols", css_extra_cols)
    GT$par$css_extra_cols <- css_extra_cols
  } else if (!is.null(GT$par$css_extra_cols)) {
    css_extra_cols <- GT$par$css_extra_cols
  } else {
    css_extra_cols <- c()
  }
  if (!is.null(substance_order)) {
    substance_order_col <- GT$par$substance_order_col <- names(substance_order)
    substance_order     <- GT$par$substance_order     <- substance_order[[1]]
    set_par(con, "substance_order_col", substance_order_col)
    set_par(con, "substance_order",     substance_order)
  } else if (!is.null(GT$par$substance_order_col)) {
    substance_order_col <- GT$par$substance_order_col
    substance_order     <- GT$par$substance_order
  } else {
    substance_order_col <- NULL
  }

  if (is.null(substance_order_col)) {
    substance_order <- dplyr::tbl(con, "substance") |>
      dplyr::pull("id")
  } else {
    substance_order <- dplyr::tbl(con, "substance") |>
      dplyr::select(tidyselect::all_of(c("id", substance_order_col))) |>
      dplyr::collect() |>
      dplyr::rename(value = 2) |>
      dplyr::arrange(match(.data$value, substance_order)) |>
      dplyr::pull("id")
  }

  css_tbl    <- dplyr::tbl(con, "simulated_css")
  conc_tbl   <- dplyr::tbl(con, "concentration")
  sample_tbl <- dplyr::tbl(con, "sample")

  if (!"C_ss" %in% colnames(conc_tbl)) {
    DBI::dbExecute(con, "ALTER TABLE concentration ADD COLUMN C_ss DOUBLE")
  }

  # Collect simulated C_ss data
  css_cols <- c(
    "id", "substance_id", "age_lb", "age_ub", "weight", "css", css_extra_cols
  )
  css_df <- css_tbl |>
    dplyr::select(tidyselect::all_of(css_cols)) |>
    dplyr::collect() |>
    tidyr::nest(css = c("id", "css")) |>
    dplyr::mutate(
      css = purrr::map(.data$css, \(df) {
        df |> dplyr::arrange(.data$id) |> dplyr::pull("css")
      }),
      idx = dplyr::row_number(),
      age_ub_p1 = .data$age_ub + 1
    )

  conc_cols <- c("id", "sample_id", "substance_id")
  sample_cols <- c("id", "location_id", "age", "weight", css_extra_cols)

  # Sample C_ss data for each location
  location_id <- dplyr::tbl(con, "location") |>
    dplyr::select("id") |>
    dplyr::arrange(.data$id) |>
    dplyr::pull()
  purrr::walk(location_id, function(location_id) {
    conc_df <- conc_tbl |>
      dplyr::select(tidyselect::all_of(conc_cols)) |>
      dplyr::left_join(
        sample_tbl |> dplyr::select(tidyselect::all_of(sample_cols)),
        by = dplyr::join_by("sample_id" == "id")
      ) |>
      dplyr::filter(.data$location_id == .env$location_id) |>
      dplyr::select(-c("location_id")) |>
      dplyr::collect()
    if (nrow(conc_df) == 0) return()
    update <- dplyr::inner_join(
      conc_df,
      css_df |> dplyr::select(-c("css")),
      by = dplyr::join_by(
        "substance_id",
        "age" >= "age_lb",
        "age" < "age_ub_p1",
        "weight",
        !!!rlang::syms(css_extra_cols)
      )
    ) |>
      dplyr::select("id", "sample_id", "substance_id", "idx")
    if (nrow(update) == 0) {
      stop("No matching rows found in 'simulated_css' for the 'concentration' ",
           "and 'sample' tables. Ensure that the substance_id, age, weight, ",
           "and any extra columns match between the tables.", call. = FALSE)
    }
    update <- update |>
      dplyr::arrange(.data$sample_id) |>
      tidyr::nest(data = "id") |>
      dplyr::arrange(match(.data$substance_id, substance_order), .data$idx) |>
      dplyr::mutate(
        C_ss = purrr::map2(.data$data, .data$idx, \(data, idx) {
          # Note: if input vector, x, to sample() is a single numeric value,
          # then sample() will sample from 1:x. Here, we want n values of x,
          # not n samples from 1:x. See Details of ?sample for more info.
          css <- css_df$css[[idx]]
          n <- nrow(data)
          if (length(css) == 1) return(rep(css, n))
          sample(css, n, replace = TRUE)
        })
      ) |>
      dplyr::select("data", "C_ss") |>
      tidyr::unnest(c("data", "C_ss"))
    update_table(con, "concentration", update, reset_seed = GT$par$reset_seed)
  })
  invisible(GT)
}

#' Calculate concentrations and risks
#'
#' Calculate internal dose, in vitro concentration, and risk estimates.
#'
#' This is a wrapper around several other functions:
#' * [calc_internal_dose()] to calculate internal dose (D_int).
#' * [calc_invitro_concentration()] to calculate in vitro concentration
#' (C_invitro).
#' * [calc_risk()] to calculate risk estimates.
#'
#' If a `risk_name` argument is provided to `...` and it is not "risk", then
#' sensitivity analysis is assumed and the `sensitivity` argument in the
#' internal dose and in vitro concentration calculations is set to TRUE.
#'
#' @param GT GeoTox object.
#' @param overwrite Logical indicating whether to overwrite existing values
#'   (default FALSE).
#' @param ... Additional arguments passed to [calc_risk()].
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [calc_internal_dose()], [calc_invitro_concentration()],
#'   [calc_risk()]
#'
#' @examples
#' # Setup required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal", 21,
#'   "00-00-1", 50, 99, "Normal", 22,
#'   "00-00-1",  0, 49,  "Obese", 61,
#'   "00-00-1", 50, 99,  "Obese", 62,
#'   "00-00-2",  0, 49, "Normal", 11,
#'   "00-00-2", 50, 99, "Normal", 12,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' hill_df <- tibble::tribble(
#'   ~assay, ~casn, ~logc, ~resp,
#'   "a1", "00-00-1",    0,  10,
#'   "a1", "00-00-1",    1,  20,
#'   "a1", "00-00-1",    2,  80,
#'   "a1", "00-00-1",    3, 100,
#'   "a1", "00-00-2", -0.5,   5,
#'   "a1", "00-00-2",  0.5,  20,
#'   "a1", "00-00-2",  1.5,  55,
#'   "a1", "00-00-2",  2.5,  60
#' )
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   set_simulated_css(css_df) |>
#'   add_exposure_rate_params() |>
#'   add_hill_params(fit_hill(hill_df, assay = "assay", substance = "casn")) |>
#'   simulate_population(exposure = exposure_df)
#'
#' # Calculate concentrations and risk
#' GT <- GT |> calc_response()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant table
#'
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' dplyr::tbl(con, "risk") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
calc_response <- function(GT, overwrite = FALSE, ...) {
  dots <- list(...)
  if ("risk_name" %in% names(dots) && dots$risk_name != "risk") {
    sensitivity <- TRUE
  } else {
    sensitivity <- FALSE
  }
  GT <- GT |>
    calc_internal_dose(
      overwrite = overwrite, sensitivity = sensitivity
    ) |>
    calc_invitro_concentration(
      overwrite = overwrite, sensitivity = sensitivity
    ) |>
    calc_risk(overwrite = overwrite, ...)
  invisible(GT)
}

#' A Generalized concentration addition style solution
#'
#' Find the effective concentration of a mixture via an objective function
#' given the concentrations and inverse, based on a regular space AC50 and
#' concentrations.
#'
#' @param conc_mix Effective concentration of the mixture in regular space.
#' @param resp Individual chemical responses.
#' @param conc Individual chemical concentrations in regular space.
#' @param max Maximal (asymtotic) response.
#' @param AC50 Concentrations of half-maximal response.
#'
#' @returns Objective value.
#' @noRd
obj_ECx <- function(conc_mix, resp, conc, max, AC50) {
  x <- hill_conc(resp, max, AC50, 1)
  p <- conc / sum(conc)
  ECx.val <- sum(p * conc_mix / x, na.rm = FALSE)
  (ECx.val - 1)^2
}

test_that("spot check", {
  
  conc_mix <- 2
  resp <- 1
  conc <- c(1, 3)
  max <- resp * 2
  AC50 <- conc
  
  expect_equal(obj_ECx(conc_mix, resp, conc, max, AC50), 0)
  
  conc_mix <- 2
  resp <- c(1, 2)
  conc <- c(1, 3)
  max <- c(3, 4)
  AC50 <- 1
  
  expect_equal(obj_ECx(conc_mix, resp, conc, max, AC50), 2.25)
  
})

#' Set sample data
#'
#' Create the 'sample' table in the GeoTox database.
#'
#' The 'sample' table consists of individual characteristics and location
#' information. At a minimum, it must contain columns with information on age,
#' weight category, and location identifier(s).
#'
#' When the `df` input contains information on age or weight category, the
#' column names must be "age" and "weight", respectively. The `location` input
#' (default "FIPS") can be a named vector to specify multiple identifier columns
#' in `df`. For example, `location = c(FIPS = "FIPS", state = "ST")` would
#' indicate that `df` contains both FIPS codes and state identifiers for
#' locations. The location column(s) will be added to the 'location' table and
#' replaced with a corresponding "location_id" column in the 'sample' table. The
#' `state = "ST"` part in the example above would rename the "ST" column in `df`
#' to "state" in the 'location' table.
#'
#' There are several other functions that can be used to create or add to the
#' 'sample' table: [simulate_age()] to generate age data, [simulate_obesity()]
#' to generate weight category data, and [simulate_population()], which is a
#' wrapper function that can generate both age and weight category data along
#' with other fields. `set_sample()` can be used in combination with these
#' functions to first set some known sample data, e.g. age, and then simulate
#' any missing fields, e.g. weight category.
#'
#' If `overwrite = TRUE`, any existing 'concentration' and 'risk' tables will be
#' dropped before creating the 'sample' table. This is because the existing
#' concentration and risk data would no longer be valid for the new sample data.
#' Further downstream tables are not dropped automatically, but can be updated
#' during subsequent simulation and analysis steps.
#'
#' @param GT GeoTox object.
#' @param df Data frame containing sample data.
#' @param location Column name(s) in `df` that contain location identifier(s)
#'   (default "FIPS").
#' @param overwrite Logical indicating whether to overwrite existing 'sample'
#'   table and remove existing 'concentration' and 'risk' tables (default
#'   FALSE).
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [simulate_age()], [simulate_obesity()], [simulate_population()]
#'
#' @examples
#' # Example sample data
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#'
#' # Set sample data
#' GT <- GeoTox() |> set_sample(sample_df)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
set_sample <- function(GT, df, location = "FIPS", overwrite = FALSE) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  foreign_keys <- list("location_id" = location)
  check_foreign_keys(df, foreign_keys)
  conc_exists <- DBI::dbExistsTable(con, "concentration")
  risk_exists <- DBI::dbExistsTable(con, "risk")
  if (overwrite) {
    if (conc_exists) {
      DBI::dbRemoveTable(con, "concentration")
    }
    if (risk_exists) {
      DBI::dbRemoveTable(con, "risk")
    }
  } else {
    if (conc_exists | risk_exists) {
      stop("The 'concentration' or 'risk' tables already exist. ",
           "Use `overwrite = TRUE` to drop them.",
           call. = FALSE)
    }
  }
  df <- fetch_foreign_keys(con, df, foreign_keys, GT$par$reset_seed)
  overwrite_table(
    con, "sample", df, foreign_keys, reset_seed = GT$par$reset_seed
  )
  invisible(GT)
}

#' Get concentration mean values
#'
#' Calculates the concentration mean values for each substance and route at each
#' location based on the data in the 'concentration' and 'sample' tables of the
#' GeoTox database. The output of this function is useful for plotting or
#' further analysis.
#'
#' The `col` parameter specifies which column in the 'concentration' table to
#' use for the mean calculation. The available choices will depend on what data
#' has been stored in the 'concentration' table during the simulation process.
#'
#' @param GT GeoTox object.
#' @param col Column name in the 'concentration' table for which to calculate
#'   the mean, grouped by substance, route, and location.
#'
#' @returns A data frame.
#' @export
#'
#' @examples
#' # Setup required tables
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' GT <- GeoTox() |>
#'   add_exposure(exposure_df) |>
#'   simulate_exposure(n = 100)
#'
#' # Calculate mean external concentration by substance and location
#' get_concentration_mean(GT, "C_ext")
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at column names in the 'concentration' table
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
get_concentration_mean <- function(GT, col) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (!DBI::dbExistsTable(con, "concentration")) {
    stop("No 'concentration' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!DBI::dbExistsTable(con, "sample")) {
    stop("No 'sample' table found in the GeoTox connection.", call. = FALSE)
  }

  out <- dplyr::tbl(con, "concentration") |>
    dplyr::select(tidyselect::all_of(
      c("substance_id", "route_id", "sample_id", col)
    )) |>
    # Get location info from sample table
    dplyr::left_join(
      dplyr::tbl(con, "sample"),
      by = dplyr::join_by("sample_id" == "id")
    ) |>
    duckplyr::as_duckdb_tibble()

  # Get summary stat
  out |>
    dplyr::summarize(
      mean = mean(.data[[col]], na.rm = TRUE),
      .by = c("substance_id", "route_id", "location_id")
    ) |>
    dplyr::collect() |>
    # Fill in missing value with NA
    # tidyr::complete(
    #   tidyr::nesting(substance_id, route_id),
    #   .data$location_id,
    #   fill = list(mean = NA_real_)
    # )
    # Notes:
    # tidyr::nesting(substance_id, route_id) has R CMD check warnings
    # tidyr::nesting(.data$substance_id, .data$route_id) cannot find columns
    # So, complete using all columns, then filter out unwanted combinations
    tidyr::complete(
      .data$substance_id,
      .data$route_id,
      .data$location_id,
      fill = list(mean = NA_real_)
    ) |>
    dplyr::filter(!all(is.na(.data$mean)), .by = c("substance_id", "route_id"))
}

#' Hill model response
#'
#' Calculate the response for a given concentration in regular space.
#'
#' This is a regular space version of
#' [tcpl::tcplHillVal()](https://cran.r-project.org/package=tcpl).
#'
#' The Hill model is defined as:
#' \deqn{resp = \frac{max}{1 + (\frac{AC50}{conc})^{n}}}
#'
#' @param conc Concentration in regular space.
#' @param max Maximal (asymptotic) response.
#' @param AC50 Concentration of half-maximal response.
#' @param n Hill coefficient (slope).
#'
#' @returns Response value.
#' @noRd
hill_val <- function(conc, max, AC50, n) {

  if (any(conc < 0) || any(AC50 < 0)) {
    stop("Input concentrations can't be negative.", call. = FALSE)
  }

  max / (1 + (AC50 / conc)^n)

}

test_that("spot check", {
  max <- 100
  expect_equal(hill_val(10,  max, 10, 1),  max /  2)
  expect_equal(hill_val( 3,  max,  9, 1),  max /  4)
  expect_equal(hill_val( 1, -max,  9, 1), -max / 10)
  expect_equal(hill_val( 5,  max, 10, 2),  max /  5)
})

test_that("corner cases", {
  expect_equal(hill_val(c(0, 5, Inf), 2, 5, 1), 0:2)
})

test_that("negatives", {
  expect_error(hill_val(-1, 2,  1, 1))
  expect_error(hill_val( 1, 2, -1, 1))
})

#' Objective function for 2- or 3-parameter Hill model
#'
#' @param par Parameters.
#' @param log10_conc Base-10 log scale concentration.
#' @param resp Response.
#'
#' @returns Objective value.
#' @noRd
obj_hill <- function(par, log10_conc, resp) {

  max        <- par[1]  # maximal response
  log10_AC50 <- par[2]  # log10(concentration of half-maximal response)
  if (length(par) == 3) {
    n        <- 1       # Hill coefficient (slope)
    err      <- par[3]
  } else {
    n        <- par[3]
    err      <- par[4]
  }

  y <- max / (1 + 10^(n * (log10_AC50 - log10_conc)))

  sum(stats::dt((resp - y) / exp(err), df = length(par), log = TRUE) - err)
}

test_that("spot check", {
  
  par <- c(2, 3, dt(0, df = 3, log = TRUE))
  log10_conc <- par[2]
  resp <- rep(par[1] / 2, 5)
  
  expect_equal(obj_hill(par, log10_conc, resp), 0)
  
  par <- c(2, 3, 4, dt(0, df = 4, log = TRUE))
  log10_conc <- par[2]
  resp <- rep(par[1] / 2, 7)
  
  expect_equal(obj_hill(par, log10_conc, resp), 0)
  
})

# Testing the obj-hill functions
test_that("2 and 3 parameter models give accurate likelihood", {

## hill function for simulation
  hill_fun <- function(x, alpha, beta, theta){
    alpha / (1 + 10^(beta * (theta - x)))
  }

## Creating regular parameters
  tp <- 1
  log10_ga <- 1
  gw <- 2
  err <- 1

## 2-parameter model

  #simulate dose-response
  params2 <- c(tp, log10_ga, err)
  log10_conc <- seq(0.1,10,length.out = 50)
  resp <- hill_fun(log10_conc, tp, log10_ga, gw)

  # 2 parameter hill model expectation
  val <- obj_hill(params2, log10_conc, resp)
  expect_equal(val, -100.201, tolerance = 0.001)

## 3-parameter model

  # 3 parameter hill model expectation
  params3 <- c(tp, log10_ga, gw, err)
  val <- obj_hill(params3, log10_conc, resp)
  expect_equal(val, -99.23463, tolerance = 0.001)


#### Edge Tests

# Really tiny parameter values

  #simulate dose-response
  params_tiny <- c(1e-10, 1e-10, 1e-10, 1e-10)

  expect_no_warning(obj_hill(params_tiny, log10_conc, resp))


  # Really large parameter values

  #simulate dose-response
  params_large <- c(1e10, 1e10, 1e10, 1e10)

  expect_no_warning(obj_hill(params_tiny, log10_conc, resp))
})

#' Prepare steady-state plasma concentrations for sensitivity analysis
#'
#' Create the 'fixed_css' table in the GeoTox database, which contains values of
#' steady-state plasma concentrations (C\eqn{_{ss}}) for sensitivity analysis.
#'
#' Several tables are required in the GeoTox database before the 'fixed_css'
#' table can be created. Typically, `set_fixed_css()` is called after using
#' [sample_simulated_css()], at which point all required tables will be present.
#'
#' The resulting 'fixed_css' table is used for sensitivity analysis where one
#' parameter is allowed to vary at a time while all other parameters are held
#' constant at fixed values.
#'
#' The `substance_order` argument can be used to specify the order in which
#' substances are evaluated when sampling C\eqn{_{ss}} values. The default is to
#' evaluate substances in the order they appear in the 'substance' table.
#' However, if a different order is needed for some reason (e.g., replication of
#' results from a previous GeoTox implementation), the user can provide a named
#' list where the name is the column in the 'substance' table to use for
#' ordering (e.g., "casn") and the value is a vector of substance identifiers in
#' the desired order. If provided, the `substance_order` will be added to the
#' GeoTox object parameter list, `GT$par`. This is only applicable to the
#' "params" section described below.
#'
#' ## Table 'fixed_css' columns:
#' \describe{
#' \item{age}{Median pre-simulated C\eqn{_{ss}} values are computed by age
#' group for each substance, then the median C\eqn{_{ss}} values are assigned
#' to each individual based on their age group.}
#' \item{weight}{Median pre-simulated C\eqn{_{ss}} values are computed by weight
#' category for each substance, then the median C\eqn{_{ss}} values are assigned
#' to each individual based on their weight category.}
#' \item{params}{Individuals are assigned the median age of their location and
#' pre-simulated C\eqn{_{ss}} values for the "Normal" weight category are
#' sampled.}
#' \item{other}{Median sampled C\eqn{_{ss}} values are computed across all
#' substances for each location after mean-imputation of missing C\eqn{_{ss}}
#' values for each substance and location. The median C\eqn{_{ss}} values are
#' assigned to each individual based on their location.}
#' }
#'
#' @param GT GeoTox object.
#' @param substance_order Named list specifying order of substance evaluation
#'   (default NULL).
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [sample_simulated_css()], [simulate_population()]
#'
#' @examples
#' # Create required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' # Note: normally the css_df would have many more rows for each combination of
#' # the non-'css' columns to allow for sampling.
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal",  1,
#'   "00-00-1", 50, 99, "Normal",  2,
#'   "00-00-1",  0, 49,  "Obese", 11,
#'   "00-00-1", 50, 99,  "Obese", 12,
#'   "00-00-2",  0, 49, "Normal", 21,
#'   "00-00-2", 50, 99, "Normal", 22,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   add_exposure(exposure_df) |>
#'   simulate_exposure() |>
#'   set_simulated_css(css_df) |>
#'   sample_simulated_css()
#'
#' # Set fixed C_ss values
#' GT <- GT |> set_fixed_css()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#' # Note: the 'age', 'weight', 'params', and 'other' columns of the
#' # 'fixed_css' table contain the C_ss values for sensitivity analysis.
#' # For example, the 'age' column doesn't contain ages, but C_ss values.
#'
#' dplyr::tbl(con, "fixed_css") |> dplyr::collect()
#'
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
set_fixed_css <- function(GT, substance_order = NULL) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  tables <- DBI::dbListTables(con)
  if (!"sample" %in% tables) {
    stop("No 'sample' table found in the GeoTox connection.", call. = FALSE)
  }
  if (!"simulated_css" %in% tables) {
    stop("No 'simulated_css' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!"substance" %in% tables) {
    stop("No 'substance' table found in the GeoTox connection.", call. = FALSE)
  }
  if (!"concentration" %in% tables) {
    stop("No 'concentration' table found in the GeoTox connection.",
         call. = FALSE)
  }

  # Update parameters
  if (!is.null(substance_order)) {
    substance_order_col <- GT$par$substance_order_col <- names(substance_order)
    substance_order     <- GT$par$substance_order     <- substance_order[[1]]
    set_par(con, "substance_order_col", substance_order_col)
    set_par(con, "substance_order",     substance_order)
  } else if (!is.null(GT$par$substance_order_col)) {
    substance_order_col <- GT$par$substance_order_col
    substance_order     <- GT$par$substance_order
  }

  if (is.null(substance_order)) {
    substance_order <- dplyr::tbl(con, "substance") |>
      dplyr::pull("id")
  } else {
    substance_order <- dplyr::tbl(con, "substance") |>
      dplyr::select(tidyselect::all_of(c("id", substance_order_col))) |>
      dplyr::collect() |>
      dplyr::rename(value = 2) |>
      dplyr::arrange(match(.data$value, substance_order)) |>
      dplyr::pull("id")
  }

  sample_tbl <- dplyr::tbl(con, "sample")
  css_tbl    <- dplyr::tbl(con, "simulated_css")

  # Initialize "fixed_css" table
  fixed_css_tbl <- dplyr::cross_join(
    sample_tbl |>
      dplyr::select(sample_id = "id"),
    css_tbl |>
      dplyr::distinct(.data$substance_id)
  ) |>
    dplyr::arrange(.data$sample_id, .data$substance_id) |>
    dplyr::mutate(id = as.integer(dplyr::row_number()), .before = 1) |>
    dplyr::compute(
      name = "fixed_css",
      temporary = FALSE,
      overwrite = TRUE
    )
  # sql <- "ALTER TABLE fixed_css ADD PRIMARY KEY (id)"
  # DBI::dbExecute(con, sql)

  #=======================================
  # age
  #=======================================

  sql <- "ALTER TABLE fixed_css ADD COLUMN age DOUBLE"
  DBI::dbExecute(con, sql)

  # Compute median C_ss by substance and age group
  css_stats <- css_tbl |>
    dplyr::summarize(
      median_css = stats::median(.data$css, na.rm = TRUE),
      .by = c("substance_id", "age_lb", "age_ub")
    )

  # Join median C_ss by substance and age group to each sample
  df <- fixed_css_tbl |>
    # Get age from sample table
    dplyr::left_join(
      sample_tbl |> dplyr::select("id", "age"),
      by = dplyr::join_by("sample_id" == "id")
    ) |>
    # Add median_css value
    dplyr::left_join(
      css_stats,
      by = dplyr::join_by(
        "substance_id" == "substance_id",
        "age" >= "age_lb",
        "age" <= "age_ub"
      )
    ) |>
    # Keep only needed columns
    dplyr::select("id", "median_css") |>
    dplyr::rename(age = "median_css")

  # Add values to fixed_css table
  update_table(
    con, "fixed_css", df, copy = FALSE, reset_seed = GT$par$reset_seed
  )

  #=======================================
  # weight
  #=======================================

  sql <- "ALTER TABLE fixed_css ADD COLUMN weight DOUBLE"
  DBI::dbExecute(con, sql)

  # Compute median C_ss by substance and weight group
  css_stats <- css_tbl |>
    dplyr::summarize(
      median_css = stats::median(.data$css, na.rm = TRUE),
      .by = c("substance_id", "weight")
    )

  # Join median C_ss by substance and weight group to each sample
  df <- fixed_css_tbl |>
    # Get weight from sample table
    dplyr::left_join(
      sample_tbl |> dplyr::select("id", "weight"),
      by = dplyr::join_by("sample_id" == "id")
    ) |>
    # Add median_css value
    dplyr::left_join(
      css_stats,
      by = dplyr::join_by("substance_id", "weight")
    ) |>
    # Keep only needed columns
    dplyr::select("id", "median_css") |>
    dplyr::rename(weight = "median_css")

  update_table(
    con, "fixed_css", df, copy = FALSE, reset_seed = GT$par$reset_seed
  )

  #=======================================
  # params
  #=======================================

  sql <- "ALTER TABLE fixed_css ADD COLUMN params DOUBLE"
  DBI::dbExecute(con, sql)

  # Sample with replacement can't be done easily in database backends. Try to
  # fetch the minimum data needed for sampling in R.

  # First, get median age by location_id.
  median_age <- sample_tbl |>
    dplyr::summarize(
      value = stats::median(.data$age, na.rm = TRUE),
      .by = "location_id"
    )

  # Then, get all C_ss values for "Normal" weight and any age group that
  # includes the median age.
  css_df <- css_tbl |>
    dplyr::filter(.data$weight == "Normal") |>
    # median_age is non-integer and can lay between age_lb and age_ub (e.g. 40.5).
    # Add a new upper bound column that is +1 for the semi_join below
    dplyr::mutate(age_ub_p1 = .data$age_ub + 1) |>
    dplyr::semi_join(
      median_age,
      by = dplyr::join_by("age_lb" <= "value", "age_ub_p1" > "value")
    ) |>
    dplyr::select("substance_id", "age_lb", "age_ub", "css") |>
    dplyr::collect() |>
    tidyr::nest(css = "css")

  # Collect median_age after using in semi_join above
  median_age <- median_age |> dplyr::collect()

  # Sample C_ss values for each location_id and median age
  purrr::walk2(
    median_age$location_id,
    median_age$value,
    \(loc, age) {

      # Collect data for location
      df <- fixed_css_tbl |>
        # Get location_id from sample table
        dplyr::left_join(
          sample_tbl |> dplyr::select("id", "location_id"),
          by = dplyr::join_by("sample_id" == "id")
        ) |>
        dplyr::filter(.data$location_id == .env$loc) |>
        dplyr::select("id", "sample_id", "substance_id") |>
        dplyr::collect()

      df <- df |>
        # Arrange for replication
        dplyr::arrange(.data$sample_id) |>
        tidyr::nest(.by = "substance_id") |>
        dplyr::arrange(match(.data$substance_id, substance_order)) |>
        # Add C_ss values to sample from
        dplyr::left_join(
          css_df |>
            dplyr::filter(
              .data$age_lb <= .env$age,
              .data$age_ub + 1 > .env$age
            ) |>
            dplyr::select("substance_id", "css"),
          by = dplyr::join_by("substance_id")
        ) |>
        # Sample C_ss with replacement
        dplyr::mutate(
          sampled_css = purrr::map2(.data$data, .data$css, \(x, y) {
            # Note: if input vector, x, to sample() is a single numeric value,
            # then sample() will sample from 1:x. Here, we want n values of x,
            # not n samples from 1:x. See Details of ?sample for more info.
            css <- y$css
            n <- nrow(x)
            if (length(css) == 1) {
              params <- rep(css, n)
            } else {
              params <- sample(css, n, replace = TRUE)
            }
            dplyr::bind_cols(
              id = x$id,
              params = params
            )
          })
        ) |>
        dplyr::select("sampled_css") |>
        tidyr::unnest(cols = c("sampled_css"))

      update_table(con, "fixed_css", df, reset_seed = GT$par$reset_seed)
    }
  )

  #=======================================
  # other
  #=======================================

  sql <- "ALTER TABLE fixed_css ADD COLUMN other DOUBLE"
  DBI::dbExecute(con, sql)

  # Compute median C_ss from samples across all substances for each location
  css_stats <- dplyr::tbl(con, "concentration") |>
    dplyr::select("sample_id", "substance_id", "C_ss") |>
    # Get location_id from sample table
    dplyr::left_join(
      sample_tbl |> dplyr::select("id", "location_id"),
      by = dplyr::join_by("sample_id" == "id")
    ) |>
    # Mean-impute C_ss for each location + substance
    dplyr::mutate(
      mean_css = mean(.data$C_ss, na.rm = TRUE),
      imputed_css = dplyr::if_else(
        is.na(.data$C_ss), .data$mean_css, .data$C_ss
      ),
      .by = c("location_id", "substance_id")
    ) |>
    # Compute median
    dplyr::summarize(
      median_css = stats::median(.data$imputed_css, na.rm = TRUE),
      .by = "location_id"
    )

  # Join median C_ss by location to each sample
  df <- fixed_css_tbl |>
    # Get location_id from sample table
    dplyr::left_join(
      sample_tbl |> dplyr::select("id", "location_id"),
      by = dplyr::join_by("sample_id" == "id")
    ) |>
    # Add median_css value
    dplyr::left_join(
      css_stats,
      by = dplyr::join_by("location_id" == "location_id")
    ) |>
    # Keep only needed columns
    dplyr::select("id", "median_css") |>
    dplyr::rename(other = "median_css")

  update_table(
    con, "fixed_css", df, copy = FALSE, reset_seed = GT$par$reset_seed
  )

  invisible(GT)
}

#' Hill model concentration
#'
#' Calculate the concentration in regular space for a given response value.
#'
#' This is a regular space version of
#' [tcpl::tcplHillConc()](https://cran.r-project.org/package=tcpl).
#'
#' The concentration is computed as:
#' \deqn{conc = AC50 * (\frac{max}{resp} - 1)^{-1 / n}}
#'
#' @param resp Response value.
#' @param max Maximal (asymptotic) response.
#' @param AC50 Concentration of half-maximal response.
#' @param n Hill coefficient (slope).
#'
#' @returns Concentration in regular space.
#' @noRd
hill_conc <- function(resp, max, AC50, n) {

  if (any(AC50 < 0)) {
    stop("Input concentration can't be negative.", call. = FALSE)
  }
  if (any((resp >= 0) != (max >= 0))) {
    stop("'resp' and 'max' must have the same sign.", call. = FALSE)
  }

  AC50 * (max / resp - 1)^(-1 / n)

}

test_that("spot check", {
  max <- 100
  expect_equal(hill_conc( max /  2,  max, 10, 1),  10)
  expect_equal(hill_conc( max /  4,  max,  9, 1),   3)
  expect_equal(hill_conc(-max / 10, -max,  9, 1),   1)
  expect_equal(hill_conc( max /  5,  max, 10, 2),   5)
})

test_that("corner cases", {
  expect_equal(hill_conc(0:2, 2, 5, 1), c(0, 5, Inf))
})

test_that("negatives", {
  expect_error(hill_conc(-1,  1,  1, 1))
  expect_error(hill_conc( 1, -1,  1, 1))
  expect_error(hill_conc( 1,  1, -1, 1))
  expect_error(hill_conc(c(1, 1), c(-1, 1), 1, 1))
})

#' Simulate obesity values
#'
#' Simulate 'weight' category values to be stored in the 'sample' table of a
#' GeoTox database.
#'
#' An 'obesity' table containing simulation data must already exist in the
#' GeoTox database, which is added using [add_obesity()].
#'
#' The inputs `obes_prev` and `obes_sd` will be assigned default values of
#' "OBESITY_CrudePrev" and "OBESITY_SD", respectively, if not provided and not
#' already set in the GeoTox object's parameters, `GT$par`. If not `NULL`, the
#' provided values will also be saved to `GT$par`.
#'
#' @param GT GeoTox object.
#' @param n Number of individuals to simulate per location (default 1000).
#'   Ignored if 'sample' table already exists.
#' @param overwrite Logical indicating whether to overwrite existing 'weight'
#'   values in the 'sample' table (default FALSE).
#' @param obes_prev Column name of obesity prevalence in the 'obesity' table
#'   (default "OBESITY_CrudePrev").
#' @param obes_sd Column name of obesity standard deviation in the 'obesity'
#'   table (default "OBESITY_SD").
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [add_obesity()], [simulate_population()]
#'
#' @examples
#' # Example obesity simulation data
#' obesity_df <- data.frame(
#'   FIPS = c(10000, 20000),
#'   OBESITY_CrudePrev = c(20, 80),
#'   OBESITY_SD = 5
#' )
#'
#' # Simulate weight category values
#' GT <- GeoTox() |>
#'   add_obesity(obesity_df) |>
#'   simulate_obesity(n = 5)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' # Replace 'obesity' table with different column names
#' names(obesity_df)[2:3] <- c("prev", "sd")
#' DBI::dbRemoveTable(con, "obesity")
#' GT |> add_obesity(obesity_df)
#'
#' # Overwrite 'weight' category values in existing 'sample' table
#' # Must specify new 'obesity' column names
#' # Notice how the column names are added to GT$par
#' str(GT$par)
#' GT <- GT |>
#'   simulate_obesity(obes_prev = "prev", obes_sd = "sd", overwrite = TRUE)
#' str(GT$par)
#'
#' # Look at updated 'sample' table
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
simulate_obesity <- function(
    GT, n = 1e3, overwrite = FALSE, obes_prev = NULL, obes_sd = NULL
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (!DBI::dbExistsTable(con, "obesity")) {
    stop("No 'obesity' table found in the GeoTox connection.", call. = FALSE)
  }

  # Update parameters
  if (!is.null(obes_prev)) {
    set_par(con, "obes_prev", obes_prev)
    GT$par$obes_prev <- obes_prev
  } else if (!is.null(GT$par$obes_prev)) {
    obes_prev <- GT$par$obes_prev
  } else {
    obes_prev <- "OBESITY_CrudePrev"
  }
  if (!is.null(obes_sd)) {
    set_par(con, "obes_sd", obes_sd)
    GT$par$obes_sd <- obes_sd
  } else if (!is.null(GT$par$obes_sd)) {
    obes_sd <- GT$par$obes_sd
  } else {
    obes_sd <- "OBESITY_SD"
  }

  # Collect all the obesity simulation input data
  obesity_cols <- c("location_id", obes_prev, obes_sd)
  obesity_df <- dplyr::tbl(con, "obesity") |>
    dplyr::select(tidyselect::all_of(obesity_cols)) |>
    dplyr::collect() |>
    dplyr::arrange(.data$location_id) |>
    dplyr::rename(tidyselect::all_of(c("prev" = obes_prev, "sd" = obes_sd)))
  n_dup <- obesity_df |> dplyr::filter(duplicated(.data$location_id)) |> nrow()
  if (n_dup > 0) {
    stop("The 'obesity' table must not contain duplicated location values.",
         call. = FALSE)
  }

  # Simulate weight data for each location
  if (!DBI::dbExistsTable(con, "sample")) {
    # Create table
    # Initialize IDs so update can be used instead of append for each location.
    # Appending to grow the table is much slower for large datasets.
    # if (inherits(con, "SQLiteConnection")) {
    #   RSQLite::initExtension(con, extension = "series")
    #   sql <- paste0(
    #     "CREATE TABLE sample AS ",
    #     "SELECT value AS id ",
    #     "FROM generate_series(1, ", n * nrow(obesity_df), ", 1)"
    #   )
    # } else {
    sql <- paste0(
      "CREATE TABLE sample AS ",
      "SELECT CAST(id AS INTEGER) AS id ",
      "FROM range(1, ", n * nrow(obesity_df) + 1, ") AS t(id)"
    )
    # }
    DBI::dbExecute(con, sql)
    # DBI::dbExecute(con, "ALTER TABLE sample ADD PRIMARY KEY (id)")
    DBI::dbExecute(con, "ALTER TABLE sample ADD COLUMN location_id INTEGER")
    DBI::dbExecute(con, "ALTER TABLE sample ADD COLUMN weight TEXT")
    # Generate weights by location
    obesity_df |>
      dplyr::mutate(idx = dplyr::row_number()) |>
      purrr::pwalk(\(location_id, prev, sd, idx) {
        sample_df <- dplyr::bind_cols(
          id = ((idx - 1) * n + 1):(idx * n),
          location_id = location_id,
          weight = .simulate_obesity(n, prev, sd)
        )
        update_table(con, "sample", sample_df, reset_seed = GT$par$reset_seed)
      })
  } else {
    sample_tbl <- dplyr::tbl(con, "sample")
    if (!"weight" %in% colnames(sample_tbl)) {
      DBI::dbExecute(con, "ALTER TABLE sample ADD COLUMN weight TEXT")
    } else if (!overwrite) {
      stop("GeoTox connection already has a 'sample' table with a 'weight' ",
           "column. Set `overwrite = TRUE` to replace the existing values.\n",
           "Warning: Overwriting will not alter downstream values.",
           call. = FALSE)
    }
    # Generate weights by location
    obesity_df |>
      dplyr::select("location_id", "prev", "sd") |>
      purrr::pwalk(\(location_id, prev, sd) {
        sample_df <- sample_tbl |>
          dplyr::filter(.data$location_id == .env$location_id) |>
          dplyr::select("id") |>
          dplyr::arrange(.data$id) |>
          dplyr::collect()
        if (nrow(sample_df) == 0) return()
        weight <- .simulate_obesity(nrow(sample_df), prev, sd)
        sample_df <- dplyr::bind_cols(sample_df, weight = weight)
        update_table(con, "sample", sample_df, reset_seed = GT$par$reset_seed)
      })
  }
  invisible(GT)
}

.simulate_obesity <- function(n, prev, sd) {
  if (is.na(prev) | is.na(sd)) return(rep(NA_character_, n))
  p <- stats::rnorm(n, prev, sd) / 100
  p[p < 0] <- 0
  p[p > 1] <- 1
  weight <- stats::rbinom(n, 1, p)
  dplyr::if_else(weight == 1, "Obese", "Normal")
}

#' Generalized concentration addition objective function
#'
#' Use to find the optimal efficacy value based on a regular space AC50 and
#' concentrations.
#'
#' @param ln_resp Natural log of individual chemical responses.
#' @param conc Individual chemical concentrations in regular space.
#' @param max Maximal (asymtotic) responses.
#' @param AC50 Concentrations of half-maximal response.
#'
#' @returns Objective value.
#' @noRd
obj_GCA <- function(ln_resp, conc, max, AC50) {
  # Solving for the efficacy on the natural log-scale. This allows for
  # better precision in the low values, e.g. 1 x 10-5
  x <- hill_conc(exp(ln_resp), max, AC50, 1)
  gca.val <- sum(conc / x, na.rm = FALSE)
  (gca.val - 1)^2
}

test_that("spot check", {
  
  resp <- 1
  ln_resp <- log(resp)
  conc <- c(1, 3)
  max <- resp * 2
  AC50 <- conc
  
  expect_equal(obj_GCA(ln_resp, conc, max, AC50), 1)
  
  resp <- c(1, 2)
  ln_resp <- log(resp)
  conc <- c(1, 3)
  max <- c(3, 4)
  AC50 <- 1
  
  expect_equal(obj_GCA(ln_resp, conc, max, AC50), 16)
  
})

#' @param quantiles Numeric vector of quantiles to calculate.
#' @export
#' @rdname get_risk_values
get_risk_quantiles <- function(
    GT,
    metric = c("GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"),
    quantiles = c(0.1, 0.25, 0.5, 0.75, 0.9),
    table_name = "risk"
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  metric <- match.arg(metric)

  if (!DBI::dbExistsTable(con, table_name)) {
    stop("No '", table_name, "' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!DBI::dbExistsTable(con, "sample")) {
    stop("No 'sample' table found in the GeoTox connection.", call. = FALSE)
  }

  out <- dplyr::tbl(con, table_name) |>
    dplyr::select(tidyselect::all_of(c("assay_id", "sample_id", metric))) |>
    # Get location info from sample table
    dplyr::left_join(
      dplyr::tbl(con, "sample"),
      by = dplyr::join_by("sample_id" == "id")
    ) |>
    # Convert to duckplyr data frame so reframe works
    duckplyr::as_duckdb_tibble()

  # Get summary stats
  out |>
      dplyr::reframe(
      quantile = quantiles,
      value = stats::quantile(
        .data[[metric]],
        probs = .data$quantile,
        na.rm = TRUE,
        names = FALSE
      ),
      .by = c("assay_id", "location_id")
    ) |>
    dplyr::collect() |>
    # Fill in missing value with NA
    tidyr::complete(
      .data$assay_id,
      .data$location_id,
      .data$quantile,
      fill = list(value = NA_real_)
    )
}

#' Calculate risk sensitivity to a single variable
#'
#' Compute risk by varying one variable while holding others fixed.
#'
#' The sensitivity analysis makes use of the C\eqn{_{ss}} values stored in the
#' 'fixed_css' table of the GeoTox database. These values are determined using
#' the pre-simulated C\eqn{_{ss}} values supplied to [set_simulated_css()] and
#' can be set using [set_fixed_css()] prior to running this function. This step
#' is automatically done when using [simulate_population()] with `sample_css =
#' TRUE`.
#'
#' There are five options for the `vary` argument:
#' \describe{
#' \item{age}{C\eqn{_{ss}} values from the 'age' column of the 'fixed_css' table
#' are used. For other cases, exposure rates are re-simulated using the median
#' age by location in the 'sample' table by calling [simulate_exposure_rate()]
#' with `sensitivity = TRUE`.}
#' \item{weight}{C\eqn{_{ss}} values from the 'weight' column of the 'fixed_css'
#' table are used.}
#' \item{css_params}{C\eqn{_{ss}} values from the 'params' column of the
#' 'fixed_css' table are used.}
#' \item{fit_params}{C\eqn{_{ss}} values from the 'other' column of the
#' 'fixed_css' table are used. For other cases, the standard deviation of
#' dose-response model fit parameters are set to zero by calling [calc_risk()]
#' with `fixed = TRUE`.}
#' \item{C_ext}{C\eqn{_{ss}} values from the 'other' column of the 'fixed_css'
#' table are used. For other cases, external concentrations are re-simulated
#' with standard deviations set to zero by calling [simulate_exposure()] with
#' `sensitivity = TRUE`.}
#' }
#' In all cases above, the resulting risk table is named
#' 'risk_sensitivity_~vary~' (e.g., 'risk_sensitivity_age') in the GeoTox
#' database.
#'
#' Inputs `rate_extra_cols`, `expos_mean`, and `expos_sd` do not need to be
#' specified again if they were already provided in a previous call and are set
#' in the GeoTox parameters (`GT$par`).
#'
#' @param GT GeoTox object.
#' @param vary Variable to vary. One of "age", "weight", "css_params",
#'   "fit_params", or "C_ext".
#' @param overwrite Logical indicating whether to overwrite existing sensitivity
#'   analysis results in the GeoTox database.
#' @param rate_extra_cols Additional columns to match from the
#'   'exposure_rate_params' table (default NULL).
#' @param expos_mean Column name of exposure concentration mean in the
#'   'exposure' table (default "mean").
#' @param expos_sd Column name of exposure concentration standard deviation in
#'   the 'exposure' table (default "sd").
#' @param max_mult Upper bound multiplier for max response (default 1.5).
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [set_simulated_css()], [set_fixed_css()], [sensitivity_analysis()]
#'
#' @examples
#' # Example setup is shown below in \dontrun().
#' # Pre-generated results will be loaded instead to avoid long example runtime.
#' 
#' \dontrun{
#' # Setup required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal", 21,
#'   "00-00-1", 50, 99, "Normal", 22,
#'   "00-00-1",  0, 49,  "Obese", 61,
#'   "00-00-1", 50, 99,  "Obese", 62,
#'   "00-00-2",  0, 49, "Normal", 11,
#'   "00-00-2", 50, 99, "Normal", 12,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' hill_df <- tibble::tribble(
#'   ~assay, ~model, ~casn, ~logc, ~resp,
#'   "a1", "human", "00-00-1",    0,  10,
#'   "a1", "human", "00-00-1",    1,  20,
#'   "a1", "human", "00-00-1",    2,  80,
#'   "a1", "human", "00-00-1",    3, 100,
#'   "a1", "human", "00-00-2", -0.5,   5,
#'   "a1", "human", "00-00-2",  0.5,  20,
#'   "a1", "human", "00-00-2",  1.5,  55,
#'   "a1", "human", "00-00-2",  2.5,  60,
#'   "a2",   "rat", "00-00-1",   -1,   0,
#'   "a2",   "rat", "00-00-1",    0,  10,
#'   "a2",   "rat", "00-00-1",    1,  30,
#'   "a2",   "rat", "00-00-1",    2,  40
#' )
#' set.seed(1234)
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   set_simulated_css(css_df) |>
#'   add_exposure_rate_params() |>
#'   add_hill_params(fit_hill(
#'     hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
#'   )) |>
#'   simulate_population(exposure = exposure_df) |>
#'   calc_response()
#'
#' # Calculate sensitivity to age
#' GT <- GT |> calc_sensitivity("age")
#' 
#' }
#'
#' # Load results from pre-generated database for this example
#' temp_dir <- tempdir()
#' zip::unzip(
#'   system.file("extdata", "sensitivity.duckdb.zip", package = "GeoTox"),
#'   junkpaths = TRUE,
#'   exdir = temp_dir
#' )
#' GT <- GeoTox(paste0(temp_dir, "/sensitivity.duckdb"))
#' 
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant table
#'
#' dplyr::tbl(con, "risk_sensitivity_age") |> dplyr::collect()
#'
#' # Compared to baseline risk table
#' dplyr::tbl(con, "risk") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
calc_sensitivity <- function(
    GT,
    vary = c("age", "weight", "css_params", "fit_params", "C_ext"),
    overwrite = FALSE,
    rate_extra_cols = NULL,
    expos_mean = NULL,
    expos_sd = NULL,
    max_mult = 1.5
) {
  rate_name <- "exposure_rate_sensitivity"
  conc_name <- "concentration_sensitivity"
  con <- get_con(GT)
  on.exit({
    # Remove temporary tables and disconnect
    if (DBI::dbExistsTable(con, rate_name)) DBI::dbRemoveTable(con, rate_name)
    if (DBI::dbExistsTable(con, conc_name)) DBI::dbRemoveTable(con, conc_name)
    DBI::dbDisconnect(con)
  })

  vary <- match.arg(vary)

  tables <- DBI::dbListTables(con)
  if (!"exposure_rate" %in% tables) {
    stop("No 'exposure_rate' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!"concentration" %in% tables) {
    stop("No 'concentration' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!"fixed_css" %in% tables) {
    stop("No 'fixed_css' table found in the GeoTox connection.",
         call. = FALSE)
  }

  risk_name <- paste0("risk_sensitivity_", vary)
  if (DBI::dbExistsTable(con, risk_name)) {
    if (!overwrite) {
      stop("GeoTox connection already has a '", risk_name, "' table. ",
           "Use `overwrite = TRUE` to replace it.", call. = FALSE)
    }
    DBI::dbRemoveTable(con, risk_name)
  }

  #=====================================
  # Exposure rates
  #=====================================
  # Create temporary table exposure_rate_sensitivity
  rate_tbl <- dplyr::tbl(con, "exposure_rate") |>
    dplyr::compute(
      name = rate_name,
      overwrite = TRUE,
      # Set temporary = FALSE since temporary tables are session-specific.
      # Remove on exit of this function.
      temporary = FALSE
    )
  # Simulate exposure_rate using median age
  if (vary != "age") {
    simulate_exposure_rate(
      GT,
      rate_extra_cols = rate_extra_cols,
      sensitivity = TRUE
    )
  }

  #=====================================
  # External concentration (C_ext)
  #=====================================
  # Create temporary table concentration_sensitivity
  conc_tbl <- dplyr::tbl(con, "concentration") |>
    dplyr::select("id", "sample_id", "substance_id", "route_id", "C_ext") |>
    dplyr::compute(
      name = conc_name,
      overwrite = TRUE,
      # Set temporary = FALSE since temporary tables are session-specific.
      # Remove on exit of this function.
      temporary = FALSE
    )
  # Simulate C_ext values for sd = NA
  if (vary != "C_ext") {
    simulate_exposure(
      GT,
      expos_mean = expos_mean,
      expos_sd = expos_sd,
      sensitivity = TRUE
    )
  }

  #=====================================
  # Steady-state plasma concentration (C_ss)
  #=====================================
  sql <- paste("ALTER TABLE", conc_name, "ADD COLUMN C_ss DOUBLE")
  DBI::dbExecute(con, sql)
  fixed_css_col <- switch(
    vary,
    age        = "age",
    weight     = "weight",
    css_params = "params",
    "other"
  )
  select_cols <- c("sample_id", "substance_id", fixed_css_col)
  update_df <- dplyr::tbl(con, conc_name) |>
    dplyr::left_join(
      dplyr::tbl(con, "fixed_css") |>
        dplyr::select(tidyselect::all_of(select_cols)),
      by = c("sample_id", "substance_id")
    ) |>
    dplyr::select(tidyselect::all_of(c("id", fixed_css_col))) |>
    dplyr::rename(C_ss = 2)
  update_table(con, conc_name, update_df, reset_seed = GT$par$reset_seed)

  #=====================================
  # Internal dose (D_int)
  #=====================================
  calc_internal_dose(GT, sensitivity = TRUE)

  #=====================================
  # in vitro concentration (C_invitro)
  #=====================================
  calc_invitro_concentration(GT, sensitivity = TRUE)

  #=====================================
  # Risk
  #=====================================
  calc_risk(
    GT,
    max_mult = max_mult,
    fixed = vary != "fit_params",
    risk_name = risk_name
  )

  invisible(GT)
}

set_par <- function(con, name, values) {
  values <- as.character(values)
  sql <- paste0("DELETE FROM par WHERE name = ?")
  DBI::dbExecute(con, sql, params = list(name))
  name <- rep(name, length(values))
  sql <- "INSERT INTO par (name, value, idx) VALUES (?, ?, ?)"
  DBI::dbExecute(con, sql, params = list(name, values, seq_along(values)))
}

get_par <- function(con) {
  sql <- "SELECT name, value FROM par ORDER BY name, idx"
  df <- DBI::dbGetQuery(con, sql)
  par <- df |>
    dplyr::summarize(value = list(.data$value), .by = "name") |>
    tibble::deframe()
  par$reset_seed <- as.logical(par$reset_seed)
  par
}

check_reserved_columns <- function(df, reserved) {
  if (any(reserved %in% names(df))) {
    stop("`df` cannot contain column(s) '", paste(reserved, collapse = "', '"),
         "'.", call. = FALSE)
  }
  invisible(TRUE)
}

check_foreign_keys <- function(df, key_list) {
  purrr::iwalk(key_list, function(keys, key_col) {
    key_table <- sub("_id$", "", key_col)
    if (!is.character(keys) || length(keys) == 0) {
      stop("`", key_table, "` must be a character vector.", call. = FALSE)
    }
    if (!all(keys %in% names(df))) {
      stop("Each `", key_table, "` value must be a column name in `df`.",
           call. = FALSE)
    }
    if (!is.null(names(keys)) && any(names(keys) == "")) {
      stop("If `", key_table, "` is a named character vector, ",
           "then all values must be named.", call. = FALSE)
    }
  })
  invisible(TRUE)
}

#=======================================
# Begin RNG state preservation
#=======================================
# These functions will preserve the user's RNG state for the purpose of
# replication of previous GeoTox results. When performing the following tasks,
# temporary tables are created by the DBI implementation where 10 random
# characters are appended to the table name. To prevent this from affecting the
# user's RNG state, the .Random.seed is reset and any new temporary tables are
# removed in case the functions are called multiple times in a row.

write_table <- function(
    con, table, df, overwrite = FALSE, field.types = NULL, reset_seed = FALSE
) {
  if (reset_seed) seed <- .Random.seed
  DBI::dbWriteTable(
    con, table, df, overwrite = overwrite, field.types = field.types
  )
  if (reset_seed) .Random.seed <<- seed
}

append_table <- function(con, table, df, reset_seed = FALSE) {
  if (reset_seed) seed <- .Random.seed
  DBI::dbAppendTable(con, table, df)
  if (reset_seed) .Random.seed <<- seed
}

update_table <- function(
    con, table, df, by = "id", copy = TRUE, reset_seed = FALSE
) {
  if (reset_seed) {
    seed <- .Random.seed
    tbls <- DBI::dbListTables(con)
  }
  dplyr::rows_update(
    dplyr::tbl(con, table),
    df,
    by = by,
    in_place = TRUE,
    unmatched = "ignore",
    copy = copy
  )
  if (reset_seed) {
    rm_tbl <- setdiff(DBI::dbListTables(con), tbls)
    if (length(rm_tbl) == 1) {
      DBI::dbRemoveTable(con, rm_tbl)
    }
    .Random.seed <<- seed
  }
}
#=======================================
# End RNG state preservation
#=======================================

fetch_foreign_keys <- function(con, df, foreign_keys, reset_seed) {
  purrr::iwalk(foreign_keys, function(keys, key_col) {
    key_table <- sub("_id$", "", key_col)
    # Add any new data to foreign key table. Even if the table already exists
    # and no new data is being added, calling add_table is simple and we don't
    # care about efficiency in these cases.
    key_df <- df |> dplyr::select(tidyselect::all_of(keys)) |> dplyr::distinct()
    # Use key_names to map the foreign key columns back to the main table
    if (!is.null(names(keys))) {
      key_df <- stats::setNames(key_df, names(keys))
      key_names <- stats::setNames(names(keys), keys)
    } else {
      key_names <- stats::setNames(keys, keys)
    }
    # The foreign key tables will not have foreign keys themselves, so we can
    # recursively call add_table with foreign_keys = NULL.
    add_table(con, key_table, key_df, reset_seed = reset_seed)
    # Get foreign key IDs
    key_df <- DBI::dbReadTable(con, key_table) |>
      dplyr::select(tidyselect::all_of(c("id", unname(key_names))))
    # Add foreign key IDs to the main table
    df <<- df |>
      dplyr::left_join(key_df, by = key_names) |>
      dplyr::select(-tidyselect::all_of(names(key_names))) |>
      dplyr::rename(tidyselect::all_of(stats::setNames("id", key_col)))
  })
  df
}

overwrite_table <- function(con, table, df, foreign_keys, reset_seed = FALSE) {
  key_cols <- c("id", names(foreign_keys))
  df <- df |>
    # Add primary key
    dplyr::mutate(id = dplyr::row_number()) |>
    # Reorder columns
    dplyr::select(tidyselect::all_of(c(key_cols, setdiff(names(df), key_cols))))
  # Write the table and add key constraints
  write_table(con, table, df, overwrite = TRUE, reset_seed = reset_seed)
  # sql <- paste("ALTER TABLE", table, "ADD PRIMARY KEY (id)")
  # DBI::dbExecute(con, sql)
  # purrr::walk(names(foreign_keys), function(key_col) {
  #   # Adding foreign keys via ALTER TABLE is not implemented yet in duckdb,
  #   # instead, just ensure they aren't NULL.
  #   sql <- paste("ALTER TABLE", table, "ALTER", key_col, "SET NOT NULL")
  #   DBI::dbExecute(con, sql)
  # })
}

add_table <- function(
    con, table, df, foreign_keys = NULL, distinct = TRUE, overwrite = FALSE,
    reset_seed = FALSE
) {
  reserved <- c("id", names(foreign_keys))
  check_reserved_columns(df, reserved)

  # Add foreign keys
  if (!is.null(foreign_keys)) {
    check_foreign_keys(df, foreign_keys)
    df <- fetch_foreign_keys(con, df, foreign_keys, reset_seed)
  }

  if (distinct) {
    df <- df |> dplyr::distinct()
  }

  if (!DBI::dbExistsTable(con, table) | overwrite) {
    overwrite_table(con, table, df, foreign_keys, reset_seed = reset_seed)
  } else {
    # Load current table
    tbl <- DBI::dbReadTable(con, table)
    # Add new columns to the existing table
    cols <- setdiff(names(df), names(tbl))
    if (length(cols) > 0) {
      for (col in cols) {
        if (is.integer(df[[col]])) {
          type <- "INTEGER"
        } else if (is.numeric(df[[col]])) {
          type <- "DOUBLE"
        } else {
          type <- "TEXT"
        }
        sql <- paste("ALTER TABLE", table, "ADD COLUMN", col, type)
        DBI::dbExecute(con, sql)
      }
      # refresh table
      tbl <- DBI::dbReadTable(con, table)
    }
    # Add new data
    # Exclude new "cols" from the join. Give them a suffix of ".new" so they
    # can be removed easily after the join.
    if (length(cols) > 0) {
      by_cols <- intersect(names(df), setdiff(names(tbl), cols))
      if (length(by_cols) > 0) {
        df <- dplyr::full_join(
          df,
          tbl,
          by = by_cols,
          suffix = c("", ".new")
        ) |>
          dplyr::select(-tidyselect::all_of(paste0(cols, ".new")))
      } else {
        # If no overlapping columns, assume all are new rows
        df <- dplyr::full_join(df, tbl, by = intersect(names(df), names(tbl)))
      }
    } else {
      df <- dplyr::full_join(df, tbl, by = intersect(names(df), names(tbl)))
    }
    # Add new column data to existing rows
    if (length(cols) > 0) {
      params <- df |>
        dplyr::filter(!is.na(.data$id)) |>
        dplyr::select(tidyselect::all_of(c("id", cols)))
      if (nrow(params) > 0) {
        update_table(con, table, params, reset_seed = reset_seed)
      }
    }
    # New rows
    params <- df |> dplyr::filter(is.na(.data$id))
    if (nrow(params) > 0) {
      # Assign new IDs
      if (nrow(tbl) == 0) {
        max_id <- 0
      } else {
        max_id <- max(tbl$id)
      }
      params <- params |> dplyr::mutate(id = dplyr::row_number() + max_id)
      # Append new rows to DB
      append_table(con, table, params, reset_seed = reset_seed)
    }
  }
}

test_that("table helpers", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Check par table
  invisible(set_par(con, "other_params", c("a", "b", "c")))
  expect_equal(
    get_par(con),
    list(
      # ordered by name, idx
      other_params = c("a", "b", "c"),
      reset_seed = FALSE
    )
  )

  # Start with a location table, use reset_seed
  location_df <- data.frame(FIPS = c("00001", "00002"))
  seed <- .Random.seed
  add_table(con, "location", location_df, reset_seed = TRUE)
  expect_equal(
    dplyr::tbl(con, "location") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~FIPS,
      1, "00001",
      2, "00002"
    )
  )
  expect_identical(.Random.seed, seed)

  # Add new rows/columns to location table, use reset_seed
  location_df <- data.frame(
    FIPS = c("00001", "00003"),
    area = c(10, 20),
    zip = c(12345L, 54321L)
  )
  seed <- .Random.seed
  add_table(con, "location", location_df, reset_seed = TRUE)
  expect_equal(
    dplyr::tbl(con, "location") |> dplyr::collect(),
    tibble::tribble(
      ~id,   ~FIPS, ~area,   ~zip,
        1, "00001",    10, 12345L,
        2, "00002",    NA,     NA,
        3, "00003",    20, 54321L
    )
  )
  expect_identical(.Random.seed, seed)

  # Reserved column
  test_df <- data.frame(id = 1, key_id = 1)
  foreign_keys <- list(key_id = "key_id")
  expect_error(
    add_table(con, "test", test_df, foreign_keys = foreign_keys),
    "`df` cannot contain column\\(s\\) 'id', 'key_id'."
  )

  # Bad foreign keys
  test_df <- data.frame(loc = c("00001"), new_col = 50)
  foreign_keys <- list(location_id = c())
  expect_error(
    add_table(con, "test", test_df, foreign_keys),
    "`location` must be a character vector."
  )
  foreign_keys <- list(location_id = c(FIPS = "loc", "other"))
  expect_error(
    add_table(con, "test", test_df, foreign_keys),
    "Each `location` value must be a column name in `df`."
  )
  foreign_keys <- list(location_id = c(FIPS = "loc", "loc"))
  expect_error(
    add_table(con, "test", test_df, foreign_keys),
    "If `location` is a named character vector, then all values must be named."
  )

  # Add sample table with named foreign keys, don't use reset_seed
  sample_df <- data.frame(loc = c("00001"), age = 50)
  foreign_keys <- list(location_id = c(FIPS = "loc"))
  seed <- .Random.seed
  add_table(con, "sample", sample_df, foreign_keys)
  expect_equal(
    dplyr::tbl(con, "sample") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~location_id, ~age,
      1, 1, 50
    )
  )
  expect_false(identical(.Random.seed, seed))

  # Add to empty table with unnamed foreign key
  empty_df <- data.frame(id = integer(0), location_id = integer(0))
  DBI::dbWriteTable(con, "empty", empty_df)
  empty_df <- data.frame(FIPS = c("00002"), new_col = "text")
  foreign_keys <- list(location_id = "FIPS")
  add_table(con, "empty", empty_df, foreign_keys)
  expect_equal(
    dplyr::tbl(con, "empty") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~location_id, ~new_col,
      1, 2, "text"
    )
  )

  # Add to table with no overlapping columns
  location_df <- data.frame(ST = c("XYZ"))
  add_table(con, "location", location_df)
  expect_equal(
    dplyr::tbl(con, "location") |> dplyr::collect(),
    tibble::tribble(
      ~id,   ~FIPS, ~area,   ~zip,   ~ST,
        1, "00001",    10, 12345L,    NA,
        2, "00002",    NA,     NA,    NA,
        3, "00003",    20, 54321L,    NA,
        4,      NA,    NA,     NA, "XYZ"
    )
  )

})

#' @export
#' @rdname get_risk_values
get_risk_sensitivity <- function(
    GT, metric = c("GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"), assay = NULL
) {
  metric <- match.arg(metric)
  tibble::tibble(
    "C_ext" = get_risk_values(
      GT, metric, assay, "risk_sensitivity_C_ext"
    ),
    "css_params" = get_risk_values(
      GT, metric, assay, "risk_sensitivity_css_params"
    ),
    "weight" = get_risk_values(
      GT, metric, assay, "risk_sensitivity_weight"
    ),
    "age" = get_risk_values(
      GT, metric, assay, "risk_sensitivity_age"
    ),
    "fit_params" = get_risk_values(
      GT, metric, assay, "risk_sensitivity_fit_params"
    ),
    "baseline" = get_risk_values(
      GT, metric, assay, "risk"
    )
  )
}

#' Simulate exposure concentrations
#'
#' Simulate external exposure (C_ext) values to be stored in the
#' 'concentration' table of a GeoTox database.
#'
#' An 'external' table containing simulation data must already exist in the
#' GeoTox database, which is added using [add_exposure()].
#'
#' The inputs `expos_mean` and `expos_sd` will be assigned default values of
#' "mean" and "sd", respectively, if not provided and not already set in the
#' GeoTox object's parameters, `GT$par`. If not `NULL`, the provided values will
#' also be saved to `GT$par`.
#'
#' If `sensitivity = TRUE`, exposure concentrations will be simulated for
#' sensitivity analysis. Typically this shouldn't be used directly by the user,
#' but rather called by [calc_sensitivity()]. In this case, the function will
#' use the 'concentration_sensitivity' table instead of the 'concentration'
#' table, and will assume that the 'sample' table already exists.
#'
#' @param GT GeoTox object.
#' @param n Number of individuals to simulate per location (default 1000).
#'   Ignored if 'sample' table already exists.
#' @param overwrite Logical indicating whether to overwrite existing 'C_ext'
#'   values in the 'concentration' table (default FALSE).
#' @param expos_mean Column name of exposure concentration mean in the
#'   'exposure' table (default "mean").
#' @param expos_sd Column name of exposure concentration standard deviation in
#'   the 'exposure' table (default "sd").
#' @param sensitivity Logical indicating whether to simulate exposures for
#'   sensitivity analysis (default FALSE).
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [add_exposure()], [simulate_population()]
#'
#' @examples
#' # Example exposure simulation data
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#'
#' # Simulate C_ext values
#' GT <- GeoTox() |>
#'   add_exposure(exposure_df) |>
#'   simulate_exposure(n = 3)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#'
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Replace 'exposure' table with different column names
#' names(exposure_df)[4:5] <- c("mu", "sigma")
#' DBI::dbRemoveTable(con, "exposure")
#' GT |> add_exposure(exposure_df)
#'
#' # Overwrite 'C_ext' values in existing 'concentration' table
#' # Must specify new 'exposure' column names
#' # Notice how the column names are added to GT$par
#' str(GT$par)
#' GT <- GT |>
#'   simulate_exposure(expos_mean = "mu", expos_sd = "sigma", overwrite = TRUE)
#' str(GT$par)
#'
#' # Look at updated 'concentration' table
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
simulate_exposure <- function(
    GT, n = 1e3, overwrite = FALSE, expos_mean = NULL, expos_sd = NULL,
    sensitivity = FALSE
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (!DBI::dbExistsTable(con, "exposure")) {
    stop("No 'exposure' table found in the GeoTox connection.", call. = FALSE)
  }
  if (sensitivity) {
    conc_name <- "concentration_sensitivity"
    # Conc table should have been created by calc_sensitivity()
    if (!DBI::dbExistsTable(con, conc_name)) {
      stop("No '", conc_name, "' table found in the GeoTox connection.",
           call. = FALSE)
    }
    # Sample table should already exist
    if (!DBI::dbExistsTable(con, "sample")) {
      stop("No 'sample' table found in the GeoTox connection.", call. = FALSE)
    }
    conc_tbl <- dplyr::tbl(con, conc_name)
  } else {
    conc_name <- "concentration"
  }

  # Update parameters
  if (!is.null(expos_mean)) {
    set_par(con, "expos_mean", expos_mean)
    GT$par$expos_mean <- expos_mean
  } else if (!is.null(GT$par$expos_mean)) {
    expos_mean <- GT$par$expos_mean
  } else {
    expos_mean <- "mean"
  }
  if (!is.null(expos_sd)) {
    set_par(con, "expos_sd", expos_sd)
    GT$par$expos_sd <- expos_sd
  } else if (!is.null(GT$par$expos_sd)) {
    expos_sd <- GT$par$expos_sd
  } else {
    expos_sd <- "sd"
  }

  # Collect all the exposure simulation input data
  exposure_cols <- c(
    "id", "location_id", "substance_id", "route_id", expos_mean, expos_sd
  )
  exposure_df <- dplyr::tbl(con, "exposure") |>
    dplyr::select(tidyselect::all_of(exposure_cols)) |>
    dplyr::collect()
  if (sensitivity) {
    exposure_df <- exposure_df |> dplyr::mutate(sd = NA_real_)
  }
  exposure_df <- exposure_df |>
    tidyr::nest(.by = "location_id") |>
    dplyr::arrange(.data$location_id)

  if (!sensitivity) {
    if (!DBI::dbExistsTable(con, "sample")) {
      # Create sample table
      sql <- paste0(
        "CREATE TABLE sample AS ",
        "SELECT CAST(id AS INTEGER) AS id ",
        "FROM range(1, ", n * nrow(exposure_df) + 1, ") AS t(id)"
      )
      DBI::dbExecute(con, sql)
      DBI::dbExecute(con, "ALTER TABLE sample ADD COLUMN location_id INTEGER")
      # Add location info
      exposure_df |>
        dplyr::mutate(idx = dplyr::row_number()) |>
        dplyr::select("location_id", "idx") |>
        purrr::pwalk(\(location_id, idx) {
          sample_df <- dplyr::bind_cols(
            id = ((idx - 1) * n + 1):(idx * n),
            location_id = location_id
          )
          update_table(con, "sample", sample_df, reset_seed = GT$par$reset_seed)
        })
    }
    if (!DBI::dbExistsTable(con, conc_name)) {
      # Create concentration table
      conc_tbl <- dplyr::cross_join(
        dplyr::tbl(con, "sample") |>
          dplyr::select("id") |>
          dplyr::rename(sample_id = "id"),
        dplyr::tbl(con, "exposure") |>
          dplyr::distinct(.data$substance_id, .data$route_id)
      ) |>
        dplyr::arrange(.data$sample_id, .data$substance_id, .data$route_id) |>
        dplyr::mutate(id = dplyr::row_number(), .before = 1) |>
        dplyr::compute(
          name = conc_name,
          temporary = FALSE
        )
      sql <- paste("ALTER TABLE", conc_name, "ADD COLUMN C_ext DOUBLE")
      DBI::dbExecute(con, sql)
    } else {
      conc_tbl <- dplyr::tbl(con, conc_name)
      if (!"C_ext" %in% colnames(conc_tbl)) {
        sql <- paste("ALTER TABLE", conc_name, "ADD COLUMN C_ext DOUBLE")
        DBI::dbExecute(con, sql)
      } else if (!overwrite) {
        stop("GeoTox connection already has a '", conc_name, "' table with a ",
             "'C_ext' column. Set `overwrite = TRUE` to replace the existing ",
             "values.\n",
             "Warning: Overwriting will not alter downstream values.",
             call. = FALSE)
      }
    }
  }

  # Loop over locations
  exposure_df |>
    purrr::pwalk(\(location_id, data) {
      sample_df <- dplyr::tbl(con, "sample") |>
        dplyr::filter(.data$location_id == .env$location_id) |>
        dplyr::select("id") |>
        dplyr::arrange(.data$id) |>
        dplyr::collect()
      if (nrow(sample_df) == 0) return()
      data <- data |> dplyr::arrange(.data$substance_id)
      update_df <- .simulate_exposure(
        data,
        expos_mean,
        expos_sd,
        n = nrow(sample_df)
      )
      colnames(update_df) <- data$id
      update_df <- update_df |>
        dplyr::as_tibble() |>
        dplyr::mutate(sample_id = sample_df$id) |>
        tidyr::pivot_longer(
          cols = !c("sample_id"),
          names_to = "exposure_id",
          values_to = "C_ext"
        ) |>
        dplyr::mutate(exposure_id = as.integer(.data$exposure_id)) |>
        dplyr::left_join(
          data |> dplyr::select("id", "substance_id", "route_id"),
          by = dplyr::join_by("exposure_id" == "id")
        ) |>
        dplyr::select("sample_id", "substance_id", "route_id", "C_ext")
      update_table(
        con, conc_name, update_df, reset_seed = GT$par$reset_seed,
        by = c("sample_id", "substance_id", "route_id")
      )
    })
  invisible(GT)
}

.simulate_exposure <- function(x, mean, sd, n) {

  mean <- x[[mean]]
  sd <- x[[sd]]

  mapply(
    function(mean, sd) {
      if (mean == 0) {
        rep(0, n)
      } else if (mean > 0 & is.na(sd)) {
        rep(mean, n)
      } else {
        truncnorm::rtruncnorm(
          n, a = 0, b = Inf, mean = mean, sd = sd
        )
      }
    },
    mean = mean,
    sd = sd,
    SIMPLIFY = FALSE
  ) |>
    do.call(what = cbind)
}

#' Set pre-simulated steady-state plasma concentrations
#'
#' Create the 'simulated_css' table in a GeoTox database, which contains
#' pre-simulated steady-state plasma concentrations (C\eqn{_{ss}}).
#'
#' The 'simulated_css' table is used by [sample_simulated_css()] to assign
#' C\eqn{_{ss}} values to individuals. The minimum required columns in the `df`
#' data frame are "age_lb", "age_ub", "weight", "css", and at least one column
#' with substance information (default "casn").
#'
#' The values for "age_lb" and "age_ub" should be non-overlapping integers
#' representing age ranges (in years). For example, two subsequent age groups
#' might be `c(0, 4)` and `c(5, 9)`. The "weight" column should contain the
#' weight category and contain values of either "Normal" or "Obese". The "css"
#' column should contain the pre-simulated C\eqn{_{ss}} values.
#'
#' The `substance` input can be a named vector to specify multiple substance
#' identifier columns in `df`. For example, `c(casn = "casn", name = "chnm")`
#' would indicate that `df` contains both CAS numbers and chemical names for
#' substances. The `name = "chnm"` part would rename the "chnm" column in `df`
#' to "name" in the 'substance' table.
#'
#' @param GT GeoTox object.
#' @param df Data frame containing simulated C\eqn{_{ss}} values for groups of
#'   population characteristics.
#' @param substance Column name(s) in `df` that contain substance identifier(s)
#'   (default "casn").
#' @param overwrite Logical indicating whether to overwrite existing
#'   'simulated_css' table (default FALSE).
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [sample_simulated_css()]
#'
#' @examples
#' # Example pre-simulated C_ss data
#' # Note: normally the css_df would have many more rows for each combination of
#' # the non-'css' columns to allow for sampling.
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal",  1,
#'   "00-00-1", 50, 99, "Normal",  2,
#'   "00-00-1",  0, 49,  "Obese", 11,
#'   "00-00-1", 50, 99,  "Obese", 12,
#'   "00-00-2",  0, 49, "Normal", 21,
#'   "00-00-2", 50, 99, "Normal", 22,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#'
#' # Set simulated C_ss values
#' GT <- GeoTox() |> set_simulated_css(css_df)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant tables
#'
#' dplyr::tbl(con, "simulated_css") |> dplyr::collect()
#'
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
set_simulated_css <- function(GT, df, substance = "casn", overwrite = FALSE) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  foreign_keys <- list("substance_id" = substance)
  check_foreign_keys(df, foreign_keys)
  cols <- c("age_lb", "age_ub", "weight", "css")
  if (!all(cols %in% names(df))) {
    stop("`df` must contain the following columns: ",
         paste(cols, collapse = ", "), ".", call. = FALSE)
  }
  if (DBI::dbExistsTable(con, "simulated_css") && !overwrite) {
    stop("Table 'simulated_css' already exists. ",
         "Set `overwrite = TRUE` to replace it.", call. = FALSE)
  }
  df <- fetch_foreign_keys(con, df, foreign_keys, GT$par$reset_seed)
  overwrite_table(
    con, "simulated_css", df, foreign_keys, reset_seed = GT$par$reset_seed
  )
  invisible(GT)
}

#' Independent action
#'
#' Calculate independent action response for a set of chemicals with Hill
#' concentration-response curves.
#'
#' The concentration is computed as:
#' \deqn{
#'   IA = E_{max} \times
#'   \left(
#'     1 - \prod\limits_{i} \left(1 - \frac{x_i}{E_{max}}\right)
#'   \right),
#' }
#' where \eqn{x_i = hill\_val(conc_i, max_i, AC50_i, n_i)} is the Hill model
#' response function for each chemical.
#'
#' @param conc Concentrations in regular space.
#' @param max Maximal (asymptotic) responses.
#' @param AC50 Concentrations of half-maximal response.
#' @param Emax Maximum mixture response.
#' @param n Hill coefficients (slopes).
#'
#' @returns Response value.
#' @noRd
calc_independent_action <- function(conc, max, AC50, Emax, n = 1) {
  resp <- hill_val(conc, max, AC50, n)
  Emax * (1 - prod((1 - resp / Emax)))
}

test_that("random spot check", {
  Emax <- 100
  conc <-  10 * runif(10) + 1e-4
  max  <- Emax * runif(10) + 1e-4
  AC50 <-  10 * runif(10) + 1e-4
  n    <- 1 + rpois(10, 5)
  IA   <- calc_independent_action(conc, max, AC50, Emax, n)

  expect_true(IA > 0 & IA < Emax)
})

#' @keywords internal
"_PACKAGE"

# Dummy function to avoid R CMD check notes about unused imports from dbplyr.
ignore_unused_imports <- function() {
  dbplyr::tbl_lazy
}

## usethis namespace: start
#' @importFrom rlang .data
#' @importFrom rlang .env
## usethis namespace: end
NULL

#' Simulate population characteristics and exposures
#'
#' Simulate age, weight category, exposure rates, and external exposures. Sample
#' from pre-simulated steady-state plasma concentrations.
#'
#' This is a wrapper around several other functions:
#' * [add_age()] and [simulate_age()] for age simulation.
#' * [add_obesity()] and [simulate_obesity()] for weight category simulation.
#' * [add_exposure()] and [simulate_exposure()] for external exposure
#' concentration simulation (C\eqn{_{ext}}).
#' * [simulate_exposure_rate()] for exposure rate simulation.
#' * [sample_simulated_css()] for sampling steady-state plasma concentrations
#' (C\eqn{_{ss}}).
#' * [set_fixed_css()] to prepare C\eqn{_{ss}} values for sensitivity analysis.
#'
#' The user can provide data frames for age, obesity, and exposure data; if any
#' of these are provided, the corresponding add and simulate functions will be
#' called. The user can also specify whether to simulate exposure rates and
#' sample C\eqn{_{ss}} values using the `simulate_rate` and `sample_css`
#' arguments, respectively. If `simulate_rate` is `TRUE`, exposure rate
#' parameters must have been added using [add_exposure_rate_params()]. If
#' `sample_css` is `TRUE`, a table of simulated C\eqn{_{ss}} values must already
#' exist using [set_simulated_css()].
#'
#' ## Additional arguments:
#' \describe{
#' \item{n}{Number of samples to simulate (default 1000). Used in
#' [simulate_age()], [simulate_obesity()], and [simulate_exposure()]. Ignored
#' if the 'sample' table already exists, in which case the existing sample sizes
#' are used.}
#' \item{location}{Column name for location ID (default "FIPS"). Used in
#' [add_age()], [add_obesity()], and [add_exposure()].}
#' \item{overwrite}{Logical indicating whether to overwrite existing values
#' (default FALSE). Used in [simulate_age()], [simulate_obesity()],
#' [simulate_exposure_rate()], and [simulate_exposure()].}
#' \item{substance}{Column name for substance ID (default "casn"). Used in
#' [add_exposure()].}
#' \item{route}{Column name for exposure route (default "route"). Used in
#' [add_exposure()].}
#' \item{rate_extra_cols}{Additional columns to include in exposure_rate table.
#' Used in [simulate_exposure_rate()].}
#' \item{obes_prev, obes_sd}{Column names for obesity prevalence and standard
#' deviation (default "OBESITY_CrudePrev" and "OBESITY_SD", respectively). Used
#' in [simulate_obesity()].}
#' \item{expos_mean, expos_sd}{Column names for exposure concentration mean and
#' standard deviation (default "mean" and "sd", respectively). Used in
#' [simulate_exposure()].}
#' \item{css_extra_cols}{Additional columns to include in simulated_css table.
#' Used in [sample_simulated_css()].}
#' \item{substance_order}{Named list specifying order of substances. Used in
#' [sample_simulated_css()] and [set_fixed_css()].}
#' }
#'
#' @param GT GeoTox object.
#' @param age Data frame with age data.
#' @param obesity Data frame with obesity data.
#' @param exposure Data frame with exposure data.
#' @param simulate_rate Logical indicating whether to simulate exposure rates.
#'   This requires that exposure rate parameters have been added using
#'   [add_exposure_rate_params()].
#' @param sample_css Logical indicating whether to sample steady-state plasma
#'   concentrations (C\eqn{_{ss}}). This requires that a table of simulated
#'   C\eqn{_{ss}} values has been set using [set_simulated_css()]. In addition,
#'   [set_fixed_css()] will be called after sampling to prepare C\eqn{_{ss}}
#'   values for sensitivity analysis.
#' @param ... Additional arguments passed to wrapped functions (see 'Additional
#'   arguments' section of 'Details').
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [add_age()], [simulate_age()], [add_obesity()],
#'   [simulate_obesity()], [add_exposure()], [simulate_exposure()],
#'   [simulate_exposure_rate()], [sample_simulated_css()], [set_fixed_css()]
#'
#' @examples
#' # Example simulation data
#'
#' age_df <- data.frame(
#'   FIPS = rep(c(10000, 20000), each = 19),
#'   AGEGRP = rep(0:18, times = 2),
#'   TOT_POP = 0
#' )
#' # FIPS 10000, populate age group 40-44
#' age_df$TOT_POP[c(1, 10)] = 100
#' # FIPS 20000, populate age groups 50-59
#' age_df$TOT_POP[c(1, 12, 13) + 19] = c(200, 100, 100)
#'
#' obesity_df <- data.frame(
#'   FIPS = c(10000, 20000),
#'   OBESITY_CrudePrev = c(20, 80),
#'   OBESITY_SD = 5
#' )
#'
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#'
#' # Note: normally the css_df would have many more rows for each combination of
#' # the non-'css' columns to allow for sampling.
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal",  1,
#'   "00-00-1", 50, 99, "Normal",  2,
#'   "00-00-1",  0, 49,  "Obese", 11,
#'   "00-00-1", 50, 99,  "Obese", 12,
#'   "00-00-2",  0, 49, "Normal", 21,
#'   "00-00-2", 50, 99, "Normal", 22,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#'
#' # Simulate population
#' GT <- GeoTox() |>
#'   add_exposure_rate_params() |>
#'   set_simulated_css(css_df) |>
#'   simulate_population(
#'     age = age_df,
#'     obesity = obesity_df,
#'     exposure = exposure_df,
#'     n = 3
#'   )
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#'
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Note: the 'age', 'weight', 'params', and 'other' columns of the
#' # 'fixed_css' table contain the C_ss values for sensitivity analysis.
#' # For example, the 'age' column doesn't contain ages, but C_ss values.
#' dplyr::tbl(con, "fixed_css") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
simulate_population <- function(
    GT, age = NULL, obesity = NULL, exposure = NULL,
    simulate_rate = TRUE, sample_css = TRUE, ...
) {
  dots <- list(...)

  n         <- dots$n         %||% 1e3
  location  <- dots$location  %||% "FIPS"
  overwrite <- dots$overwrite %||% FALSE

  # Prevent partial matching issue
  if ("substance" %in% names(dots)) {
    # Still allow for dots$substance = NULL
    substance <- dots$substance %||% "casn"
  } else {
    substance <- "casn"
  }

  # Age
  if (!is.null(age)) {
    GT <- GT |>
      add_age(age, location = location) |>
      simulate_age(n = n, overwrite = overwrite)
  }

  # Exposure rate
  con <- get_con(GT)
  age_col_exists <- DBI::dbExistsTable(con, "sample") &&
    "age" %in% colnames(dplyr::tbl(con, "sample"))
  rate_tbl_exists <- DBI::dbExistsTable(con, "exposure_rate")
  DBI::dbDisconnect(con)
  if (simulate_rate && (age_col_exists & (!rate_tbl_exists | overwrite))) {
    GT <- GT |>
      simulate_exposure_rate(
        rate_extra_cols = dots$rate_extra_cols,
        overwrite       = overwrite
      )
  }

  # Weight
  if (!is.null(obesity)) {
    GT <- GT |>
      add_obesity(obesity, location = location) |>
      simulate_obesity(
        n         = n,
        overwrite = overwrite,
        obes_prev = dots$obes_prev,
        obes_sd   = dots$obes_sd
      )
  }

  # Exposure
  if (!is.null(exposure)) {
    GT <- GT |>
      add_exposure(
        exposure,
        location  = location,
        # This can result in a partial match error if "substance" is not in dots
        # but "substance_order" is.
        # substance = dots$substance %||% "casn",
        substance = substance,
        route     = dots$route %||% "route"
      ) |>
      simulate_exposure(
        n          = n,
        overwrite  = overwrite,
        expos_mean = dots$expos_mean,
        expos_sd   = dots$expos_sd
      )
  }

  # C_ss
  if (sample_css) {
    GT <- GT |>
      sample_simulated_css(
        css_extra_cols  = dots$css_extra_cols,
        substance_order = dots$substance_order
      ) |>
      set_fixed_css()
  }

  invisible(GT)
}

#' Calculate internal dose
#'
#' Calculates internal dose (D_int) as the product of external concentration
#' ('C_ext' in the 'concentration' table) and exposure rate ('rate' in the
#' 'exposure_rate' table), and stores the results in the 'D_int' column of the
#' 'concentration' table in the GeoTox database.
#'
#' If `sensitivity = TRUE`, 'D_int' will be calculated for sensitivity
#' analysis. Typically this shouldn't be used directly by the user, but rather
#' called by [calc_sensitivity()]. In this case, the function will use the
#' 'concentration_sensitivity' and 'exposure_rate_sensitivity' tables instead of
#' the 'concentration' and 'exposure_rate' tables.
#'
#' @param GT GeoTox object.
#' @param overwrite Logical indicating whether to overwrite existing 'D_int'
#'   values in the 'concentration' table (default FALSE).
#' @param sensitivity Logical indicating whether to simulate internal dose for
#'   sensitivity analysis (default FALSE).
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [calc_response()]
#'
#' @examples
#' # Setup required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   add_exposure_rate_params() |>
#'   simulate_population(exposure = exposure_df, sample_css = FALSE)
#'
#' # Calculate internal dose
#' GT <- GT |> calc_internal_dose()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant tables
#'
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' dplyr::tbl(con, "exposure_rate") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
calc_internal_dose <- function(
    GT, overwrite = FALSE, sensitivity = FALSE
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (sensitivity) {
    conc_name <- "concentration_sensitivity"
    rate_name <- "exposure_rate_sensitivity"
  } else {
    conc_name <- "concentration"
    rate_name <- "exposure_rate"
  }

  if (!DBI::dbExistsTable(con, conc_name)) {
    stop("No '", conc_name, "' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!DBI::dbExistsTable(con, rate_name)) {
    stop("No '", rate_name, "' table found in the GeoTox connection.",
         call. = FALSE)
  }

  conc_tbl <- dplyr::tbl(con, conc_name)
  rate_tbl <- dplyr::tbl(con, rate_name)

  if (!"D_int" %in% colnames(conc_tbl)) {
    sql <- paste("ALTER TABLE", conc_name, "ADD COLUMN D_int DOUBLE")
    DBI::dbExecute(con, sql)
  } else if (!overwrite) {
    stop("GeoTox connection already has a 'D_int' column in the '", conc_name,
         "' table. Set `overwrite = TRUE` to replace the existing values.\n",
         "Warning: Overwriting will not alter downstream values.",
         call. = FALSE)
  }

  update_df <- conc_tbl |>
    dplyr::left_join(
      rate_tbl,
      by = dplyr::join_by("sample_id", "route_id")
    ) |>
    dplyr::mutate(D_int = .data$C_ext * .data$rate) |>
    dplyr::select("id", "D_int")

  # reset_seed can be left as FALSE because it is only altered if the update_df
  # is copied into the database.
  update_table(con, conc_name, update_df, copy = FALSE)

  invisible(GT)
}

#' GeoTox Data
#'
#' Sample data for use in vignettes and function examples. See the Package
#' Data vignette, \code{vignette("package_data", package = "GeoTox")}, for
#' details on how this data was gathered.
#'
#' @format A list with items:
#' \describe{
#'   \item{exposure}{2020 AirToxScreen exposure concentrations for a subset of
#'   chemicals in North Carolina counties.}
#'   \item{dose_response}{Subset of chemicals curated by ICE cHTS as active
#'   within a set of assays.}
#'   \item{age}{County population estimates for 7/1/2020 in North Carolina.}
#'   \item{obesity}{CDC PLACES obesity data for North Carolina counties in
#'   2020.}
#'   \item{simulated_css}{Simulated steady-state plasma concentrations for
#'   various age groups and obesity status combinations.}
#'   \item{boundaries}{County and state boundaries for North Carolina in 2020.}
#' }
"geo_tox_data"

#' Simulate exposure rates
#'
#' Simulate exposure rate values for each sample and exposure route based on
#' parameters in the `exposure_rate_params` table.
#'
#' An 'exposure_rate_params' table must already exist in the GeoTox database,
#' which is added using [add_exposure_rate_params()]. There must also be a
#' 'sample' table with an 'age' column, which can be created using
#' [simulate_age()] or [set_sample()]. 'location' and 'route' tables must also
#' exist, but they are created when adding the rate parameters and samples.
#'
#' If `rate_exta_cols` is provided it will be added to the GeoTox object
#' parameter list, `GT$par`. These columns must exist in the
#' 'exposure_rate_params' table and will be used to match between the 'sample'
#' and 'exposure_rate_params' tables when simulating exposure rates.
#'
#' Typically this function will be called with `sensitivity` set to `FALSE`. In
#' this case, the function will create an 'exposure_rate' table with simulated
#' exposure rates for each sample and exposure route.
#'
#' If `sensitivity` is `TRUE`, exposure rates will be simulated for sensitivity
#' analysis. Typically this shouldn't be used directly by the user, but rather
#' called by [calc_sensitivity()]. In this case, the existing 'exposure_rate'
#' table will be copied to the 'exposure_rate_sensitivity' table where the rate
#' values will be overwritten.
#'
#' @param GT GeoTox object.
#' @param rate_extra_cols Additional columns to match from the
#'   'exposure_rate_params' table (default NULL).
#' @param overwrite Logical indicating whether to overwrite existing
#'   'exposure_rate' table (default FALSE).
#' @param sensitivity Logical indicating whether to simulate exposure rates for
#'   sensitivity analysis (default FALSE).
#'
#' @returns The updated GeoTox object, invisibly.
#' @export
#' @seealso [add_exposure_rate_params()], [simulate_population()]
#'
#' @examples
#' # Setup required tables
#' # Note: 'gender' is ignored when using the default rate params
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight, ~gender,
#'   10000, 25, "Normal",   "male",
#'   10000, 35,  "Obese",   "male",
#'   20000, 50, "Normal", "female"
#' )
#' GT <- GeoTox() |>
#'   add_exposure_rate_params() |>
#'   set_sample(sample_df)
#'
#' # Simulate exposure rates
#' GT |> simulate_exposure_rate()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "exposure_rate") |> dplyr::collect()
#'
#' dplyr::tbl(con, "sample") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Replace exposure rate params with a new table that includes gender
#' params_df <- tibble::tribble(
#'   ~age_lb, ~age_ub, ~gender, ~mean, ~sd,
#'    0, 49,   "male", 10, 1,
#'   50, 99,   "male", 20, 1,
#'    0, 49, "female", 30, 1,
#'   50, 99, "female", 40, 1
#' )
#' DBI::dbRemoveTable(con, "exposure_rate_params")
#' GT |> add_exposure_rate_params(params = params_df)
#'
#' # Overwrite 'rate' values in existing 'exposure_rate' table
#' # Must specify additional column names
#' # Notice how the column names are added to GT$par
#' str(GT$par)
#' GT <- GT |>
#'   simulate_exposure_rate(rate_extra_cols = c("gender"), overwrite = TRUE)
#' str(GT$par)
#'
#' # Look at updated 'exposure_rate' table
#' dplyr::tbl(con, "exposure_rate") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
simulate_exposure_rate <- function(
    GT, rate_extra_cols = NULL, overwrite = FALSE, sensitivity = FALSE
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (!sensitivity) {
    # Create exposure_rate table
    rate_name <- "exposure_rate"
    tables <- DBI::dbListTables(con)
    if (!"sample" %in% tables) {
      stop("No 'sample' table found in the GeoTox connection.", call. = FALSE)
    }
    if (!"location" %in% tables) {
      stop("No 'location' table found in the GeoTox connection.", call. = FALSE)
    }
    if (!"route" %in% tables) {
      stop("No 'route' table found in the GeoTox connection.", call. = FALSE)
    }
    if (!"exposure_rate_params" %in% tables) {
      stop("No 'exposure_rate_params' table found in the GeoTox connection.",
           call. = FALSE)
    }
    if (rate_name %in% tables & !overwrite) {
      stop("GeoTox connection already has an '", rate_name, "' table. ",
           "Set `overwrite = TRUE` to replace it.\n",
           "Warning: Overwriting will not alter downstream values.",
           call. = FALSE)
    }
    rate_tbl <- dplyr::cross_join(
      dplyr::tbl(con, "sample") |>
        dplyr::select("id") |>
        dplyr::rename(sample_id = "id"),
      dplyr::tbl(con, "route") |>
        dplyr::select("id") |>
        dplyr::rename(route_id = "id")
    ) |>
      dplyr::arrange(.data$sample_id, .data$route_id) |>
      dplyr::compute(
        name = rate_name,
        temporary = FALSE,
        overwrite = TRUE
      )
    sql <- paste("ALTER TABLE", rate_name, "ADD COLUMN rate DOUBLE")
    DBI::dbExecute(con, sql)
  } else {
    # Table already created in calc_sensitivity()
    rate_name <- "exposure_rate_sensitivity"
    if (!DBI::dbExistsTable(con, rate_name)) {
      stop("No '", rate_name, "' table found in the GeoTox connection.",
           call. = FALSE)
    }
    rate_tbl <- dplyr::tbl(con, rate_name)
  }

  # Update parameters
  if (!is.null(rate_extra_cols)) {
    set_par(con, "rate_extra_cols", rate_extra_cols)
    GT$par$rate_extra_cols <- rate_extra_cols
  } else if (!is.null(GT$par$rate_extra_cols)) {
    rate_extra_cols <- GT$par$rate_extra_cols
  } else {
    rate_extra_cols <- c()
  }

  # Collect all rate params
  params_cols <- c(
    "route_id", "age_lb", "age_ub", "mean", "sd", rate_extra_cols
  )
  params_df <- dplyr::tbl(con, "exposure_rate_params") |>
    dplyr::select(tidyselect::all_of(params_cols)) |>
    dplyr::collect()

  # Needed columns
  sample_cols <- c("id", "location_id", "age", rate_extra_cols)
  select_cols <- c("sample_id", "route_id", "age", rate_extra_cols)

  # Loop over locations
  location_id <- dplyr::tbl(con, "location") |>
    dplyr::select("id") |>
    dplyr::arrange(.data$id) |>
    dplyr::pull()
  purrr::walk(location_id, function(loc) {
    # Each sample + route should only get mapped to a single param mean + sd, so
    # a "many-to-one" relationship is used. However, this isn't supported on
    # database backends, so collect() needs to be done first.
    rate_df <- rate_tbl |>
      dplyr::left_join(
        dplyr::tbl(con, "sample") |>
          dplyr::select(tidyselect::all_of(sample_cols)),
        by = dplyr::join_by("sample_id" == "id")
      ) |>
      dplyr::filter(.data$location_id == .env$loc) |>
      dplyr::select(tidyselect::all_of(select_cols))
    if (sensitivity) {
      rate_df <- rate_df |>
        dplyr::mutate(age = stats::median(.data$age, na.rm = TRUE))
    }
    rate_df <- rate_df |>
      dplyr::arrange(.data$sample_id, .data$route_id) |>
      dplyr::collect()
    if (nrow(rate_df) == 0) return()
    rate_df <- rate_df |>
      dplyr::left_join(
        params_df,
        by = dplyr::join_by(
          "age" >= "age_lb",
          "age" < "age_ub",
          "route_id",
          !!!rlang::syms(rate_extra_cols)
        ),
        relationship = "many-to-one"
      ) |>
      dplyr::mutate(
        rate = truncnorm::rtruncnorm(1, 0, Inf, .data$mean, .data$sd)
      ) |>
      dplyr::select("sample_id", "route_id", "rate")

    update_table(
      con, rate_name, rate_df, by = c("sample_id", "route_id"),
      reset_seed = GT$par$reset_seed
    )
  })
  invisible(GT)
}

#' Add age simulation data
#'
#' Create or add to the 'age' table in a GeoTox database.
#'
#' The simulation data must have columns "AGEGRP" and "TOT_POP" and at least one
#' column containing location information (default "FIPS"). Each location must
#' have 19 rows for AGEGRP 0-18, where 1-18 are age groups in increments of 5,
#' e.g. AGEGRP = 5 would be ages 20 to 24, and AGEGRP = 0 is the combination of
#' all age groups. The age data is used by [simulate_age()] to generate age
#' samples for each location.
#'
#' The `location` input can be a named vector to specify multiple identifier
#' columns in `df`. For example, `location = c(FIPS = "FIPS", state = "ST")`
#' would indicate that `df` contains both FIPS codes and state identifiers for
#' locations. The `state = "ST"` part would rename the "ST" column in `df` to
#' "state" in the 'location' table.
#'
#' @param GT GeoTox object.
#' @param df Data frame with age simulation data.
#' @param location Column name(s) in `df` that contain location identifier(s)
#'   (default "FIPS").
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [simulate_age()]
#'
#' @examples
#' # Example age simulation data
#' age_df <- data.frame(
#'   FIPS = rep(c(10000, 20000), each = 19),
#'   AGEGRP = rep(0:18, times = 2),
#'   TOT_POP = 0
#' )
#' # FIPS 10000, populate age group 40-44
#' age_df$TOT_POP[c(1, 10)] = 100
#' # FIPS 20000, populate age groups 50-59
#' age_df$TOT_POP[c(1, 12, 13) + 19] = c(200, 100, 100)
#'
#' # Add age simulation data to GeoTox database
#' GT <- GeoTox() |> add_age(age_df)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "age") |> dplyr::filter(TOT_POP > 0) |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
add_age <- function(GT, df, location = "FIPS") {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  foreign_keys <- list("location_id" = location)
  add_table(con, "age", df, foreign_keys, reset_seed = GT$par$reset_seed)
  invisible(GT)
}

#=======================================
# Assay
#=======================================
add_assay <- function(GT, df, substance = "casn") {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  foreign_keys <- list("substance_id" = substance)
  add_table(con, "assay", df, foreign_keys, reset_seed = GT$par$reset_seed)
  invisible(GT)
}

#' Add exposure simulation data
#'
#' Create or add to the 'exposure' table in a GeoTox database.
#'
#' The simulation data must contain columns for exposure mean and standard
#' deviation (default "mean" and "sd", respectfully), at least one column
#' containing location information (default "FIPS"), at least one column
#' containing substance information (default "casn"), and one column containing
#' route information (default "route"). The exposure data is used by
#' [simulate_exposure()] to generate external exposure concentration samples for
#' each location.
#'
#' The `location` and `substance` inputs can be named vectors to specify
#' multiple identifier columns in `df`. For example, `substance = c(casn =
#' "casn", name = "chnm")` would indicate that `df` contains both CAS numbers
#' and chemical names for substances. The `name = "chnm"` part would rename the
#' "chnm" column in `df` to "name" in the 'substance' table.
#'
#' @param GT GeoTox object.
#' @param df Data frame with exposure simulation data.
#' @param location Column name(s) in `df` that contain location identifier(s)
#'   (default "FIPS").
#' @param substance Column name(s) in `df` that contain substance identifier(s)
#'   (default "casn").
#' @param route Column name in `df` that contains route identifier (default
#'   "route").
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [simulate_exposure()]
#'
#' @examples
#' # Example exposure simulation data
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#'
#' # Add exposure simulation data to GeoTox database
#' GT <- GeoTox() |> add_exposure(exposure_df)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#' dplyr::tbl(con, "exposure") |> dplyr::collect()
#' dplyr::tbl(con, "location") |> dplyr::collect()
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Add another substance with a new field and route
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~chnm, ~route, ~mean, ~sd,
#'   10000, "00-00-3", "chem3", "drinking", 100, 1,
#'   20000, "00-00-3", "chem3", "drinking", 200, 1
#' )
#' GT |> add_exposure(exposure_df, substance = c(casn = "casn", name = "chnm"))
#'
#' # Look at updated tables
#'
#' dplyr::tbl(con, "exposure") |> dplyr::collect()
#'
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#'
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
add_exposure <- function(
    GT, df, location = "FIPS", substance = "casn", route = "route"
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  foreign_keys = list(
    "location_id" = location,
    "substance_id" = substance,
    "route_id" = route
  )
  add_table(con, "exposure", df, foreign_keys, reset_seed = GT$par$reset_seed)
  invisible(GT)
}

#' Add exposure rate simulation data
#'
#' Create or add to the 'exposure_rate_params' table in a GeoTox database.
#'
#' There are two routes with built-in default exposure rate parameters:
#' "inhalation" and "drinking". If `params` is not provided (`NULL`), the
#' default parameters for the specified `route` will be used. If `params` is
#' provided, it must contain columns "age_lb", "age_ub", "mean", and "sd". The
#' exposure rate parameters are used by [simulate_exposure_rate()] to generate
#' exposure rate samples for each individual in the population.
#'
#' The build-in parameters come from the following sources:
#' * Inhalation: [EPA Exposure Factors Handbook (2015), Table 6.7](https://www.epa.gov/sites/default/files/2015-09/documents/efh-chapter06.pdf)
#' in m\eqn{^3}/kg-day. The "mean" and "sd" values are the average of male and
#' female values.
#' * Drinking: [EPA Exposure Factors Handbook (2019), Table 3-30](https://www.epa.gov/sites/default/files/2019-02/documents/efh_-_chapter_3_update.pdf)
#' in mL/kg-day.
#'
#' @param GT GeoTox object.
#' @param route Exposure route (default "inhalation").
#' @param params Data frame with exposure rate parameters. If `NULL`, default
#'   parameters for the specified `route` will be used (default `NULL`).
#' @param overwrite Logical indicating whether to overwrite existing exposure
#'   rate parameters for the specified `route` (default `FALSE`).
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [simulate_exposure_rate()]
#'
#' @examples
#' # Add both default params to GeoTox database
#' GT <- GeoTox() |>
#'   add_exposure_rate_params() |>
#'   add_exposure_rate_params(route = "drinking")
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant tables
#' params_tbl <- dplyr::tbl(con, "exposure_rate_params") |> dplyr::collect()
#'
#' params_tbl |> dplyr::filter(route_id == 1)
#'
#' params_tbl |> dplyr::filter(route_id == 2)
#'
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Add custom params with new column (gender), assign to route "custom"
#' params_df <- tibble::tribble(
#'   ~age_lb, ~age_ub, ~gender, ~mean, ~sd,
#'    0, 49, "male",   10, 1,
#'   50, 99, "male",   20, 1,
#'    0, 49, "female", 30, 1,
#'   50, 99, "female", 40, 1
#' )
#' GT |> add_exposure_rate_params(params_df, route = "custom")
#'
#' # Look at updated tables
#' params_tbl <- dplyr::tbl(con, "exposure_rate_params") |> dplyr::collect()
#'
#' params_tbl |> dplyr::filter(route_id == 1)
#'
#' params_tbl |> dplyr::filter(route_id == 2)
#'
#' params_tbl |> dplyr::filter(route_id == 3)
#'
#' dplyr::tbl(con, "route") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
add_exposure_rate_params <- function(
    GT, route = "inhalation", params = NULL, overwrite = FALSE
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  if (is.null(params)) {
    if (route == "inhalation") {
      # https://www.epa.gov/sites/default/files/2015-09/documents/efh-chapter06.pdf
      # Table 6.7 Distribution percentiles of physiological daily inhalation
      # rates per unit body weight (m^3/kg-day) for free living normal weight
      # males and females aged 2 months to 96 years
      params <- tibble::tribble(
        ~age, ~male.mean, ~male.sd, ~female.mean, ~female.sd,
        0, 0.495, 0.08, 0.48, 0.075,
        1, 0.48,  0.06, 0.45, 0.08,
        2, 0.44,  0.04, 0.44, 0.07,
        5, 0.42,  0.05, 0.40, 0.05,
        7, 0.37,  0.06, 0.35, 0.06,
        11, 0.30,  0.05, 0.27, 0.05,
        23, 0.25,  0.04, 0.23, 0.04,
        30, 0.24,  0.03, 0.24, 0.04,
        40, 0.23,  0.04, 0.21, 0.04,
        65, 0.19,  0.03, 0.17, 0.04
      ) |>
        dplyr::rename(age_lb = "age") |>
        dplyr::mutate(
          age_ub = dplyr::lead(.data$age_lb, default = 150),
          mean = rowMeans(dplyr::pick("male.mean", "female.mean")),
          sd = rowMeans(dplyr::pick("male.sd", "female.sd")),
          route = .env$route
        ) |>
        dplyr::select("route", "age_lb", "age_ub", "mean", "sd")
    } else if (route == "drinking") {
      # https://www.epa.gov/sites/default/files/2019-02/documents/efh_-_chapter_3_update.pdf
      # Table 3-30
      # mL/kg-day
      params <- tibble::tribble(
        ~age, ~mean, ~sd,
        0, 43.5, 42.5,
        1, 46.8, 28.1,
        4, 37.9, 21.8,
        7, 26.9, 15.3,
        11, 20.2, 11.6,
        15, 16.4,  9.6,
        20, 18.6, 10.7,
        45, 22,   10.8,
        65, 21.9,  9.9,
        75, 21.6,  9.5
      ) |>
        dplyr::rename(age_lb = "age") |>
        dplyr::mutate(
          age_ub = dplyr::lead(.data$age_lb, default = 150),
          route = route
        ) |>
        dplyr::select("route", "age_lb", "age_ub", "mean", "sd")
    } else {
      stop("Default exposure params do not exist for route '", route, "'.",
           call. = FALSE)
    }
  } else {
    cols <- c("age_lb", "age_ub", "mean", "sd")
    if (!all(cols %in% names(params))) {
      stop("`params` must contain the following columns: ",
           paste(cols, collapse = ", "), ".", call. = FALSE)
    }
    cols <- c("route", cols)
    if ("route" %in% names(params)) {
      stop("`params` should not contain a 'route' column. ",
           "The route is set by the `route` argument.", call. = FALSE)
    }
    params <- params |>
      dplyr::mutate(route = .env$route) |>
      dplyr::select(tidyselect::all_of(c(cols, setdiff(names(params), cols))))
  }
  add_table(
    con, "route", data.frame(route = route), reset_seed = GT$par$reset_seed
  )
  if (DBI::dbExistsTable(con, "exposure_rate_params")) {
    route_id <- dplyr::tbl(con, "route") |>
      dplyr::filter(.data$route == .env$route) |>
      dplyr::pull("id")
    n <- dplyr::tbl(con, "exposure_rate_params") |>
      dplyr::filter(.data$route_id == .env$route_id) |>
      dplyr::count() |>
      dplyr::pull("n")
    if (n > 0) {
      if (!overwrite) {
        stop("Exposure rate parameters for route '", route, "' already exist. ",
             "Set `overwrite = TRUE` to replace them.", call. = FALSE)
      } else {
        sql <- paste0(
          "DELETE FROM exposure_rate_params WHERE route_id = ", route_id
        )
        DBI::dbExecute(con, sql)
      }
    }
  }
  add_table(
    con, "exposure_rate_params", params,
    foreign_keys = list("route_id" = "route"), reset_seed = GT$par$reset_seed
  )
  invisible(GT)
}

#' Add Hill model parameters
#'
#' Create or add to the 'hill_params' table in a GeoTox database.
#'
#' @param GT GeoTox object.
#' @param hill_params List output from [fit_hill()].
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [fit_hill()]
#'
#' @examples
#' # Example Hill model data
#' hill_df <- tibble::tribble(
#'   ~assay, ~model, ~casn, ~logc, ~resp,
#'   "a1", "human", "00-00-1",    0,  10,
#'   "a1", "human", "00-00-1",    1,  20,
#'   "a1", "human", "00-00-1",    2,  80,
#'   "a1", "human", "00-00-1",    3, 100,
#'   "a1", "human", "00-00-2", -0.5,   5,
#'   "a1", "human", "00-00-2",  0.5,  20,
#'   "a1", "human", "00-00-2",  1.5,  55,
#'   "a1", "human", "00-00-2",  2.5,  60,
#'   "a2",   "rat", "00-00-1",   -1,   0,
#'   "a2",   "rat", "00-00-1",    0,  10,
#'   "a2",   "rat", "00-00-1",    1,  30,
#'   "a2",   "rat", "00-00-1",    2,  40
#' )
#' hill_params <- fit_hill(
#'   hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
#' )
#'
#' # Add Hill model parameters to GeoTox database
#' GT <- GeoTox() |> add_hill_params(hill_params)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "hill_params") |> dplyr::collect()
#'
#' dplyr::tbl(con, "assay") |> dplyr::collect()
#'
#' dplyr::tbl(con, "substance") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
add_hill_params <- function(GT, hill_params) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  if (is.null(hill_params$assay) & is.null(hill_params$substance)) {
    foreign_keys <- NULL
  } else {
    foreign_keys <- list()
    if (!is.null(hill_params$assay)) {
      foreign_keys$assay_id <- hill_params$assay
    }
    if (!is.null(hill_params$substance)) {
      foreign_keys$substance_id <- hill_params$substance
    }
  }
  add_table(
    con, "hill_params", hill_params$fit, foreign_keys,
    reset_seed = GT$par$reset_seed
  )
  invisible(GT)
}

#=======================================
# Location
#=======================================
add_location <- function(GT, df) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  add_table(con, "location", df, reset_seed = GT$par$reset_seed)
  invisible(GT)
}

#' Add obesity simulation data
#'
#' Create or add to the 'obesity' table in a GeoTox database.
#'
#' The simulation data must contain columns for obesity prevalence and standard
#' deviation (default "OBESITY_CrudePrev" and "OBESITY_SD", respectfully) and at
#' least one column containing location information (default "FIPS"). The
#' obesity data is used by [simulate_obesity()] to generate weight category
#' samples for each location.
#'
#' The `location` input can be a named vector to specify multiple identifier
#' columns in `df`. For example, `location = c(FIPS = "FIPS", state = "ST")`
#' would indicate that `df` contains both FIPS codes and state identifiers for
#' locations. The `state = "ST"` part would rename the "ST" column in `df` to
#' "state" in the 'location' table.
#'
#' @param GT GeoTox object.
#' @param df Data frame with obesity simulation data.
#' @param location Column name(s) in `df` that contain location identifier(s)
#'   (default "FIPS").
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [simulate_obesity()]
#'
#' @examples
#' # Example obesity simulation data
#' obesity_df <- data.frame(
#'   FIPS = c(10000, 20000),
#'   OBESITY_CrudePrev = c(20, 80),
#'   OBESITY_SD = 5
#' )
#'
#' # Add obesity simulation data to GeoTox database
#' GT <- GeoTox() |> add_obesity(obesity_df)
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at created tables
#'
#' dplyr::tbl(con, "obesity") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' # Add another location with additional information
#' obesity_df <- data.frame(
#'  FIPS = 30000,
#'  ST = "State3",
#'  OBESITY_CrudePrev = 50,
#'  OBESITY_SD = 10
#' )
#' GT |> add_obesity(obesity_df, location = c(FIPS = "FIPS", state = "ST"))
#'
#' # Look at updated tables
#'
#' dplyr::tbl(con, "obesity") |> dplyr::collect()
#'
#' dplyr::tbl(con, "location") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
add_obesity <- function(GT, df, location = "FIPS") {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  foreign_keys <- list("location_id" = location)
  add_table(con, "obesity", df, foreign_keys, reset_seed = GT$par$reset_seed)
  invisible(GT)
}

#=======================================
# Route
#=======================================
add_route <- function(GT, df) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  add_table(con, "route", df, reset_seed = GT$par$reset_seed)
  invisible(GT)
}

#=======================================
# Substance
#=======================================
add_substance <- function(GT, df) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  add_table(con, "substance", df, reset_seed = GT$par$reset_seed)
  invisible(GT)
}

test_that("add tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  #=====================================
  # Tables w/o foreign keys
  #=====================================

  # exposure_rate_params
  params_df <- tibble::tribble(
    ~age_lb, ~age_ub, ~gender, ~other, ~mean, ~sd,
     0, 49, "male",   "other", 10, 2,
    50, 99, "male",   "other", 20, 3,
     0, 49, "female", "other", 15, 4,
    50, 99, "female", "other", 25, 5
  )
  GT |>
    add_exposure_rate_params() |>
    add_exposure_rate_params(route = "drinking") |>
    add_exposure_rate_params(route = "custom", params = params_df)
  # no default values
  expect_error(
    add_exposure_rate_params(GT, route = "dermal"),
    "Default exposure params do not exist for route 'dermal'."
  )
  # bad colnames
  params_df <- tibble::tribble(
    ~age_lb, ~age_ub, ~gender, ~other, ~mean,
     0, 49, "male",   "other", 10,
    50, 99, "male",   "other", 20,
     0, 49, "female", "other", 15,
    50, 99, "female", "other", 25
  )
  expect_error(
    add_exposure_rate_params(GT, route = "bad_params", params = params_df),
    "`params` must contain the following columns:"
  )
  params_df$sd <- 2
  params_df$route <- "bad_params"
  expect_error(
    add_exposure_rate_params(GT, route = "bad_params", params = params_df),
    "`params` should not contain a 'route' column."
  )
  # overwrite
  expect_error(
    add_exposure_rate_params(GT),
    "Exposure rate parameters for route 'inhalation' already exist."
  )
  add_exposure_rate_params(GT, overwrite = TRUE)
  expect_equal(
    dplyr::tbl(con, "exposure_rate_params") |> dplyr::collect() |>
      dplyr::summarize(n = dplyr::n(), .by = route_id) |>
      dplyr::arrange(route_id) |> dplyr::collect(),
    tibble::tribble(
      ~route_id, ~n,
      1, 10,
      2, 10,
      3, 4
    )
  )

  # location
  location_df <- data.frame(FIPS = c("00001", "00002"))
  add_location(GT, location_df)
  expect_equal(
    dplyr::tbl(con, "location") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~FIPS,
      1, "00001",
      2, "00002"
    )
  )

  # route
  route_df <- data.frame(route = "dermal")
  add_route(GT, route_df)
  expect_equal(
    dplyr::tbl(con, "route") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~route,
      1, "inhalation",
      2, "drinking",
      3, "custom",
      4, "dermal"
    )
  )

  # substance
  substance_df <- data.frame(
    casn = c("50-00-0", "50-00-1"),
    chnm = c("chem1", "chem2")
  )
  add_substance(GT, substance_df)
  expect_equal(
    dplyr::tbl(con, "substance") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~casn, ~chnm,
      1, "50-00-0", "chem1",
      2, "50-00-1", "chem2"
    )
  )

  #=====================================
  # Tables w/ foreign keys
  #=====================================

  # age
  age_df <- tibble::tribble(
    ~FIPS, ~age_group, ~count,
    "00001", 0, 1000,
    "00001", 1, 1200,
    "00002", 0, 800,
    "00002", 1, 900
  )
  add_age(GT, age_df)
  expect_equal(
    dplyr::tbl(con, "age") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~location_id, ~age_group, ~count,
      1, 1, 0, 1000,
      2, 1, 1, 1200,
      3, 2, 0, 800,
      4, 2, 1, 900
    )
  )

  # assay
  assay_df <- tibble::tribble(
    ~name, ~casn,
    "assay1", "50-00-0",
    "assay2", "50-00-0",
    "assay1", "50-00-1"
  )
  add_assay(GT, assay_df)
  expect_equal(
    dplyr::tbl(con, "assay") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~substance_id, ~name,
      1, 1, "assay1",
      2, 1, "assay2",
      3, 2, "assay1"
    )
  )

  # exposure
  exposure_df <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~value,
    "00001", "50-00-0", "inhalation", 5,
    "00001", "50-00-1", "drinking", 10,
    "00002", "50-00-0", "inhalation", 7
  )
  add_exposure(GT, exposure_df)
  expect_equal(
    dplyr::tbl(con, "exposure") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~location_id, ~substance_id, ~route_id, ~value,
      1, 1, 1, 1, 5,
      2, 1, 2, 2, 10,
      3, 2, 1, 1, 7
    )
  )

  # obesity
  obesity_df <- tibble::tribble(
    ~FIPS, ~prev,
    "00001", 0.3,
    "00002", 0.25
  )
  add_obesity(GT, obesity_df)
  expect_equal(
    dplyr::tbl(con, "obesity") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~location_id, ~prev,
      1, 1, 0.3,
      2, 2, 0.25
    )
  )

  #=====================================
  # Tables w/ optional foreign keys
  #=====================================

  # hill_params w/o foreign keys
  hill_params <- list(
    fit = data.frame(par1 = 1:2, par2 = 3:4),
    assay = NULL,
    substance = NULL
  )
  add_hill_params(GT, hill_params)
  expect_equal(
    dplyr::tbl(con, "hill_params") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~par1, ~par2,
      1, 1, 3,
      2, 2, 4
    )
  )

  # hill_params w/o foreign keys
  hill_params$fit$name = c("assay1", "assay2")
  hill_params$fit$casn = c("50-00-0", "50-00-1")
  hill_params$assay = "name"
  hill_params$substance = "casn"
  add_hill_params(GT, hill_params)
  expect_equal(
    dplyr::tbl(con, "hill_params") |> dplyr::collect(),
    tibble::tribble(
      ~id, ~par1, ~par2, ~assay_id, ~substance_id,
      1, 1, 3, 1, 1,
      2, 2, 4, 2, 2
    )
  )
})

test_that("sensitivity analysis", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  set.seed(1234)

  # Setup tables
  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~weight,
    "00001", 25, "Normal",
    "00001", 35, "Normal",
    "00002", 45, "Obese",
  )
  exposure_df <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mean, ~sd,
    "00001", "50-00-0", "inhalation", 10, 2,
    "00001", "50-00-1", "inhalation", 20, 5,
    "00002", "50-00-0", "inhalation", 15, 3,
    "00002", "50-00-1", "inhalation", 25, 4
  )
  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight, ~css,
    "50-00-0",  0, 44, "Normal", 1,
    "50-00-0", 45, 89, "Normal", 2,
    "50-00-0",  0, 44, "Obese", 11,
    "50-00-0", 45, 89, "Obese", 12,
    "50-00-1",  0, 44, "Normal", 3,
    "50-00-1", 45, 89, "Normal", 4,
    "50-00-1",  0, 44, "Obese", 13,
    "50-00-1", 45, 89, "Obese", 14
  )
  hill_df <- rbind(
    data.frame(
      casn = "50-00-0",
      logc = -3:3,
      resp = 5 / (1 + 10^(1.2 * (0.4 - rep(-3:3, each = 3)))) + rnorm(21)
    ),
    data.frame(
      casn = "50-00-01",
      logc = -3:3,
      resp = 10 / (1 + 10^(0.8 * (1.4 - rep(-3:3, each = 3)))) + rnorm(21)
    )
  )
  hill_fit <- fit_hill(hill_df, substance = "casn")
  GT <- GT |>
    set_sample(sample_df) |>
    set_simulated_css(css_df) |>
    add_exposure_rate_params() |>
    add_exposure(exposure_df) |>
    add_hill_params(hill_fit) |>
    simulate_exposure_rate() |>
    simulate_exposure() |>
    sample_simulated_css() |>
    set_fixed_css()

  expect_no_error(
    sensitivity_analysis(GT)
  )

  # Check results
  tables <- c(
    "risk_sensitivity_age",
    "risk_sensitivity_weight",
    "risk_sensitivity_css_params",
    "risk_sensitivity_fit_params",
    "risk_sensitivity_C_ext"
  )
  res <- tables |>
    purrr::map(\(x) dplyr::tbl(con, x) |> dplyr::collect()) |>
    rlang::set_names(tables)
  res <- res |>
    purrr::map(\(x) x |>
      dplyr::mutate(dplyr::across(tidyselect::where(is.numeric), \(x) {
        signif(x, 4)
      }))
    )
  expect_snapshot(res)
})

test_that("missing tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    calc_sensitivity(GT),
    "No 'exposure_rate' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "exposure_rate", data.frame(id = 1))

  expect_error(
    calc_sensitivity(GT),
    "No 'concentration' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "concentration", data.frame(id = 1))

  expect_error(
    calc_sensitivity(GT),
    "No 'fixed_css' table found in the GeoTox connection."
  )
})

test_that("calc sensitivity", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  set.seed(1234)

  # Setup tables
  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~weight,
    "00001", 25, "Normal",
    "00001", 35, "Normal",
    "00002", 45, "Obese",
  )
  exposure_df <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mean, ~sd,
    "00001", "50-00-0", "inhalation", 10, 2,
    "00001", "50-00-1", "inhalation", 20, 5,
    "00002", "50-00-0", "inhalation", 15, 3,
    "00002", "50-00-1", "inhalation", 25, 4
  )
  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight, ~css,
    "50-00-0",  0, 44, "Normal", 1,
    "50-00-0", 45, 89, "Normal", 2,
    "50-00-0",  0, 44, "Obese", 11,
    "50-00-0", 45, 89, "Obese", 12,
    "50-00-1",  0, 44, "Normal", 3,
    "50-00-1", 45, 89, "Normal", 4,
    "50-00-1",  0, 44, "Obese", 13,
    "50-00-1", 45, 89, "Obese", 14
  )
  hill_df <- rbind(
    data.frame(
      casn = "50-00-0",
      logc = -3:3,
      resp = 5 / (1 + 10^(1.2 * (0.4 - rep(-3:3, each = 3)))) + rnorm(21)
    ),
    data.frame(
      casn = "50-00-01",
      logc = -3:3,
      resp = 10 / (1 + 10^(0.8 * (1.4 - rep(-3:3, each = 3)))) + rnorm(21)
    )
  )
  hill_fit <- fit_hill(hill_df, substance = "casn")
  GT <- GT |>
    set_sample(sample_df) |>
    set_simulated_css(css_df) |>
    add_exposure_rate_params() |>
    add_exposure(exposure_df) |>
    add_hill_params(hill_fit) |>
    simulate_exposure_rate() |>
    simulate_exposure() |>
    sample_simulated_css() |>
    set_fixed_css()

  expect_no_error(
    calc_sensitivity(GT)
  )

  # Overwrite
  expect_error(
    calc_sensitivity(GT),
    "GeoTox connection already has a 'risk_sensitivity_age' table."
  )
  calc_sensitivity(GT, overwrite = TRUE)

  # Check results
  sens_tbl <- dplyr::tbl(con, "risk_sensitivity_age") |> dplyr::collect()
  sens_tbl <- sens_tbl |>
    dplyr::mutate(dplyr::across(tidyselect::where(is.numeric), \(x) {
      signif(x, 4)
    }))
  expect_snapshot(sens_tbl)
})

test_that("other inputs", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  set.seed(1234)

  # Setup tables
  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~weight, ~gender, ~other,
    "00001", 25, "Normal", "female", "other",
    "00001", 35, "Normal", "male",   "other",
    "00002", 45, "Obese",  "female", "other"
  )
  params <- tibble::tribble(
    ~age_lb, ~age_ub, ~gender, ~other, ~mean, ~sd,
    20, 30, "male",   "other",  0.1, 2,
    30, 40, "male",   "other",  1,   0,
    40, 50, "male",   "other", 10,   2,
    20, 30, "female", "other", 25,   0,
    30, 40, "female", "other", 50,   3,
    40, 50, "female", "other", 75,   3
  )
  exposure_df <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mu, ~sigma,
    "00001", "50-00-0", "custom", 10, 2,
    "00001", "50-00-1", "custom", 20, 5,
    "00002", "50-00-0", "custom", 15, 3,
    "00002", "50-00-1", "custom", 25, 4
  )
  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight, ~css,
    "50-00-0",  0, 44, "Normal", 1,
    "50-00-0", 45, 89, "Normal", 2,
    "50-00-0",  0, 44, "Obese", 11,
    "50-00-0", 45, 89, "Obese", 12,
    "50-00-1",  0, 44, "Normal", 3,
    "50-00-1", 45, 89, "Normal", 4,
    "50-00-1",  0, 44, "Obese", 13,
    "50-00-1", 45, 89, "Obese", 14
  )
  hill_df <- rbind(
    data.frame(
      casn = "50-00-0",
      logc = -3:3,
      resp = 5 / (1 + 10^(1.2 * (0.4 - rep(-3:3, each = 3)))) + rnorm(21)
    ),
    data.frame(
      casn = "50-00-01",
      logc = -3:3,
      resp = 10 / (1 + 10^(0.8 * (1.4 - rep(-3:3, each = 3)))) + rnorm(21)
    )
  )
  hill_fit <- fit_hill(hill_df, substance = "casn")
  GT <- GT |>
    set_sample(sample_df) |>
    set_simulated_css(css_df) |>
    add_exposure_rate_params(route = "custom", params = params) |>
    add_exposure(exposure_df) |>
    add_hill_params(hill_fit) |>
    simulate_exposure_rate(rate_extra_cols = c("gender", "other")) |>
    simulate_exposure(expos_mean = "mu", expos_sd = "sigma") |>
    sample_simulated_css() |>
    set_fixed_css()

  expect_no_error(
    calc_sensitivity(GT, vary = "C_ext")
  )

  # Check results
  sens_tbl <- dplyr::tbl(con, "risk_sensitivity_C_ext") |> dplyr::collect()
  sens_tbl <- sens_tbl |>
    dplyr::mutate(dplyr::across(tidyselect::where(is.numeric), \(x) {
      signif(x, 4)
    }))
  expect_snapshot(sens_tbl)
})

test_that("missing tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    simulate_exposure_rate(GT),
    "No 'sample' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "sample", data.frame(id = 1, location_id = 1, age = 30))

  expect_error(
    simulate_exposure_rate(GT),
    "No 'location' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "location", data.frame(id = 1, FIPS = "00001"))

  expect_error(
    simulate_exposure_rate(GT),
    "No 'route' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "route", data.frame(id = 1, name = "inhalation"))

  expect_error(
    simulate_exposure_rate(GT),
    "No 'exposure_rate_params' table found in the GeoTox connection."
  )

  expect_error(
    simulate_exposure_rate(GT, sensitivity = TRUE),
    "No 'exposure_rate_sensitivity' table found in the GeoTox connection."
  )

})

test_that("custom params", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~gender, ~other,
    "37003", 25, "female", "other",
    "37003", 35, "male",   "other",
    "37005", 45, "female", "other"
  )

  params <- tibble::tribble(
    ~age_lb, ~age_ub, ~gender, ~other, ~mean, ~sd,
    20, 30, "male",   "other",  0.1, 2,
    30, 40, "male",   "other",  1,   0,
    40, 50, "male",   "other", 10,   2,
    20, 30, "female", "other", 25,   0,
    30, 40, "female", "other", 50,   3,
    40, 50, "female", "other", 75,   3
  )

  GT <- GT |>
    set_sample(sample_df) |>
    add_exposure_rate_params(route = "custom", params = params) |>
    simulate_exposure_rate(rate_extra_cols = c("gender", "other"))

  expect_equal(GT$par$rate_extra, c("gender", "other"))

  # Using stored GT$par
  simulate_exposure_rate(GT, overwrite = TRUE)

  rate_tbl <- dplyr::tbl(con, "exposure_rate") |> dplyr::collect()
  expect_equal(rate_tbl$rate[1:2], c(25, 1))
})

test_that("simulate exposure rate", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  sample_df <- tibble::tribble(
    ~FIPS, ~age,
    "37003", 25,
    "37003", 35,
    "37005", 45
  )

  GT |>
    set_sample(sample_df) |>
    add_location(data.frame(FIPS = "37007")) |> # extra location
    add_exposure_rate_params(route = "inhalation") |>
    simulate_exposure_rate()

  rate_tbl <- dplyr::tbl(con, "exposure_rate") |> dplyr::collect()
  expect_equal(nrow(rate_tbl), 3)
  expect_true(all(!is.na(rate_tbl$rate)))

  # Overwrite
  expect_error(
    simulate_exposure_rate(GT),
    "GeoTox connection already has an 'exposure_rate' table."
  )

  # Sensitivity
  rate_sens_tbl <- dplyr::tbl(con, "exposure_rate") |>
    dplyr::compute(
      name = "exposure_rate_sensitivity",
      temporary = FALSE
    )
  simulate_exposure_rate(GT, sensitivity = TRUE)
  rate_sens_tbl <- dplyr::tbl(con, "exposure_rate_sensitivity") |> dplyr::collect()
  expect_equal(nrow(rate_tbl), 3)
  expect_true(all(!is.na(rate_sens_tbl$rate)))
})

#' Calculate in vitro concentration
#'
#' Calculates the in vitro concentration (C_invitro) as the product of steady
#' state plasma concentration (C_ss) and internal dose (D_int), and stores the
#' results in the 'C_invitro' column of the 'concentration' table in the GeoTox
#' database.
#'
#' If `sensitivity = TRUE`, 'C_invitro' will be calculated for sensitivity
#' analysis. Typically this shouldn't be used directly by the user, but rather
#' called by [calc_sensitivity()]. In this case, the function will use the
#' 'concentration_sensitivity' table instead of the 'concentration' table.
#'
#' @param GT GeoTox object.
#' @param overwrite Logical indicating whether to overwrite existing 'C_invitro'
#'   values in the 'concentration' table (default FALSE).
#' @param sensitivity Logical indicating whether to simulate in vitro
#'   concentration for sensitivity analysis (default FALSE).
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [calc_internal_dose()], [calc_response()]
#'
#' @examples
#' # Setup required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal", 21,
#'   "00-00-1", 50, 99, "Normal", 22,
#'   "00-00-1",  0, 49,  "Obese", 61,
#'   "00-00-1", 50, 99,  "Obese", 62,
#'   "00-00-2",  0, 49, "Normal", 11,
#'   "00-00-2", 50, 99, "Normal", 12,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   set_simulated_css(css_df) |>
#'   add_exposure_rate_params() |>
#'   simulate_population(exposure = exposure_df) |>
#'   calc_internal_dose()
#'
#' # Calculate in vitro concentration
#' GT <- GT |> calc_invitro_concentration()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant tables
#' dplyr::tbl(con, "concentration") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
calc_invitro_concentration <- function(
    GT, overwrite = FALSE, sensitivity = FALSE
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  if (sensitivity) {
    conc_name <- "concentration_sensitivity"
  } else {
    conc_name <- "concentration"
  }

  if (!DBI::dbExistsTable(con, conc_name)) {
    stop("No '", conc_name, "' table found in the GeoTox connection.",
         call. = FALSE)
  }

  conc_tbl <- dplyr::tbl(con, conc_name)

  if (!"C_invitro" %in% colnames(conc_tbl)) {
    sql <- paste("ALTER TABLE", conc_name, "ADD COLUMN C_invitro DOUBLE")
    DBI::dbExecute(con, sql)
  } else {
    if (!overwrite) {
      stop("GeoTox connection already has a 'C_invitro' column in the '",
           conc_name, "' table. Set `overwrite = TRUE` to replace the ",
           "existing values.\n",
           "Warning: Overwriting will not alter downstream values.",
           call. = FALSE)
    }
  }

  update_df <- conc_tbl |>
    dplyr::mutate(C_invitro = .data$C_ss * .data$D_int) |>
    dplyr::select("id", "C_invitro")

  # reset_seed can be left as FALSE because it is only altered if the update_df
  # is copied into the database.
  update_table(con, conc_name, update_df, copy = FALSE)

  invisible(GT)
}

#' Calculate risk scores
#'
#' Calculate generalized concentration addition (GCA) and independent action
#' (IA) risk scores and hazard quotients (HQ) based on in vitro concentration
#' data and Hill parameters.
#'
#' This function requires that the 'concentration', 'sample', and 'hill_params'
#' tables are present in the GeoTox object. Typically these tables will have
#' been created by prior calls to [calc_invitro_concentration()] and
#' [add_hill_params()].
#'
#' The risk scores are calculated for each sample and assay combination and
#' stored in the 'risk' table in the GeoTox object, unless a different
#' `risk_name` is provided. Supplying a different `risk_name` shouldn't be done
#' directly by the user, but rather by calling [calc_sensitivity()].
#'
#' @param GT GeoTox object.
#' @param max_mult Upper bound multiplier for max response (default 1.5).
#' @param fixed Logical indicating whether to set standard deviation parameters
#'   of Hill fit to zero (default FALSE). Used in sensitivity analysis.
#' @param overwrite Logical indicating whether to overwrite existing risk table
#'   (default FALSE).
#' @param risk_name Table name to store risk results (default "risk"). Values
#'   other than "risk" are used in sensitivity analysis.
#'
#' @returns The same GeoTox object, invisibly.
#' @export
#' @seealso [add_hill_params()], [calc_invitro_concentration()],
#'   [calc_response()]
#'
#' @examples
#' # Setup required tables
#' sample_df <- tibble::tribble(
#'   ~FIPS, ~age, ~weight,
#'   10000, 25, "Normal",
#'   10000, 35,  "Obese",
#'   20000, 50, "Normal"
#' )
#' exposure_df <- tibble::tribble(
#'   ~FIPS, ~casn, ~route, ~mean, ~sd,
#'   10000, "00-00-1", "inhalation", 10, 1,
#'   10000, "00-00-2", "inhalation", 20, 1,
#'   20000, "00-00-1", "inhalation", 30, 1,
#'   20000, "00-00-2", "inhalation", 40, 1
#' )
#' css_df <- tibble::tribble(
#'   ~casn, ~age_lb, ~age_ub, ~weight, ~css,
#'   "00-00-1",  0, 49, "Normal", 21,
#'   "00-00-1", 50, 99, "Normal", 22,
#'   "00-00-1",  0, 49,  "Obese", 61,
#'   "00-00-1", 50, 99,  "Obese", 62,
#'   "00-00-2",  0, 49, "Normal", 11,
#'   "00-00-2", 50, 99, "Normal", 12,
#'   "00-00-2",  0, 49,  "Obese", 31,
#'   "00-00-2", 50, 99,  "Obese", 32
#' )
#' hill_df <- tibble::tribble(
#'   ~assay, ~casn, ~logc, ~resp,
#'   "a1", "00-00-1",    0,  10,
#'   "a1", "00-00-1",    1,  20,
#'   "a1", "00-00-1",    2,  80,
#'   "a1", "00-00-1",    3, 100,
#'   "a1", "00-00-2", -0.5,   5,
#'   "a1", "00-00-2",  0.5,  20,
#'   "a1", "00-00-2",  1.5,  55,
#'   "a1", "00-00-2",  2.5,  60
#' )
#' GT <- GeoTox() |>
#'   set_sample(sample_df) |>
#'   set_simulated_css(css_df) |>
#'   add_exposure_rate_params() |>
#'   add_hill_params(fit_hill(hill_df, assay = "assay", substance = "casn")) |>
#'   simulate_population(exposure = exposure_df) |>
#'   calc_internal_dose() |>
#'   calc_invitro_concentration()
#'
#' # Calculate risk
#' GT <- GT |> calc_risk()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # Look at relevant table
#' dplyr::tbl(con, "risk") |> dplyr::collect()
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
calc_risk <- function(
    GT, max_mult = 1.5, fixed = FALSE, overwrite = FALSE, risk_name = "risk"
) {
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))

  tables <- DBI::dbListTables(con)

  if (risk_name == "risk") {
    conc_name <- "concentration"
  } else {
    conc_name <- "concentration_sensitivity"
  }
  if (!conc_name %in% tables) {
    stop("No '", conc_name, "' table found in the GeoTox connection.",
         call. = FALSE)
  }
  if (!"sample" %in% tables) {
    stop("No 'sample' table found in the GeoTox connection.", call. = FALSE)
  }
  if (!"hill_params" %in% tables) {
    stop("No 'hill_params' table found in the GeoTox connection.",
         call. = FALSE)
  }

  conc_tbl   <- dplyr::tbl(con, conc_name)
  sample_tbl <- dplyr::tbl(con, "sample")
  hill_tbl   <- dplyr::tbl(con, "hill_params")

  # Check for required columns
  conc_cols <- c("sample_id", "substance_id", "C_invitro")
  sample_cols <- c("id", "location_id")
  hill_cols <- c(
    "substance_id", "tp", "tp.sd", "logAC50", "logAC50.sd", "logc_min",
    "logc_max", "resp_min", "resp_max"
  )
  if ("assay_id" %in% colnames(hill_tbl)) {
    hill_cols <- c("assay_id", hill_cols)
    assay_exists <- TRUE
    update_by <- c("assay_id", "sample_id")
  } else {
    assay_exists <- FALSE
    update_by <- c("sample_id")
  }
  conc_tbl <- conc_tbl |>
    dplyr::select(tidyselect::all_of(conc_cols))
  sample_tbl <- sample_tbl |>
    dplyr::select(tidyselect::all_of(sample_cols))
  hill_df <- hill_tbl |>
    dplyr::select(tidyselect::all_of(hill_cols)) |>
    dplyr::collect()
  if (assay_exists) {
    hill_df <- hill_df |>
      tidyr::nest(.by = "assay_id")
  } else {
    hill_df <- hill_df |>
      tidyr::nest() |>
      dplyr::mutate(assay_id = NA_integer_)
  }

  # If all needed input data are available, check for existing risk table
  if (risk_name %in% tables) {
    if (!overwrite) {
      stop("GeoTox connection already has a '", risk_name, "' table. ",
           "Use `overwrite = TRUE` to replace it.", call. = FALSE)
    }
    DBI::dbRemoveTable(con, risk_name)
  }

  # Create risk table
  if (GT$par$reset_seed) {
    seed <- .Random.seed
    tbls <- DBI::dbListTables(con)
  }
  risk_tbl <- dplyr::cross_join(
    # Have sample_tbl first so hill_df is copied into the database.
    # If hill_df is first, then the sample table is copied into memory.
    sample_tbl |>
      dplyr::select("id") |>
      dplyr::rename(sample_id = "id"),
    hill_df |> dplyr::select("assay_id"),
    copy = TRUE
  ) |>
    dplyr::select("assay_id", "sample_id") |>
    dplyr::arrange(.data$assay_id, .data$sample_id) |>
    dplyr::compute(
      name = risk_name,
      temporary = FALSE
    )
  if (GT$par$reset_seed) {
    rm_tbl <- setdiff(DBI::dbListTables(con), c(tbls, risk_name))
    if (length(rm_tbl) == 1) {
      DBI::dbRemoveTable(con, rm_tbl)
    }
    .Random.seed <<- seed
  }
  sql <- paste("ALTER TABLE", risk_name, "ADD COLUMN \"GCA.Eff\" DOUBLE")
  DBI::dbExecute(con, sql)
  sql <- paste("ALTER TABLE", risk_name, "ADD COLUMN \"IA.Eff\" DOUBLE")
  DBI::dbExecute(con, sql)
  sql <- paste("ALTER TABLE", risk_name, "ADD COLUMN \"GCA.HQ.10\" DOUBLE")
  DBI::dbExecute(con, sql)
  sql <- paste("ALTER TABLE", risk_name, "ADD COLUMN \"IA.HQ.10\" DOUBLE")
  DBI::dbExecute(con, sql)

  # Compute risk

  # Loop over location
  location_id <- dplyr::tbl(con, "location") |>
    dplyr::select("id") |>
    dplyr::arrange(.data$id) |>
    dplyr::pull()
  purrr::walk(location_id, \(location_id) {

    # Collect concentration data for all samples in the location
    conc_df <- conc_tbl |>
      dplyr::inner_join(
        sample_tbl |>
          dplyr::filter(location_id == .env$location_id) |>
          dplyr::select("id"),
        by = dplyr::join_by("sample_id" == "id")
      ) |>
      dplyr::filter(is.finite(.data$C_invitro) & .data$C_invitro > 0) |>
      dplyr::collect()

    if (nrow(conc_df) == 0) return(NULL)

    # Loop over assays
    update_df <- purrr::map2(
      hill_df$assay_id,
      hill_df$data,
      \(assay_id, hill_params) {
        risk <- .calc_risk(conc_df, hill_params, max_mult, fixed)
        if (is.null(risk)) return(NULL)
        dplyr::bind_cols(assay_id = assay_id, risk)
      }
    ) |>
      dplyr::bind_rows()

    update_table(
      con, risk_name, update_df, reset_seed = GT$par$reset_seed,
      by = update_by
    )
  })

  invisible(GT)
}

.calc_risk <- function(df, hill_params, max_mult, fixed) {

  interval <- c(-50,50)

  sample_id <- df |>
    dplyr::distinct(.data$sample_id) |>
    dplyr::pull() |>
    sort()

  resp_val <- pmax(abs(hill_params$resp_max), abs(hill_params$resp_min))

  tp <- lapply(sample_id, \(x) {
    out <- truncnorm::rtruncnorm(
      1,
      a    = 0,
      b    = resp_val * max_mult,
      mean = abs(hill_params$tp),
      sd   = if (fixed) 0 else hill_params$tp.sd
    )
    # Replace NAs with 0 (can occur when tp and tp.sd are 0)
    out[is.na(out)] <- 0
    out
  }) |>
    unlist()

  AC50 <- lapply(sample_id, \(x) {
    logAC50 <- truncnorm::rtruncnorm(
      1,
      a    = hill_params$logc_min - 2.0001,
      b    = hill_params$logc_max + 0.5001,
      mean = hill_params$logAC50,
      sd   = if (fixed) 0 else hill_params$logAC50.sd
    )
    10^logAC50
  }) |>
    unlist()

  hill_params <- dplyr::bind_cols(
    sample_id = rep(sample_id, each = nrow(hill_params)),
    substance_id = rep(hill_params$substance_id, times = length(sample_id)),
    tp = tp,
    AC50 = AC50
  )

  df <- df |>
    dplyr::inner_join(
      hill_params,
      by = c("sample_id", "substance_id")
    ) |>
    # Order for replication
    dplyr::arrange(
      .data$sample_id,
      match(.data$substance_id, hill_params$substance_id)
    )

  if (nrow(df) == 0) return(NULL)

  df <- df |>
    tidyr::nest(.by = sample_id) |>
    dplyr::mutate(
      data = purrr::map(.data$data, \(x) {

        mixture.result <- stats::optimize(
          obj_GCA,
          interval = interval,
          conc = x$C_invitro,
          max = x$tp,
          AC50 = x$AC50
        )
        GCA.eff <- exp(mixture.result$minimum)

        Emax_resp <- stats::optimize(
          obj_GCA,
          interval = interval,
          conc = x$C_invitro * 10^14,
          max= x$tp,
          AC50 = x$AC50
        )
        Emax <- exp(Emax_resp$minimum)

        IA.eff <- calc_independent_action(
          x$C_invitro, x$tp, x$AC50, Emax
        )

        E10 <- Emax * 0.1

        EC10.result <- stats::optimize(
          obj_ECx,
          interval = c(-1000,1000),
          resp = E10,
          conc = x$C_invitro,
          max = x$tp,
          AC50 = x$AC50
        )
        EC10.GCA <- EC10.result$minimum

        E10.by.chem <- x$tp * 0.1
        AC10 <- hill_conc(E10.by.chem, x$tp, x$AC50, 1)

        sCi <- sum(x$C_invitro)
        GCA.HQ.10 <- if (EC10.GCA > 0) sCi / EC10.GCA else NA_real_
        IA.HQ.10 <- sum(x$C_invitro / AC10)

        c(GCA.eff, IA.eff, GCA.HQ.10, IA.HQ.10)
      })
    )

  df <- dplyr::bind_cols(
    sample_id = df$sample_id,
    as.data.frame(do.call(rbind, df$data))
  )
  colnames(df)[-1] <- c("GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10")
  tibble::as_tibble(df)
}

# Define DuckDB storage locations to prevent information messages from being
# printed when a connection is opened. See `?duckdb::duckdb_storage` for details.
# duckdb.home for DuckDB v1.5.5, duckdb.extension_directory for earlier versions.
tempdir <- withr::local_tempdir()
op <- options(duckdb.home = tempdir, duckdb.extension_directory = tempdir)

withr::defer(options(op), teardown_env())

#' Fit 2- or 3-parameter Hill model
#'
#' Fit a 2-parameter (fixed slope) or 3-parameter (variable slope) Hill model to
#' concentration-response data.
#'
#' The input `x` data frame must contain columns specified by `conc` and `resp`
#' arguments representing the log10-transformed concentration and response,
#' respectively.
#'
#' Optional `assay` and `substance` identifiers can be named vectors that are
#' used to fit multiple substances and/or assays. For example, `assay = c(name =
#' "assay", "model")` would indicate that `x` contains both and assay name and
#' model. The `name = "assay"` part would rename the "assay" column in `x` to
#' "name" in the 'assay' table when aded with [add_hill_params()].
#'
#' Returned column 'tp' is the top asymptote and 'logAC50' is the 50% response
#' concentration. If the computation of the standard deviations of these two
#' parameters fails, then the standard deviation is set equal to the parameter
#' estimate and is indicated by the respective imputed flag being TRUE.
#'
#' @param x Data frame of dose response data.
#' @param conc Column name of base-10 log scaled concentration (default "logc").
#' @param resp Column name of response (default "resp").
#' @param fixed_slope Logical indicating whether to fit a 2-parameter (TRUE) or
#'   3-parameter (FALSE) Hill model (default TRUE).
#' @param assay Column name of assay identifier(s) (optional, default NULL).
#' @param substance Column name of substance identifier(s) (optional, default
#'   NULL).
#'
#' @returns A list with elements 'fit', 'assay', 'substance'. The 'fit' element
#'   is a data frame of fit parameters, while the 'assay' and 'substance'
#'   elements indicate the column names used for assay and substance
#'   identifiers, respectively.
#' @export
#' @seealso [add_hill_params()]
#'
#' @examples
#' hill_df <- tibble::tribble(
#'   ~assay, ~model, ~casn, ~logc, ~resp,
#'   "a1", "human", "00-00-1",    0,  10,
#'   "a1", "human", "00-00-1",    1,  20,
#'   "a1", "human", "00-00-1",    2,  80,
#'   "a1", "human", "00-00-1",    3, 100,
#'   "a1", "human", "00-00-2", -0.5,   5,
#'   "a1", "human", "00-00-2",  0.5,  20,
#'   "a1", "human", "00-00-2",  1.5,  55,
#'   "a1", "human", "00-00-2",  2.5,  60,
#'   "a2",   "rat", "00-00-1",   -1,   0,
#'   "a2",   "rat", "00-00-1",    0,  10,
#'   "a2",   "rat", "00-00-1",    1,  30,
#'   "a2",   "rat", "00-00-1",    2,  40
#' )
#'
#' # Fit 2-parameter Hill model
#' fit_hill(
#'   hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
#' )
#'
#' # Fit 3-parameter Hill model
#' fit_hill(hill_df, assay = "assay", substance = "casn", fixed_slope = FALSE)
fit_hill <- function(
    x, conc = "logc", resp = "resp", fixed_slope = TRUE, assay = NULL,
    substance = NULL
) {
  if (!inherits(x, "data.frame")) {
    stop("`x` must be a data.frame")
  }
  if (!all(c(conc, resp) %in% names(x))) {
    stop("`x` must contain columns named by 'conc' and 'resp' inputs")
  }
  if (!is.null(substance) && !all(substance %in% names(x))) {
    stop("`x` must contain column(s) named by 'substance' when not NULL")
  }
  if (!is.null(assay) && !all(assay %in% names(x))) {
    stop("`x` must contain column(s) named by 'assay' when not NULL")
  }

  assay_val <- unname(assay)
  substance_val <- unname(substance)
  cols <- c(assay_val, substance_val, "x" = conc, "y" = resp)

  fit <- x |>
    dplyr::select(tidyselect::all_of(cols)) |>
    tidyr::nest(
      .by = tidyselect::all_of(c(assay_val, substance_val)),
      .key = "df"
    ) |>
    dplyr::mutate(
      df = purrr::map(.data$df, \(x) .fit_hill(x, "x", "y", fixed_slope)),
      df = purrr::map(.data$df, \(x) .extract_hill_params(x))) |>
    tidyr::unnest("df") |>
    dplyr::arrange(dplyr::across(tidyselect::all_of(c(assay_val, substance_val))))

  list(
    fit = fit,
    assay = assay,
    substance = substance
  )
}

.fit_hill <- function(x, conc = "logc", resp = "resp", fixed_slope = TRUE) {

  log10_conc <- x[[conc]]
  resp <- x[[resp]]

  # Compute initial values
  resp_medians <- tapply(resp, log10_conc, stats::median)
  resp_mad <- stats::mad(resp)

  resp_init <- resp_medians[which.max(abs(resp_medians))]
  conc_init <- as.numeric(names(resp_init)) - 0.5
  err_init <- ifelse(resp_mad > 0, log(resp_mad), .Machine$double.eps)

  if (fixed_slope) {
    par_init <- c(resp_init, conc_init, err_init)
  } else {
    par_init <- c(resp_init, conc_init, 1.2, err_init)
  }

  # Determine bounds
  resp_max <- max(resp)
  resp_min <- min(resp)
  resp_val <- max(abs(c(resp_max, resp_min)))
  log10_conc_min <- min(log10_conc)
  log10_conc_max <- max(log10_conc)

  bounds <- as.data.frame(rbind(
    c(   -1.2 * resp_val,       1.2 * resp_val), # top asymptote
    c(log10_conc_min - 2, log10_conc_max + 0.5), # log10(AC50)
    c(               0.3,                    8), # slope
    c(              -Inf,                  Inf)  # err
  ))
  colnames(bounds) <- c("lower", "upper")

  if (fixed_slope) {
    # remove slope bound
    bounds <- bounds[-3, ]
  }

  # Fit data
  fit <- stats::optim(
    par_init,
    fn = obj_hill,
    method = "L-BFGS-B",
    log10_conc = log10_conc,
    resp = resp,
    lower = bounds$lower,
    upper = bounds$upper,
    hessian = TRUE,
    control = list(
      fnscale = -1,
      maxit = 10000
    )
  )

  sds <- suppressWarnings(sqrt(diag(solve(-fit$hessian))))
  if (fixed_slope) {
    par <- c(fit$par[1:2], 1, fit$par[3])
    sds <- c(sds[1:2], 0, sds[3])
  } else {
    par <- fit$par
  }

  # Return results
  out <- list(
    par = par,
    sds = sds,
    val = fit$value,
    convergence = fit$convergence,
    AIC = 2 * length(fit$par) - 2 * fit$value,
    logc_max = log10_conc_max,
    logc_min = log10_conc_min,
    resp_max = resp_max,
    resp_min = resp_min
  )
  names(out$par) <- names(out$sds) <- c("tp", "logAC50", "slope", "t-error")

  out
}

.extract_hill_params <- function(fit) {

  # Extract parameters
  cols <- c("tp" = "par.tp", "tp.sd" = "sds.tp",
            "logAC50" = "par.logAC50", "logAC50.sd" = "sds.logAC50",
            "slope" = "par.slope", "slope.sd" = "sds.slope",
            "logc_min", "logc_max", "resp_min", "resp_max", "AIC")
  params <- tibble::as_tibble(t(unlist(fit))) |>
    dplyr::select(tidyselect::all_of(cols))

  # Impute sd NAs with mean for tp and logAC50
  params |>
    dplyr::mutate(
      tp.sd.imputed = is.na(.data$tp.sd),
      tp.sd = dplyr::if_else(.data$tp.sd.imputed,
                             .data$tp,
                             .data$tp.sd),
      logAC50.sd.imputed = is.na(.data$logAC50.sd),
      logAC50.sd = dplyr::if_else(.data$logAC50.sd.imputed,
                                  .data$logAC50,
                                  .data$logAC50.sd))
}

test_that("bad inputs", {
  # Input should be a data frame
  expect_error(
    fit_hill(c()),
    "`x` must be a data.frame"
  )
  # Expected required column names
  expect_error(
    fit_hill(data.frame(x = 0, y = 0)),
    "`x` must contain columns named by 'conc' and 'resp' inputs"
  )
  # Expected optional column names
  df <- data.frame(logc = c(-1, 1), resp = c(100, 0))
  expect_error(
    fit_hill(df, substance = "missing"),
    "`x` must contain column\\(s\\) named by 'substance' when not NULL"
  )
  expect_error(
    fit_hill(df, assay = "missing"),
    "`x` must contain column\\(s\\) named by 'assay' when not NULL"
  )
})

test_that("output column names", {
  df <- data.frame(
    logc = -3:3,
    resp = 5 / (1 + 10^(1.2 * (0.4 - rep(-3:3, each = 3)))) + rnorm(21)
  )
  colnames <- c("tp", "tp.sd", "logAC50", "logAC50.sd", "slope", "slope.sd",
                "logc_min", "logc_max", "resp_min", "resp_max",
                "AIC", "tp.sd.imputed", "logAC50.sd.imputed")

  out <- fit_hill(df)
  expect_equal(names(out$fit), colnames)
  expect_null(out$substance)
  expect_null(out$assay)

  df <- rbind(df, df)
  df$substance_col <- rep(c("A", "B"), each = 21)

  out <- fit_hill(df, substance = "substance_col")
  expect_equal(names(out$fit), c("substance_col", colnames))
  expect_equal(out$substance, "substance_col")
  expect_null(out$assay)

  df <- rbind(df, df)
  df$assay_col <- rep(c("a", "b"), each = 42)

  out <- fit_hill(df, substance = "substance_col", assay = "assay_col")
  expect_equal(names(out$fit), c("assay_col", "substance_col", colnames))
  expect_equal(out$substance, "substance_col")
  expect_equal(out$assay, "assay_col")
})

test_that("internal", {

  x <- data.frame(logc = c(-1, 0, 1), resp = c(10, 5, 0))

  out <- .fit_hill(x)

  expect_equal(as.numeric(out$par["slope"]),
               1)

  out <- .fit_hill(x, fixed_slope = FALSE)

  expect_true(as.numeric(out$par["slope"]) > 1)

  res <- .extract_hill_params(out)

  colnames <- c("tp", "tp.sd", "logAC50", "logAC50.sd", "slope", "slope.sd",
                "logc_min", "logc_max", "resp_min", "resp_max",
                "AIC", "tp.sd.imputed", "logAC50.sd.imputed")

  expect_equal(names(res), colnames)
})

#' GeoTox S3 object
#'
#' Create a GeoTox object and connect to the underlying database.
#'
#' The `dbname` will point to a DuckDB database file. If the file does not
#' already exist, a new database will be created. Additional arguments passed
#' via `...` will be forwarded to [duckdb::duckdb()].
#'
#' The `reset_seed` parameter is necessary for replicating results from the
#' previous GeoTox implementation. Some database functions create temporary
#' tables where a random string is appended to the table name. This advances the
#' `.Random.seed` state and causes functions that use randomness to produce
#' different results between the previous and current implementations. Setting
#' `reset_seed = TRUE` will reset the user's `.Random.seed` to the value it had
#' before certain database operations.
#'
#' Various parameters will be stored in the 'par' table. These parameters will
#' also be loaded into the GeoTox object as a list in `GT$par`. The value for
#' `reset_seed` can only be set on initial GeoTox object creation and will be
#' loaded from the 'par' table for existing databases.
#'
#' @param dbname Database file name. Default is a temporary file.
#' @param reset_seed Logical indicating whether to reset the user's global
#'   `.Random.seed` after certain database operations (default `FALSE`).
#' @param ... Additional arguments passed to [duckdb::duckdb()].
#'
#' @returns For `GeoTox()`, a GeoTox S3 object. For `get_con()`, a database
#'   connection.
#' @export
#'
#' @examples
#' # Create a GeoTox object
#' GT <- GeoTox()
#'
#' # Open a connection to GeoTox database
#' con <- get_con(GT)
#'
#' # List database tables
#' DBI::dbListTables(con)
#'
#' # Look at the GT parameters
#' dplyr::tbl(con, "par") |> dplyr::collect()
#' str(GT$par)
#'
#' # Clean up example
#' DBI::dbDisconnect(con)
#' file.remove(GT$db_info$dbdir)
GeoTox <- function(
    dbname = tempfile(fileext = ".duckdb"), reset_seed = FALSE, ...
) {
  db_info <- list(dbdir = dbname, ...)
  GT <- structure(
    list(db_info = db_info),
    class = "GeoTox"
  )
  con <- get_con(GT)
  on.exit(DBI::dbDisconnect(con))
  if (!DBI::dbExistsTable(con, "par")) {
    sql <- "CREATE TABLE par (name TEXT, value TEXT, idx INTEGER)"
    DBI::dbExecute(con, sql)
    set_par(con, "reset_seed", reset_seed)
  }
  GT$par <- get_par(con)
  GT
}

#' @param GT GeoTox object.
#' @export
#' @rdname GeoTox
get_con <- function(GT) {
  drv <- do.call(duckdb::duckdb, GT$db_info)
  DBI::dbConnect(drv)
}

#' @export
print.GeoTox <- function(x, ...) {
  con <- get_con(x)
  on.exit(DBI::dbDisconnect(con))

  tables <- DBI::dbListTables(con)
  get_n <- function(table_name) {
    if (table_name %in% tables) {
      dplyr::tbl(con, table_name) |> dplyr::count() |> dplyr::pull("n")
    } else {
      "NA"
    }
  }
  if ("sample" %in% tables) {
    n_sample <- dplyr::tbl(con, "sample") |>
      dplyr::count(.data$location_id) |>
      dplyr::pull("n") |>
      range()
    if (n_sample[1] == n_sample[2]) {
      n_sample <- n_sample[1]
    } else {
      n_sample <- paste0("[", paste(n_sample, collapse = ", "), "]")
    }
  } else {
    n_sample <- NA
  }
  if ("concentration" %in% tables) {
    conc_cols <- setdiff(
      colnames(dplyr::tbl(con, "concentration")),
      c("id", "sample_id", "substance_id", "route_id")
    )
  } else {
    conc_cols <- "NA"
  }
  if ("risk" %in% tables) {
    risk_cols <- setdiff(
      colnames(dplyr::tbl(con, "risk")),
      c("assay_id", "sample_id")
    )
  } else {
    risk_cols <- "NA"
  }
  sens_tbls <- grep("^risk_sensitivity_", tables, value = TRUE)
  if (length(sens_tbls) > 0) {
    sens_tbls <- sub("^risk_sensitivity_", "", sens_tbls)
  } else {
    sens_tbls <- "NA"
  }

  cat("GeoTox object\n")
  cat("Database info:\n")
  purrr::iwalk(x$db_info, \(value, name) {
    if (!is.atomic(value)) value <- class(value)
    cat("  ", name, ": ", value, "\n", sep = "")
  })
  cat("Reset seed: ", x$par$reset_seed, "\n", sep = "")
  cat("Assays: ", get_n("assay"), "\n", sep = "")
  cat("Substances: ", get_n("substance"), "\n", sep = "")
  cat("Locations: ", get_n("location"), "\n", sep = "")
  cat("Population: ", n_sample, "\n", sep = "")
  cat("Concentrations: ", paste(conc_cols, collapse = ", "), "\n", sep = "")
  cat("Risk: ", paste(risk_cols, collapse = ", "), "\n", sep = "")
  cat("Sensitivity: ", paste(sens_tbls, collapse = ", "), "\n", sep = "")
}

test_that("print", {
  # Extra input, "config", for duckdb driver
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"), config = list())
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Empty object
  expect_no_error(capture_output(print(GT)))

  # Range of sample sizes, no info for other tables
  sample_df <- data.frame(FIPS = c("00001", "00001", "00002"))
  conc_df <- data.frame(id = 1, sample_id = 1, substance_id = 1, route_id = 1)
  risk_df <- data.frame(assay_id = 1, sample_id = 1)
  GT |> set_sample(sample_df)
  DBI::dbWriteTable(con, "concentration", conc_df)
  DBI::dbWriteTable(con, "risk", risk_df)
  expect_no_error(capture_output(print(GT)))

  # Single sample size, info for other tables
  sample_df <- data.frame(FIPS = c("00001", "00002"))
  conc_df <- data.frame(
    id = 1, sample_id = 1, substance_id = 1, route_id = 1, C_ext = 1
  )
  risk_df <- data.frame(assay_id = 1, sample_id = 1, GCA.Eff = 1)
  GT |> set_sample(sample_df, overwrite = TRUE)
  DBI::dbWriteTable(con, "concentration", conc_df, overwrite = TRUE)
  DBI::dbWriteTable(con, "risk", risk_df, overwrite = TRUE)
  DBI::dbWriteTable(con, "risk_sensitivity_age", risk_df)
  expect_no_error(capture_output(print(GT)))
})

test_that("missing table", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  expect_error(
    simulate_obesity(GT),
    "No 'obesity' table found in the GeoTox connection."
  )
})

test_that("duplicate location", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))
  bad_df <- data.frame(
    location_id = 1,
    OBESITY_CrudePrev = c(10, 20),
    OBESITY_SD = c(5, 10)
  )
  DBI::dbWriteTable(con, "obesity", bad_df)
  expect_error(
    simulate_obesity(GT),
    "The 'obesity' table must not contain duplicated location values."
  )
})

test_that("simulate weight", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  obesity_df <- data.frame(
    location_id = c(1, 2, 3),
    OBESITY_CrudePrev = c(10, NA_real_, 30),
    OBESITY_SD = c(5, 10, 15)
  )
  DBI::dbWriteTable(con, "obesity", obesity_df)

  # No sample table
  expect_no_error(
    simulate_obesity(GT, n = 10)
  )

  # Sample table exists with age column
  expect_error(
    simulate_obesity(GT),
    "GeoTox connection already has a 'sample' table with a 'weight' column."
  )

  # Sample table exists without weight column, has an unmatched location_id
  sample_df <- data.frame(
    id = 1:6,
    location_id = c(1, 1, 2, 2, 2, 4)
  )
  DBI::dbWriteTable(con, "sample", sample_df, overwrite = TRUE)
  expect_no_error(
    simulate_obesity(GT)
  )

  # Custom inputs
  colnames(obesity_df)[2:3] <- c("prev", "sd")
  DBI::dbWriteTable(con, "obesity", obesity_df, overwrite = TRUE)
  GT <- simulate_obesity(GT, obes_prev = "prev", obes_sd = "sd", overwrite = TRUE)

  # Stored GT params
  simulate_obesity(GT, overwrite = TRUE)

  # Check output values
  sample_tbl <- dplyr::tbl(con, "sample") |> dplyr::collect()
  x <- split(sample_tbl$weight, sample_tbl$location_id)
  expect_length(x[[1]], 2)
  expect_length(x[[2]], 3)
  expect_length(x[[3]], 1)
  expect_true(all(x[[1]] %in% c("Normal", "Obese")))
  expect_true(all(is.na(x[[2]])))
  expect_true(is.na(x[[3]]))
})

test_that("missing tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    simulate_exposure(GT),
    "No 'exposure' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "exposure", data.frame(a = 1))
  expect_error(
    simulate_exposure(GT, sensitivity = TRUE),
    "No 'concentration_sensitivity' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "concentration_sensitivity", data.frame(a = 1))
  expect_error(
    simulate_exposure(GT, sensitivity = TRUE),
    "No 'sample' table found in the GeoTox connection."
  )
})

test_that("sample table already exists", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup needed tables
  sample_df <- tibble::tibble(
    FIPS = c(rep("00001", 2), rep("00002", 3))
  )
  exposure_df <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mean, ~sd,
    "00001", "50-00-0", "inhalation",  0, 2,
    "00001", "64-17-5", "inhalation", 20, 5,
    "00002", "50-00-0", "inhalation", 15, 3,
    "00002", "64-17-5", "inhalation", 25, 4,
    "00003", "50-00-0", "inhalation", 12, 2
  )
  GT |> set_sample(sample_df) |> add_exposure(exposure_df)

  # No concentration table
  expect_no_error(simulate_exposure(GT))

  # Overwrite
  expect_error(
    simulate_exposure(GT),
    "GeoTox connection already has a 'concentration' table with a 'C_ext' column."
  )

  # Concentration table w/o C_ext
  DBI::dbExecute(con, "ALTER TABLE concentration DROP COLUMN C_ext")
  expect_no_error(simulate_exposure(GT))

  # Output
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(nrow(conc_tbl), 10)
  expect_true("C_ext" %in% colnames(conc_tbl))

  # Sensitivity
  conc_sens_tbl <- dplyr::tbl(con, "concentration") |>
    dplyr::select(id, sample_id, substance_id, route_id, C_ext) |>
    dplyr::compute(
      name = "concentration_sensitivity",
      overwrite = TRUE,
      temporary = FALSE
    )
  expect_no_error(simulate_exposure(GT, sensitivity = TRUE))
})

test_that("sample table doesn't exist", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup needed tables
  exposure_df <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mean, ~sd,
    "00001", "50-00-0", "inhalation",  0, 2,
    "00001", "64-17-5", "inhalation", 20, 5,
    "00002", "50-00-0", "inhalation", 15, 3,
    "00002", "64-17-5", "inhalation", 25, 4,
    "00003", "50-00-0", "inhalation", 12, 2
  )
  GT |> add_exposure(exposure_df)

  # Simulate
  expect_no_error(simulate_exposure(GT, n = 4))

  # Output
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(nrow(conc_tbl), 24) # 3 locations * 4 samples * 2 substances
  expect_true("C_ext" %in% colnames(conc_tbl))
  sample_tbl <- dplyr::tbl(con, "sample") |> dplyr::collect()
  expect_equal(nrow(sample_tbl), 12) # 3 locations * 4 samples
})

test_that("other column names", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup needed tables
  sample_df <- tibble::tibble(
    FIPS = c(rep("00001", 2), rep("00002", 3))
  )
  exposure_df <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mu, ~sigma,
    "00001", "50-00-0", "inhalation", 10, 2,
    "00001", "64-17-5", "inhalation", 20, 5,
    "00002", "50-00-0", "inhalation", 15, 3,
    "00002", "64-17-5", "inhalation", 25, 4,
    "00003", "50-00-0", "inhalation", 12, 2
  )
  GT |> set_sample(sample_df) |> add_exposure(exposure_df)

  # Input column names
  expect_no_error(
    GT <- simulate_exposure(GT, expos_mean = "mu", expos_sd = "sigma")
  )

  # Using stored GT$par values
  expect_no_error(
    simulate_exposure(GT, overwrite = TRUE)
  )
})

test_that("data order", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup needed tables
  exposure_df_1 <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mean, ~sd,
    "00001", "50-00-0", "inhalation",  0, 0,
    "00001", "64-17-5", "inhalation", 20, 0
  )
  # Reverse substance order for this location
  exposure_df_2 <- tibble::tribble(
    ~FIPS, ~casn, ~route, ~mean, ~sd,
    "00002", "64-17-5", "inhalation", 25, 0,
    "00002", "50-00-0", "inhalation", 15, 0
  )
  GT |> add_exposure(exposure_df_1) |> add_exposure(exposure_df_2)

  # Simulate
  expect_no_error(simulate_exposure(GT, n = 1))

  # Output
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(nrow(conc_tbl), 4) # 2 locations * 2 substances
  expect_true("C_ext" %in% colnames(conc_tbl))
  sample_tbl <- dplyr::tbl(con, "sample") |> dplyr::collect()
  expect_equal(nrow(sample_tbl), 2) # 2 locations * 1 sample
  # Check order of C_ext values, should be ordered by location then substance
  expect_equal(
    conc_tbl |> dplyr::arrange(id) |> dplyr::pull(C_ext),
    c(0, 20, 15, 25)
  )
})

test_that("set sample", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Normal inputs
  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~weight,
    10000, 25, "Normal"
  )
  set_sample(GT, sample_df)
  sample_tbl <- DBI::dbReadTable(con, "sample")
  expect_equal(sample_tbl$age, 25)
  location_tbl <- DBI::dbReadTable(con, "location")
  expect_equal(nrow(location_tbl), 1)
  expect_equal(colnames(location_tbl), c("id", "FIPS"))

  # Concentration and risk tables exit
  DBI::dbWriteTable(con, "concentration", tibble::tibble(id = 1))
  DBI::dbWriteTable(con, "risk", tibble::tibble(id = 1))
  expect_error(
    set_sample(GT, sample_df),
    "The 'concentration' or 'risk' tables already exist."
  )
  expect_no_error(
    set_sample(GT, sample_df, overwrite = TRUE)
  )
  expect_true(!DBI::dbExistsTable(con, "concentration"))
  expect_true(!DBI::dbExistsTable(con, "risk"))

  # Different location
  sample_df <- tibble::tribble(
    ~FIPS, ~zip, ~age,
    20000, "12345", 30
  )
  # 'concentration' and 'risk' tables don't exist, so overwrite isn't needed
  set_sample(GT, sample_df, location = c("FIPS", "zip"))
  sample_tbl <- DBI::dbReadTable(con, "sample")
  expect_equal(sample_tbl$age, 30)
  location_tbl <- DBI::dbReadTable(con, "location")
  expect_equal(nrow(location_tbl), 2)
  expect_equal(colnames(location_tbl), c("id", "FIPS", "zip"))
  expect_true(is.na(location_tbl$zip[[1]]))
})

test_that("get risk values", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup tables
  risk_df <- tibble::tribble(
    ~assay_id, ~sample_id, ~GCA.Eff,
    1, 1, 10,
    1, 2, NA_real_,
    1, 3, 30,
    1, 4, 40,
    2, 2, 60
  )
  assay_df <- tibble::tribble(
    ~id, ~name,
    1, "Assay A",
    2, "Assay B"
  )

  # Missing table
  expect_error(
    get_risk_values(GT),
    "No 'risk' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "risk", risk_df)

  # No assay table
  expect_warning(
    out <- get_risk_values(GT, assay = c(name = "Assay A")),
    "Ignoring assay parameter."
  )
  expect_equal(
    out,
    risk_df |> dplyr::arrange(sample_id) |> dplyr::pull(GCA.Eff)
  )

  # With assay table
  DBI::dbWriteTable(con, "assay", assay_df)
  expect_error(
    get_risk_values(GT, assay = "Assay A"),
    "A single assay must be specified with the 'assay' input as a named vector"
  )
  expect_error(
    get_risk_values(GT, assay = c(bad_name = "Assay A")),
    "Column 'bad_name' not found in assay table."
  )
  expect_error(
    get_risk_values(GT, assay = c(name = "Missing Assay")),
    "No results found for assay 'Missing Assay'."
  )
  out <- get_risk_values(GT, assay = c(name = "Assay A"))
  expect_equal(out, c(10, NA_real_, 30, 40))
  out <- get_risk_values(GT, assay = c(name = "Assay B"))
  expect_equal(out, 60)
  out <- get_risk_values(GT, assay = c(id = 2))
  expect_equal(out, 60)
})

test_that("simulate population", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  age_df <- data.frame(
    FIPS = rep(c("00001", "00002"), each = 19),
    AGEGRP = rep(0:18, times = 2),
    TOT_POP = 0
  )
  # location 1, age group 40-44
  age_df$TOT_POP[c(1, 10)]  = 100
  # location 2, age group 5-9 and 50-54
  age_df$TOT_POP[c(3, 12) + 19] = 100
  age_df$TOT_POP[20] <- 200

  obesity_df <- data.frame(
    FIPS = c("00001", "00002"),
    OBESITY_CrudePrev = c(10, 20),
    OBESITY_SD = c(5, 10)
  )

  exposure_df <- data.frame(
    FIPS = c("00001", "00002"),
    casn = c("50-00-0", "50-00-0"),
    route = "inhalation",
    mean = c(10, 20),
    sd = c(2, 4)
  )

  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight, ~css,
    "50-00-0",  0, 44, "Normal",  1,
    "50-00-0", 45, 89, "Normal",  2,
    "50-00-0",  0, 44, "Obese",  11,
    "50-00-0", 45, 89, "Obese",  12
  )

  expect_no_error(
    GT |>
      set_simulated_css(css_df) |>
      add_exposure_rate_params() |>
      simulate_population(
        age = age_df,
        obesity = obesity_df,
        exposure = exposure_df,
        n = 3
      )
  )

  # Add another exposure substance column
  exposure_df$chnm <- c("chem1", "chem2")
  DBI::dbRemoveTable(con, "exposure")
  expect_no_error(
    simulate_population(
      GT,
      exposure = exposure_df,
      substance = c("casn", "chnm"),
      simulate_rate = FALSE,
      overwrite = TRUE
    )
  )

  # Check results
  sample_df <- dplyr::tbl(con, "sample") |> dplyr::collect()
  expect_equal(nrow(sample_df), 6)
  conc_df <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(nrow(conc_df), 6)
  expect_equal(
    colnames(conc_df),
    c("id", "sample_id", "substance_id", "route_id", "C_ext", "C_ss")
  )
})

test_that("get risk sensitivity", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup tables
  risk_df <- data.frame(
    sample_id = c(1, 2),
    GCA.Eff = c(10, NA_real_)
  )
  risk_sensitivity_C_ext <- data.frame(
    sample_id = c(1, 2),
    GCA.Eff = c(20)
  )
  risk_sensitivity_css_params <- data.frame(
    sample_id = c(1, 2),
    GCA.Eff = c(30, 40)
  )
  risk_sensitivity_weight <- data.frame(
    sample_id = c(1, 2),
    GCA.Eff = c(50, 60)
  )
  risk_sensitivity_age <- data.frame(
    sample_id = c(1, 2),
    GCA.Eff = c(80)
  )
  risk_sensitivity_fit_params <- data.frame(
    sample_id = c(1, 2),
    GCA.Eff = c(90)
  )
  DBI::dbWriteTable(con, "risk", risk_df)
  DBI::dbWriteTable(con, "risk_sensitivity_C_ext", risk_sensitivity_C_ext)
  DBI::dbWriteTable(con, "risk_sensitivity_css_params", risk_sensitivity_css_params)
  DBI::dbWriteTable(con, "risk_sensitivity_weight", risk_sensitivity_weight)
  DBI::dbWriteTable(con, "risk_sensitivity_age", risk_sensitivity_age)
  DBI::dbWriteTable(con, "risk_sensitivity_fit_params", risk_sensitivity_fit_params)

  out <- get_risk_sensitivity(GT)
  expect_equal(
    out,
    tibble::tribble(
      ~C_ext, ~css_params, ~weight, ~age, ~fit_params, ~baseline,
      20, 30, 50, 80, 90, 10,
      20, 40, 60, 80, 90, NA_real_
    )
  )
})

test_that("missing tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    calc_internal_dose(GT),
    "No 'concentration' table found in the GeoTox connection."
  )
  expect_error(
    calc_internal_dose(GT, sensitivity = TRUE),
    "No 'concentration_sensitivity' table found in the GeoTox connection."
  )
  df <- tibble::tribble(
    ~id, ~sample_id, ~route_id, ~C_ext,
    1, 1, 1, 0.5
  )
  write_table(con, "concentration", df)
  write_table(con, "concentration_sensitivity", df)
  expect_error(
    calc_internal_dose(GT),
    "No 'exposure_rate' table found in the GeoTox connection."
  )
  expect_error(
    calc_internal_dose(GT, sensitivity = TRUE),
    "No 'exposure_rate_sensitivity' table found in the GeoTox connection."
  )
})

test_that("calculate internal dose", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Don't match rows to make sure join is working
  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~route_id, ~C_ext,
    1, 1, 1, 0.5,
    2, 1, 2, 0.3,
    3, 2, 1, 0.2
  )
  rate_df <- tibble::tribble(
    ~sample_id, ~route_id, ~rate,
    2, 1, 20,
    1, 2, 5,
    1, 1, 10
  )
  write_table(con, "concentration", conc_df)
  write_table(con, "exposure_rate", rate_df)

  # Call function
  calc_internal_dose(GT)

  # Check results
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(conc_tbl$D_int, c(0.5 * 10, 0.3 * 5, 0.2 * 20))

  # Overwrite error
  expect_error(
    calc_internal_dose(GT),
    "GeoTox connection already has a 'D_int' column in the 'concentration' table."
  )

  # Modify input data and overwrite results
  sql <- "UPDATE concentration SET C_ext = C_ext + 1"
  invisible(DBI::dbExecute(con, sql))
  calc_internal_dose(GT, overwrite = TRUE)

  # Check results
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(conc_tbl$D_int, c(1.5 * 10, 1.3 * 5, 1.2 * 20))
})

test_that("missing table", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  expect_error(
    simulate_age(GT),
    "No 'age' table found in the GeoTox connection."
  )
})

test_that("incorrect nrows", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))
  bad_df <- data.frame(
    location_id = 1,
    AGEGRP = 0:9,
    TOT_POP = 100
  )
  DBI::dbWriteTable(con, "age", bad_df)
  expect_error(
    simulate_age(GT),
    "The 'age' table must contain 19 rows for each location."
  )
})

test_that("simulate age", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  age_df <- data.frame(
    location_id = rep(1:2, each = 19),
    AGEGRP = rep(0:18, times = 2),
    TOT_POP = 0
  )
  # location 1, age group 40-44
  age_df$TOT_POP[c(1, 10)]  = 100
  # location 2, age groups 50-59
  age_df$TOT_POP[c(1, 12, 13) + 19] = c(200, 100, 100)
  DBI::dbWriteTable(con, "age", age_df)

  # No sample table
  expect_no_error(
    simulate_age(GT, n = 5)
  )

  # Sample table exists with age column
  expect_error(
    simulate_age(GT),
    "GeoTox connection already has a 'sample' table with an 'age' column."
  )

  # Sample table exists without age column, has an unmatched location_id (id 3)
  sample_df <- data.frame(
    id = 1:6,
    location_id = c(1, 1, 2, 2, 2, 3)
  )
  DBI::dbWriteTable(con, "sample", sample_df, overwrite = TRUE)
  # And an extra location_id in the age table (id 4)
  extra_location <- data.frame(
    location_id = 4,
    AGEGRP = 0:18,
    TOT_POP = 0
  )
  append_table(con, "age", extra_location)
  expect_no_error(
    simulate_age(GT)
  )

  # Check output values
  sample_tbl <- dplyr::tbl(con, "sample") |> dplyr::collect()
  x <- split(sample_tbl$age, sample_tbl$location_id)
  expect_length(x[[1]], 2)
  expect_length(x[[2]], 3)
  expect_length(x[[3]], 1)
  expect_true(all(x[[1]] >= 40 & x[[1]] <= 44))
  expect_true(all(x[[2]] >= 50 & x[[2]] <= 59))
  expect_true(is.na(x[[3]]))
})

test_that("normal risk", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  risk_name <- "risk"
  conc_name <- "concentration"
  rate_name <- "exposure_rate"

  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~substance_id, ~route_id, ~C_ext, ~C_ss,
    1, 1, 1, 1, 0.5, 0.7,
    2, 1, 2, 2, 0.3, 0.5,
    3, 2, 1, 1, 0.4, 0.6,
    4, 2, 2, 2, 0.6, 0.8
  )
  rate_df <- tibble::tribble(
    ~sample_id, ~route_id, ~rate,
    1, 1, 5,
    1, 2, 10,
    2, 1, 15,
    2, 2, 20
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 1,
    2, 1
  )
  location_df <- tibble::tribble(
    ~id, ~FIPS,
    1, "00001"
  )
  hill_df <- tibble::tribble(
    ~substance_id,
    ~tp, ~tp.sd, ~logAC50, ~logAC50.sd, ~logc_min, ~logc_max, ~resp_min, ~resp_max,
    1, 0.5, 0.1, -5, 0.2, -8, -2, 0.0, 1.0,
    2, 0.6, 0.1, -6, 0.2, -9, -3, 0.0, 1.0
  )
  write_table(con, conc_name, conc_df)
  write_table(con, rate_name, rate_df)
  write_table(con, "sample", sample_df)
  write_table(con, "location", location_df)
  write_table(con, "hill_params", hill_df)

  set.seed(1234)
  GT <- GT |> calc_response()

  risk_tbl <- dplyr::tbl(con, risk_name) |> dplyr::collect()
  risk_cols <- c(
    "assay_id", "sample_id", "GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"
  )
  expect_true(all(risk_cols %in% colnames(risk_tbl)))

  risk_tbl <- risk_tbl |>
    dplyr::mutate(dplyr::across(tidyselect::where(is.numeric), \(x) {
      signif(x, 4)
    }))
  expect_snapshot(risk_tbl)
})

test_that("risk sensitivity", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  risk_name <- "risk_sensitivity"
  conc_name <- "concentration_sensitivity"
  rate_name <- "exposure_rate_sensitivity"

  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~substance_id, ~route_id, ~C_ext, ~C_ss,
    1, 1, 1, 1, 0.5, 0.7,
    2, 1, 2, 2, 0.3, 0.5,
    3, 2, 1, 1, 0.4, 0.6,
    4, 2, 2, 2, 0.6, 0.8
  )
  rate_df <- tibble::tribble(
    ~sample_id, ~route_id, ~rate,
    1, 1, 5,
    1, 2, 10,
    2, 1, 15,
    2, 2, 20
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 1,
    2, 1
  )
  location_df <- tibble::tribble(
    ~id, ~FIPS,
    1, "00001"
  )
  hill_df <- tibble::tribble(
    ~substance_id,
    ~tp, ~tp.sd, ~logAC50, ~logAC50.sd, ~logc_min, ~logc_max, ~resp_min, ~resp_max,
    1, 0.5, 0.1, -5, 0.2, -8, -2, 0.0, 1.0,
    2, 0.6, 0.1, -6, 0.2, -9, -3, 0.0, 1.0
  )
  write_table(con, conc_name, conc_df)
  write_table(con, rate_name, rate_df)
  write_table(con, "sample", sample_df)
  write_table(con, "location", location_df)
  write_table(con, "hill_params", hill_df)

  set.seed(1234)
  GT <- GT |> calc_response(risk_name = risk_name, max_mult = 2, fixed = TRUE)

  risk_tbl <- dplyr::tbl(con, risk_name) |> dplyr::collect()
  risk_cols <- c(
    "assay_id", "sample_id", "GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"
  )
  expect_true(all(risk_cols %in% colnames(risk_tbl)))

  risk_tbl <- risk_tbl |>
    dplyr::mutate(dplyr::across(tidyselect::where(is.numeric), \(x) {
      signif(x, 4)
    }))
  expect_snapshot(risk_tbl)
})

test_that("missing tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    sample_simulated_css(GT),
    "No 'simulated_css' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "simulated_css", data.frame(id = 1))
  expect_error(
    sample_simulated_css(GT),
    "No 'concentration' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "concentration", data.frame(id = 1))
  expect_error(
    sample_simulated_css(GT),
    "No 'sample' table found in the GeoTox connection."
  )
})

test_that("sample simulated css", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup tables
  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~weight,
    10000, 30, "Normal",
    10000, 50, "Normal",
    10000, 30,  "Obese",
    10000, 50,  "Obese"
  )
  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight, ~css,
    "00-00-1",  0, 49, "Normal",  1,
    "00-00-1",  0, 49, "Normal",  1, # a second copy to hit test case
    "00-00-1", 50, 99, "Normal",  2,
    "00-00-1",  0, 49,  "Obese", 11,
    "00-00-1", 50, 99,  "Obese", 12,
    "00-00-2",  0, 49, "Normal", 21,
    "00-00-2", 50, 99, "Normal", 22,
    "00-00-2",  0, 49,  "Obese", 31,
    "00-00-2", 50, 99,  "Obese", 32
  ) |>
    # Mix up order to test join, keep first row for substance order
    dplyr::arrange(c(1, sample(2:9)))
  GT |> set_simulated_css(css_df) |> set_sample(sample_df)
  conc_df <- dplyr::cross_join(
    dplyr::tbl(con, "sample") |> dplyr::select(sample_id = id),
    dplyr::tbl(con, "substance") |> dplyr::select(substance_id = id)
  ) |>
    dplyr::mutate(id = dplyr::row_number(), .before = 1) |>
    dplyr::compute(name = "concentration", temporary = FALSE)

  sample_simulated_css(GT)

  conc_tbl <- dplyr::tbl(con, "concentration") |>
    dplyr::collect() |> dplyr::arrange(id)
  expect_equal(conc_tbl$C_ss, c(1, 2, 11, 12, 21, 22, 31, 32))
})

test_that("extra inputs", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup tables
  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~weight, ~sign, ~no_match,
    10000, 30, "Normal", "+", "x",
    10000, 50, "Normal", "-", "x",
    10000, 30,  "Obese", "+", "x",
    10000, 50,  "Obese", "-", "x"
  )
  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight, ~css, ~sign, ~no_match,
    "00-00-1",  0, 49, "Normal",   1, "+", "no_match",
    "00-00-1", 50, 99, "Normal",   2, "+", "no_match",
    "00-00-1",  0, 49, "Normal",  -1, "-", "no_match",
    "00-00-1", 50, 99, "Normal",  -2, "-", "no_match",
    "00-00-1",  0, 49,  "Obese",  11, "+", "no_match",
    "00-00-1", 50, 99,  "Obese",  12, "+", "no_match",
    "00-00-1",  0, 49,  "Obese", -11, "-", "no_match",
    "00-00-1", 50, 99,  "Obese", -12, "-", "no_match",
    "00-00-2",  0, 49, "Normal",  21, "+", "no_match",
    "00-00-2", 50, 99, "Normal",  22, "+", "no_match",
    "00-00-2",  0, 49, "Normal", -21, "-", "no_match",
    "00-00-2", 50, 99, "Normal", -22, "-", "no_match",
    "00-00-2",  0, 49,  "Obese",  31, "+", "no_match",
    "00-00-2", 50, 99,  "Obese",  32, "+", "no_match",
    "00-00-2",  0, 49,  "Obese", -31, "-", "no_match",
    "00-00-2", 50, 99,  "Obese", -32, "-", "no_match",
  )
  GT |> set_simulated_css(css_df) |> set_sample(sample_df)
  conc_df <- dplyr::cross_join(
    dplyr::tbl(con, "sample") |> dplyr::select(sample_id = id),
    dplyr::tbl(con, "substance") |> dplyr::select(substance_id = id)
  ) |>
    dplyr::mutate(id = dplyr::row_number(), .before = 1) |>
    dplyr::compute(name = "concentration", temporary = FALSE)
  # Add extra location
  GT |> add_location(data.frame(FIPS = 20000))

  # Bad extra cols
  expect_error(
    sample_simulated_css(GT, css_extra_cols = c("no_match")),
    "No matching rows found in 'simulated_css'"
  )

  # Extra cols and substance order
  css_extra_cols <- c("sign")
  substance_order <- list(casn = c("00-00-2", "00-00-1"))
  expect_no_error(
    GT <- GT |>
      sample_simulated_css(
        css_extra_cols  = css_extra_cols,
        substance_order = substance_order
      )
  )

  # Run again using stored parameters
  expect_no_error(
    sample_simulated_css(GT)
  )

  # Check results
  expect_equal(GT$par$css_extra_cols, css_extra_cols)
  expect_equal(GT$par$substance_order_col, "casn")
  expect_equal(GT$par$substance_order, c("00-00-2", "00-00-1"))
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(conc_tbl$C_ss, c(1, -2, 11, -12, 21, -22, 31, -32))
})

test_that("set simulated css", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Missing css column
  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight,
    "00-00-1", 0, 49, "Normal",
  )
  expect_error(
    set_simulated_css(GT, css_df),
    "`df` must contain the following columns: age_lb, age_ub, weight, css."
  )

  # Okay
  css_df$css <- 1
  expect_no_error(
    set_simulated_css(GT, css_df)
  )

  # Overwrite
  expect_error(
    set_simulated_css(GT, css_df),
    "Table 'simulated_css' already exists."
  )

  # Other substance column
  names(css_df)[1] <- "chem_id"
  expect_no_error(
    set_simulated_css(GT, css_df, substance = "chem_id", overwrite = TRUE)
  )
})

test_that("missing tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    calc_risk(GT),
    "No 'concentration' table found in the GeoTox connection."
  )
  expect_error(
    calc_risk(GT, risk_name = "risk_sensitivity"),
    "No 'concentration_sensitivity' table found in the GeoTox connection."
  )
  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~substance_id, ~C_invitro,
    1, 1, 1, 0.5,
    2, 1, 2, 0.3
  )
  write_table(con, "concentration", conc_df)
  expect_error(
    calc_risk(GT),
    "No 'sample' table found in the GeoTox connection."
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 101,
    2, 102
  )
  write_table(con, "sample", sample_df)
  expect_error(
    calc_risk(GT),
    "No 'hill_params' table found in the GeoTox connection."
  )
})

test_that("no assay_id, reset_seed, extra loc", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"), reset_seed = TRUE)
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~substance_id, ~C_invitro,
    1, 1, 1, 0.5,
    2, 1, 2, 0.3,
    3, 2, 1, 0.4,
    4, 2, 2, 0.6
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 1,
    2, 1
  )
  location_df <- tibble::tribble(
    ~id, ~FIPS,
    1, "00001",
    2, "00002"
  )
  hill_df <- tibble::tribble(
    ~substance_id,
    ~tp, ~tp.sd, ~logAC50, ~logAC50.sd, ~logc_min, ~logc_max, ~resp_min, ~resp_max,
    1, 0.5, 0.1, -5, 0.2, -8, -2, 0.0, 1.0,
    2, 0.6, 0.1, -6, 0.2, -9, -3, 0.0, 1.0
  )
  write_table(con, "concentration", conc_df)
  write_table(con, "sample", sample_df)
  write_table(con, "location", location_df)
  write_table(con, "hill_params", hill_df)

  GT <- calc_risk(GT)

  risk_tbl <- dplyr::tbl(con, "risk") |> dplyr::collect()
  risk_cols <- c(
    "assay_id", "sample_id", "GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"
  )
  expect_true(all(risk_cols %in% colnames(risk_tbl)))
  expect_length(risk_tbl$assay_id, 2)
  expect_true(all(is.na(risk_tbl$assay_id)))
})

test_that("with assay_id, extra hill", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~substance_id, ~C_invitro,
    1, 1, 1, 0.5,
    2, 1, 2, 0.3,
    3, 2, 1, 0.4,
    4, 2, 2, 0.6
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 1,
    2, 1
  )
  location_df <- tibble::tribble(
    ~id, ~FIPS,
    1, "00001"
  )
  hill_df <- tibble::tribble(
    ~assay_id, ~substance_id,
    ~tp, ~tp.sd, ~logAC50, ~logAC50.sd, ~logc_min, ~logc_max, ~resp_min, ~resp_max,
    1, 1,  0.5, 0.1, -5, 0.2,  -8, -2, 0.0, 1.0,
    2, 2, -0.6, 0.1, -6, 0.2,  -9, -3, 0.0, 1.0,
    3, 3,  0.7, 0.1, -7, 0.2, -10, -4, 0.0, 1.0
  )
  write_table(con, "concentration", conc_df)
  write_table(con, "sample", sample_df)
  write_table(con, "location", location_df)
  write_table(con, "hill_params", hill_df)

  GT <- calc_risk(GT)

  risk_tbl <- dplyr::tbl(con, "risk") |> dplyr::collect()
  risk_cols <- c(
    "assay_id", "sample_id", "GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"
  )
  expect_true(all(risk_cols %in% colnames(risk_tbl)))
  expect_length(risk_tbl$assay_id, 6)
  expect_equal(sum(is.na(risk_tbl$GCA.Eff)), 2)
  expect_true(all(!is.na(risk_tbl$assay_id)))
})

test_that("overwrite", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~substance_id, ~C_invitro,
    1, 1, 1, 0.5
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 1
  )
  location_df <- tibble::tribble(
    ~id, ~FIPS,
    1, "00001"
  )
  hill_df <- tibble::tribble(
    ~substance_id,
    ~tp, ~tp.sd, ~logAC50, ~logAC50.sd, ~logc_min, ~logc_max, ~resp_min, ~resp_max,
    1, 0.5, 0.1, -5, 0.2, -8, -2, 0.0, 1.0
  )
  write_table(con, "concentration", conc_df)
  write_table(con, "sample", sample_df)
  write_table(con, "location", location_df)
  write_table(con, "hill_params", hill_df)

  GT <- calc_risk(GT)

  expect_error(
    calc_risk(GT),
    "GeoTox connection already has a 'risk' table."
  )

  set.seed(1234)
  GT <- calc_risk(GT, overwrite = TRUE)

  risk_tbl <- dplyr::tbl(con, "risk") |> dplyr::collect()
  expect_equal(nrow(risk_tbl), 1)

  risk_tbl <- risk_tbl |>
    dplyr::mutate(dplyr::across(tidyselect::where(is.numeric), \(x) {
      signif(x, 4)
    }))
  expect_snapshot(risk_tbl)
})

test_that("missing tables", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    set_fixed_css(GT),
    "No 'sample' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "sample", data.frame(id = 1, location_id = 1))

  expect_error(
    set_fixed_css(GT),
    "No 'simulated_css' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "simulated_css", data.frame(id = 1, location_id = 1))

  expect_error(
    set_fixed_css(GT),
    "No 'substance' table found in the GeoTox connection."
  )

  DBI::dbWriteTable(con, "substance", data.frame(id = 1, location_id = 1))

  expect_error(
    set_fixed_css(GT),
    "No 'concentration' table found in the GeoTox connection."
  )
})

test_that("set fixed css", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  set.seed(1234)

  # Setup tables
  sample_df <- tibble::tribble(
    ~FIPS, ~age, ~weight,
    10000, 30, "Normal",
    10000, 50, "Normal",
    10000, 30,  "Obese",
    10000, 50,  "Obese",
    20000, 60, "Normal",
    20000, 40, "Normal",
    20000, 60,  "Obese",
    20000, 40,  "Obese"
  )
  css_df <- tibble::tribble(
    ~casn, ~age_lb, ~age_ub, ~weight, ~css,
    "00-00-1",  0, 49, "Normal",  1,
    "00-00-1",  0, 49, "Normal",  1, # a second copy to hit test case
    "00-00-1", 50, 99, "Normal",  2,
    "00-00-1",  0, 49,  "Obese", 11,
    "00-00-1", 50, 99,  "Obese", 12,
    "00-00-2",  0, 49, "Normal", 21,
    "00-00-2", 50, 99, "Normal", 22,
    "00-00-2",  0, 49,  "Obese", 31,
    "00-00-2", 50, 99,  "Obese", 32
  )
  GT |> set_sample(sample_df) |> set_simulated_css(css_df)
  conc_df <- dplyr::cross_join(
    dplyr::tbl(con, "sample") |> dplyr::select(sample_id = id),
    dplyr::tbl(con, "substance") |> dplyr::select(substance_id = id)
  ) |>
    dplyr::arrange(substance_id, sample_id) |>
    dplyr::mutate(id = dplyr::row_number(), .before = 1) |>
    dplyr::compute(name = "concentration", temporary = FALSE)
  sample_simulated_css(GT)

  # Basic use
  set_fixed_css(GT)

  # Input substance_order
  GT <- GT |>
    set_fixed_css(substance_order = list(casn = c("00-00-2", "00-00-1")))
  expect_equal(GT$par$substance_order_col, "casn")
  expect_equal(GT$par$substance_order, c("00-00-2", "00-00-1"))

  # Using stored GT$par
  set_fixed_css(GT)

  # Check results
  fixed_css_tbl <- dplyr::tbl(con, "fixed_css") |>
    dplyr::collect() |> dplyr::arrange(id)
  expect_equal(
    fixed_css_tbl,
    tibble::tribble(
      ~id, ~sample_id, ~substance_id, ~age, ~weight, ~params, ~other,
       1, 1, 1,  1,    1,  1, 16.5,
       2, 1, 2, 26, 21.5, 21, 16.5,
       3, 2, 1,  7,    1,  1, 16.5,
       4, 2, 2, 27, 21.5, 21, 16.5,
       5, 3, 1,  1, 11.5,  1, 16.5,
       6, 3, 2, 26, 31.5, 21, 16.5,
       7, 4, 1,  7, 11.5,  1, 16.5,
       8, 4, 2, 27, 31.5, 21, 16.5,
       9, 5, 1,  7,    1,  2, 16.5,
      10, 5, 2, 27, 21.5, 22, 16.5,
      11, 6, 1,  1,    1,  2, 16.5,
      12, 6, 2, 26, 21.5, 22, 16.5,
      13, 7, 1,  7, 11.5,  2, 16.5,
      14, 7, 2, 27, 31.5, 22, 16.5,
      15, 8, 1,  1, 11.5,  2, 16.5,
      16, 8, 2, 26, 31.5, 22, 16.5
    )
  )
})

test_that("missing table", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  expect_error(
    calc_invitro_concentration(GT),
    "No 'concentration' table found in the GeoTox connection."
  )
  expect_error(
    calc_invitro_concentration(GT, sensitivity = TRUE),
    "No 'concentration_sensitivity' table found in the GeoTox connection."
  )
})

test_that("calculate in vitro concentration", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  conc_df <- tibble::tribble(
    ~id, ~sample_id, ~route_id, ~C_ss, ~D_int,
    1, 1, 1, 0.5, 2,
    2, 1, 2, 0.3, 3,
    3, 2, 1, 0.2, 4
  )
  write_table(con, "concentration", conc_df)

  # Call function
  GT <- calc_invitro_concentration(GT)

  # Check results
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(conc_tbl$C_invitro, conc_df$C_ss * conc_df$D_int)

  # Overwrite error
  expect_error(
    calc_invitro_concentration(GT),
    "GeoTox connection already has a 'C_invitro' column in the 'concentration' table. Set `overwrite = TRUE` to replace the existing values.\nWarning: Overwriting will not alter downstream values."
  )

  # Modify input data and overwrite results
  sql <- "UPDATE concentration SET C_ss = C_ss + 1"
  invisible(DBI::dbExecute(con, sql))
  GT <- calc_invitro_concentration(GT, overwrite = TRUE)

  # Check results
  conc_tbl <- dplyr::tbl(con, "concentration") |> dplyr::collect()
  expect_equal(conc_tbl$C_invitro, (conc_df$C_ss + 1) * conc_df$D_int)
})

test_that("get concentration mean", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup tables
  conc_df <- tibble::tribble(
    ~substance_id, ~route_id, ~sample_id, ~C_ext,
    1, 1, 1, 10,
    1, 1, 2, NA_real_,
    1, 1, 3, 30,
    1, 1, 4, 40,
    2, 1, 2, 60
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 1,
    2, 1,
    3, 1,
    4, 2
  )

  # Missing tables
  expect_error(
    get_concentration_mean(GT, "C_ext"),
    "No 'concentration' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "concentration", conc_df)
  expect_error(
    get_concentration_mean(GT, "C_ext"),
    "No 'sample' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "sample", sample_df)

  # Get output
  out <- get_concentration_mean(GT, "C_ext")

  # Compare output
  expect_equal(
    out,
    tibble::tribble(
      ~substance_id, ~route_id, ~location_id, ~mean,
      1, 1, 1, 20,
      1, 1, 2, 40,
      2, 1, 1, 60,
      2, 1, 2, NA_real_
    )
  )
})

test_that("get risk quantiles", {
  GT <- GeoTox(withr::local_tempfile(fileext = ".duckdb"))
  con <- get_con(GT)
  withr::defer(DBI::dbDisconnect(con))

  # Setup tables
  risk_df <- tibble::tribble(
    ~assay_id, ~sample_id, ~GCA.Eff,
    1, 1, 10,
    1, 2, NA_real_,
    1, 3, 30,
    1, 4, 40,
    2, 2, 60
  )
  sample_df <- tibble::tribble(
    ~id, ~location_id,
    1, 1,
    2, 1,
    3, 1,
    4, 2
  )

  # Missing tables
  expect_error(
    get_risk_quantiles(GT),
    "No 'risk' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "risk", risk_df)
  expect_error(
    get_risk_quantiles(GT),
    "No 'sample' table found in the GeoTox connection."
  )
  DBI::dbWriteTable(con, "sample", sample_df)

  # Get output
  out <- get_risk_quantiles(GT, quantiles = c(0.1, 0.5))

  # Compare output
  expect_equal(
    out,
    tibble::tribble(
      ~assay_id, ~location_id, ~quantile, ~value,
      1, 1, 0.1, 12,
      1, 1, 0.5, 20,
      1, 2, 0.1, 40,
      1, 2, 0.5, 40,
      2, 1, 0.1, 60,
      2, 1, 0.5, 60,
      2, 2, 0.1, NA_real_,
      2, 2, 0.5, NA_real_
    )
  )
})

---
title: "GeoTox Package Data"
vignette: >
  %\VignetteIndexEntry{GeoTox Package Data}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
format:
  html:
    toc: true
    toc-expand: true
execute:
  eval: false
---

The GeoTox package includes example data `geo_tox_data`. This article provides a description of the data and example code for how it was gathered.

:::{.callout-note}
The websites linked in this vignette were last accessed on March 20, 2026. The latest R package versions were used at that time.
:::

:::{.callout-warning}
The package data are being pulled from various sources and are connected using FIPS codes. FIPS codes are a simple way to connect data, but they can change. For example, Connecticut began the process of going from 8 legacy counties to 9 planning regions starting in 2022 and became effective in 2024.
:::

# Setup

Load packages and create an empty list to store data.

```{r}
#| label: setup
library(dplyr)
library(httk)
library(httr2)
library(purrr)
library(readr)
library(readxl)
library(sf)
library(stringr)
library(tibble)
library(tidyr)

geo_tox_data <- list()
```

# Chemicals

## External exposure

Download modeled exposure data from [AirToxScreen](https://www.epa.gov/AirToxScreen){target="_blank" rel="noopener noreferrer"}. Results from 2020 for a subset of chemicals in North Carolina counties are included in the GeoTox package data as `geo_tox_data$exposure`.

:::{.callout-note}
The [2020](https://www.epa.gov/AirToxScreen/2020-airtoxscreen-assessment-results){target="_blank" rel="noopener noreferrer"} version has census block level data, which is provided in several large files separated by region. The [2019](https://www.epa.gov/AirToxScreen/2019-airtoxscreen-assessment-results){target="_blank" rel="noopener noreferrer"} version has census tract level data, which is provided as a single file and is much smaller to download.
:::

```{r}
#| label: exposure
filename <- "Region4b_2020ATS_Exposure_Concentrations.xlsx"
tmp <- tempfile(filename)
download.file(
  paste0(
    "https://gaftp.epa.gov/rtrmodeling_public/AirToxScreen/2020/",
    "Exposure%20Concentrations/",
    filename
  ),
  tmp
)
exposure <- read_xlsx(tmp)

# Normalization function
min_max_norm = function(x) {
  min_x <- min(x, na.rm = TRUE)
  max_x <- max(x, na.rm = TRUE)
  if (min_x == max_x) {
    rep(0, length(x))
  } else {
    (x - min_x) / (max_x - min_x)
  }
}

geo_tox_data$exposure <- exposure |>
  # North Carolina counties
  filter(State == "NC", !grepl("0$", FIPS)) |>
  # Aggregate chemicals by county
  summarize(
    across(
      -c(State:Population), # Note: use "-c(State:Tract)" for 2019 data
      list(mean = mean, sd = sd),
      .names = "{col}|||{fn}"
    ),
    .by = FIPS
  ) |>
  pivot_longer(-FIPS) |>
  separate_wider_delim(name, "|||", names = c("chemical", "stat")) |>
  pivot_wider(names_from = stat) |>
  # Normalize concentrations
  mutate(norm = min_max_norm(mean), .by = chemical)
```

### Get chemical identifiers from CompTox Dashboard

The exposure data is labeled with chemical names, but names can vary across resources and a more reliable way to connect datasets is by using Chemical Abstracts Service Registry Number (CAS-RN) identifiers. Use the batch search feature of the [CompTox Dashboard](https://comptox.epa.gov/dashboard/batch-search){target="_blank" rel="noopener noreferrer"} to retrieve CAS-RN and PREFERRED_NAME for the chemicals in the exposure dataset.

```{r}
#| label: comptox
# Copy/paste the chemical names into the batch search field
# https://comptox.epa.gov/dashboard/batch-search
cat(geo_tox_data$exposure |> distinct(chemical) |> pull(), sep = "\n")
# Export results with "CAS-RN" identifiers as a csv file, then process in R

# Remove rows without clear chemical matches, investigate manually if desired
exposure_casrn <- read_csv("CCD-Batch-Search.csv") |>
  filter(DTXSID != "N/A", !grepl("WARNING", FOUND_BY))

# Update exposure data with CompTox Dashboard data
geo_tox_data$exposure <- geo_tox_data$exposure |>
  inner_join(exposure_casrn, by = join_by(chemical == INPUT)) |>
  select(FIPS, casn = CASRN, chnm = PREFERRED_NAME, mean, sd, norm)
```

## Active chemicals

Use the Integrated Chemical Environment ([ICE](https://ice.ntp.niehs.nih.gov/){target="_blank" rel="noopener noreferrer"})
curated high-throughput screening ([cHTS](https://ice.ntp.niehs.nih.gov/DATASETDESCRIPTION?section=cHTS){target="_blank" rel="noopener noreferrer"}) data to identify active chemicals for a given set of assays.

:::{.callout-note}
The following can be used to get a list of available assays for a given chemical set.

```{r}
#| label: available-assays
# Get all supported assays
help_text <- request("https://ice.ntp.niehs.nih.gov/api/v1/search/help") |>
  req_perform() |>
  resp_body_string()
supported_assays <- str_split_i(help_text, "Supported assay\\(s\\):", 2) |>
  str_split_1("\",\"") |>
  str_replace_all(c("\"" = "")) |>
  str_trim()

# Search for assays available for given chemids
chemids <- unique(geo_tox_data$exposure$casn)
resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/search") |>
  req_body_json(list(chemids = chemids, assays = supported_assays)) |>
  req_perform()
result <- resp |> resp_body_json() |> pluck("endPoints")
fields <- c("assay", "casrn", "dtxsid", "substanceName", "endpoint", "value")
df <- map(fields, \(x) map_chr(result, x)) |>
  set_names(fields) |>
  bind_cols()
available_assays <- df |> distinct(assay) |> pull() |> sort()
```
:::

```{r}
#| label: cHTS-hits
get_cHTS_hits <- function(assays = NULL, chemids = NULL) {
  if (is.null(assays) & is.null(chemids)) {
    stop("Must provide at least one of 'assays' or 'chemids'")
  }

  # Format query parameters
  req_params <- list()

  if (!is.null(assays)) {
    if (!is.list(assays)) assays <- as.list(assays)
    req_params$assays <- assays
  }

  if (!is.null(chemids)) {
    if (!is.list(chemids)) chemids <- as.list(chemids)
    req_params$chemids <- chemids
  }

  # Query ICE API
  resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/search") |>
    req_body_json(req_params) |>
    req_perform()

  if (resp$status_code != 200) {
    stop("Failed to retrieve data from ICE API")
  }

  # Return active chemicals
  result <- resp |> resp_body_json() |> pluck("endPoints")

  fields <- c("assay", "casrn", "dtxsid", "substanceName", "endpoint", "value")

  map(fields, \(x) map_chr(result, x)) |>
    set_names(fields) |>
    bind_cols() |>
    filter(endpoint == "Call", value == "Active") |>
    select(-c(endpoint, value)) |>
    distinct()
}

assays <- c(
  "APR_HepG2_p53Act_1hr",
  "APR_HepG2_p53Act_24hr",
  "APR_HepG2_p53Act_72hr",
  "ATG_p53_CIS",
  "TOX21_DT40_LUC",
  "TOX21_DT40_100_LUC",
  "TOX21_DT40_657_LUC",
  "TOX21_ELG1_LUC_Agonist",
  "TOX21_H2AX_HTRF_CHO_Agonist_ratio",
  "TOX21_p53_BLA_p1_ratio",
  "TOX21_p53_BLA_p2_ratio",
  "TOX21_p53_BLA_p3_ratio",
  "TOX21_p53_BLA_p4_ratio",
  "TOX21_p53_BLA_p5_ratio"
)

chemids <- unique(geo_tox_data$exposure$casn)

cHTS_hits <- get_cHTS_hits(assays = assays, chemids = chemids)
```

## Dose-response

Use the ICE API to retrieve dose-response data for selected assays and chemicals, which is included in the GeoTox package data as `geo_tox_data$dose_response`.

```{r}
#| label: dose-response
get_ICE_dose_resp <- function(assays = NULL, chemids = NULL) {
  if (is.null(assays) & is.null(chemids)) {
    stop("Must provide at least one of 'assays' or 'chemids'")
  }

  # Format query parameters
  req_params <- list()

  if (!is.null(assays)) {
    if (!is.list(assays)) assays <- as.list(assays)
    req_params$assays <- assays
  }

  if (!is.null(chemids)) {
    if (!is.list(chemids)) chemids <- as.list(chemids)
    req_params$chemids <- chemids
  }

  # Query ICE API
  resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/curves") |>
    req_body_json(req_params) |>
    req_perform()

  if (resp$status_code != 200) {
    stop("Failed to retrieve data from ICE API")
  }

  # Return dose-response data
  result <- resp |> resp_body_json() |> pluck("curves")

  map(result, function(x) {
    tibble(
      endp = x[["endpoint"]],
      casn = x[["casrn"]],
      chnm = x[["substance"]],
      call = x[["call"]],
      logc = map_dbl(x[["concentrationResponses"]], "concentration") |> log10(),
      resp = map_dbl(x[["concentrationResponses"]], "response")
    )
  }) |>
    bind_rows()
}

assays <- unique(cHTS_hits$assay)

chemids <- intersect(cHTS_hits$casrn, geo_tox_data$exposure$casn)

dose_response <- get_ICE_dose_resp(assays = assays, chemids = chemids)

# Only keep active calls for assay/chemical combinations
geo_tox_data$dose_response <- dose_response |>
  filter(call == "Active") |>
  select(-call)

# Update dose-response chemical names using CompTox Dashboard data
geo_tox_data$dose_response <- geo_tox_data$dose_response |>
  inner_join(exposure_casrn, by = join_by(casn == CASRN)) |>
  select(endp, casn, chnm = PREFERRED_NAME, logc, resp)
```

# Population

## Age

Download age data from the [U.S. Census Bureau](https://www.census.gov/){target="_blank" rel="noopener noreferrer"} by searching for "County Population by Characteristics". A subset of data for North Carolina from [2020](https://www.census.gov/programs-surveys/popest/technical-documentation/research/evaluation-estimates/2020-evaluation-estimates/2010s-county-detail.html){target="_blank" rel="noopener noreferrer"} is included in the GeoTox package data as `geo_tox_data$age`.

```{r}
#| label: age
# Data for North Carolina
url <- paste0(
  "https://www2.census.gov/programs-surveys/popest/datasets/",
  "2020-2024/counties/asrh/cc-est2024-alldata-37.csv"
)
age <- read_csv(url)

geo_tox_data$age <- age |>
  # 7/1/2020 population estimate
  filter(YEAR == 2) |>
  # Create FIPS
  mutate(FIPS = str_c(STATE, COUNTY)) |>
  # Keep selected columns
  select(FIPS, AGEGRP, TOT_POP)
```

## Obesity status

Search [CDC data](https://data.cdc.gov/){target="_blank" rel="noopener noreferrer"} for "places county data". Go to the desired dataset webpage, e.g. [2020 county data](https://data.cdc.gov/500-Cities-Places/PLACES-County-Data-GIS-Friendly-Format-2020-releas/mssc-ksj7/about_data){target="_blank" rel="noopener noreferrer"}, and download the data by selecting Actions &rarr; API &rarr; Download file. A subset of data for North Carolina is included in the GeoTox package data as `geo_tox_data$obesity`.

```{r}
#| label: obesity-status
places <- read_csv(
  "PLACES__County_Data_(GIS_Friendly_Format),_2020_release.csv"
)

# Convert confidence interval to standard deviation
extract_SD <- function(x) {
  range <- as.numeric(str_split_1(str_sub(x, 2, -2), ","))
  diff(range) / 3.92
}

geo_tox_data$obesity <- places |>
  # North Carolina Counties
  filter(StateAbbr == "NC") |>
  # Select obesity data
  select(FIPS = CountyFIPS, OBESITY_CrudePrev, OBESITY_Crude95CI) |>
  # Change confidence interval to standard deviation
  rowwise() |>
  mutate(OBESITY_SD = extract_SD(OBESITY_Crude95CI)) |>
  ungroup() |>
  select(-OBESITY_Crude95CI)
```

## Steady-state plasma concentration (C_ss)

Use the `httk` package to generate C_ss values for combinations of age group and weight status for each chemical. The generation of these values is a time-intensive step, so one approach is to generate populations of C_ss values initially and then sample them later. The simulated C_ss values are included in the GeoTox package data as `geo_tox_data$simulated_css`.

```{r}
#| label: C_ss
set.seed(2345)
n_samples <- 500

# Get CASN for which httk simulation is possible. Try using load_dawson2021,
# load_sipes2017, or load_pradeep2020 to increase availability.
load_sipes2017()
casn <- intersect(
  unique(geo_tox_data$dose_response$casn),
  get_cheminfo(suppress.messages = TRUE)
)

# Define population demographics for httk simulation
pop_demo <- cross_join(
  tibble(age_group = list(
    c(0, 2), c(3, 5), c(6, 10), c(11, 15), c(16, 20), c(21, 30), c(31, 40),
    c(41, 50), c(51, 60), c(61, 70), c(71, 100)
  )),
  tibble(weight = c("Normal", "Obese"))
)

# Create wrapper function around httk steps
simulate_css <- function(
  chem.cas,
  agelim_years,
  weight_category,
  samples,
  verbose = TRUE
) {
  if (verbose) {
    cat(
      chem.cas,
      paste0("(", paste(agelim_years, collapse = ", "), ")"),
      weight_category,
      "\n"
    )
  }

  httkpop <- list(
    method = "vi",
    gendernum = NULL,
    agelim_years = agelim_years,
    agelim_months = NULL,
    weight_category = weight_category,
    reths = c(
      "Mexican American",
      "Other Hispanic",
      "Non-Hispanic White",
      "Non-Hispanic Black",
      "Other"
    )
  )

  css <- try(
    suppressWarnings({
      mcs <- create_mc_samples(
        chem.cas = chem.cas,
        samples = samples,
        httkpop.generate.arg.list = httkpop,
        suppress.messages = TRUE
      )

      calc_analytic_css(
        chem.cas = chem.cas,
        parameters = mcs,
        model = "3compartmentss",
        suppress.messages = TRUE
      )
    }),
    silent = TRUE
  )

  # Return
  if (is(css, "try-error")) {
    warning("simulate_css() failed to generate data for CASN ", chem.cas)
    list(NA)
  } else {
    list(css)
  }
}

# Simulate C_ss values
simulated_css <- map(casn, function(chem.cas) {
  pop_demo |>
    rowwise() |>
    mutate(
      css = simulate_css(.env$chem.cas, age_group, weight, .env$n_samples)
    ) |>
    ungroup()
})
simulated_css <- setNames(simulated_css, casn)

# Remove CASN that failed simulate_css
casn_keep <- map_lgl(simulated_css, function(df) {
  !(length(df$css[[1]]) == 1 && is.na(df$css[[1]]))
})
simulated_css <- simulated_css[casn_keep]

geo_tox_data$simulated_css <- simulated_css |>
  map(\(x) {
    x |>
      mutate(
        age_lb = map_int(age_group, first),
        age_ub = map_int(age_group, last)
      ) |>
      select(age_lb, age_ub, weight, css) |>
      unnest(css)
  }) |>
  enframe(name = "casn") |>
  unnest(value)
```

# Retain overlap

Retain only those chemicals found in exposure, dose-response and C_ss datasets.

```{r}
#| label: retain
casn_keep <- unique(geo_tox_data$simulated_css$casn)

geo_tox_data$exposure <- geo_tox_data$exposure |>
  filter(casn %in% casn_keep)

geo_tox_data$dose_response <- geo_tox_data$dose_response |>
  filter(casn %in% casn_keep)
```

# County/State boundaries

Download cartographic boundary files for counties and states from the [U.S. Census Bureau](https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.html){target="_blank" rel="noopener noreferrer"}. The geometry data for North Carolina counties and the state are included in the GeoTox package data as `geo_tox_data$boundaries`.

```{r}
#| label: boundaries
county <- st_read("cb_2020_us_county_5m/cb_2020_us_county_5m.shp")
state <- st_read("cb_2020_us_state_5m/cb_2020_us_state_5m.shp")

geo_tox_data$boundaries <- list(
  county = county |>
    filter(STATEFP == 37) |>
    select(FIPS = GEOID, geometry),
  state = state |>
    filter(STATEFP == 37) |>
    select(geometry)
)
```

---
title: "GeoTox Package Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{GeoTox Package Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This package includes example data `geo_tox_data`. Below is a description of the data and example code for how it was gathered. Please refer to the articles on the package <a href="https://niehs.github.io/GeoTox/" target="_blank">website</a> for examples of how the data are used.

> **NOTE:** The websites used in this vignette were last accessed, and data were generated using the latest package versions, on January 12, 2026.

> **NOTE:** FIPS codes can change. Since data is being pulled from various sources, ensure that the FIPS values can be used to connect data across these sources. For example, in 2022 Connecticut began the process of going from 8 legacy counties to 9 planning regions.

```{r setup, message=FALSE}
library(dplyr)
library(sf)
library(tidyr)
library(readr)
library(stringr)
library(purrr)
library(readxl)
library(httk)
library(httr2)

geo_tox_data <- list()
```

# Chemical data

## Exposure data

Download modeled exposure data from <a href="https://www.epa.gov/AirToxScreen" target="_blank">AirToxScreen</a>. Results from AirToxScreen 2019 for a subset of chemicals in North Carolina counties are included in the package data as `geo_tox_data$exposure`.

> **NOTE:** The 2019 data can be found following the "Previous air toxics assessments" link.

```{r, eval=FALSE}
filename <- "2019_Toxics_Exposure_Concentrations.xlsx"
tmp <- tempfile(filename)
download.file(
  paste0("https://www.epa.gov/system/files/documents/2022-12/", filename),
  tmp
)
exposure <- read_xlsx(tmp)

# Normalization function
min_max_norm = function(x) {
  min_x <- min(x, na.rm = TRUE)
  max_x <- max(x, na.rm = TRUE)
  if (min_x == max_x) {
    rep(0, length(x))
  } else {
    (x - min_x) / (max_x - min_x)
  }
}

geo_tox_data$exposure <- exposure |> 
  # North Carolina counties
  filter(State == "NC", !grepl("0$", FIPS)) |> 
  # Aggregate chemicals by county
  summarize(across(-c(State:Tract), c(mean, sd)), .by = FIPS) |> 
  pivot_longer(-FIPS, names_to = "chemical") |> 
  mutate(stat = if_else(grepl("_1$", chemical), "mean", "sd"),
         chemical = gsub('.{2}$', '', chemical)) |> 
  pivot_wider(names_from = stat) |> 
  # Normalize concentrations
  mutate(norm = min_max_norm(mean), .by = chemical)
```

### Get CASRN and PREFERRED_NAME from CompTox Dashboard

```{r, eval=FALSE}
# Copy/paste the chemical names into the batch search field
# https://comptox.epa.gov/dashboard/batch-search
cat(geo_tox_data$exposure |> distinct(chemical) |> pull(), sep = "\n")
# Export results with "CAS-RN" identifiers as a csv file, then process in R

exposure_casrn <- read_csv("CCD-Batch-Search.csv",
                           show_col_types = FALSE) |> 
  filter(DTXSID != "N/A",
         !grepl("WARNING", FOUND_BY))

# Update exposure data with CompTox Dashboard data
geo_tox_data$exposure <- geo_tox_data$exposure |> 
  inner_join(exposure_casrn, by = join_by(chemical == INPUT)) |> 
  select(FIPS, casn = CASRN, chnm = PREFERRED_NAME, mean, sd, norm)
```

## ICE cHTS data

Use <a href="https://ice.ntp.niehs.nih.gov/" target="_blank">ICE</a> cHTS data to identify active chemicals for a given set of assays.

```{r, eval=FALSE}
get_cHTS_hits <- function(assays = NULL, chemids = NULL) {
  
  if (is.null(assays) & is.null(chemids)) {
    stop("Must provide at least one of 'assays' or 'chemids'")
  }
  
  # Format query parameters
  req_params <- list()
  
  if (!is.null(assays)) {
    if (!is.list(assays)) assays <- as.list(assays)
    req_params$assays <- assays
  }
  
  if (!is.null(chemids)) {
    if (!is.list(chemids)) chemids <- as.list(chemids)
    req_params$chemids <- chemids
  }
  
  # Query ICE API
  resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/search") |> 
    req_body_json(req_params) |> 
    req_perform()
  
  if (resp$status_code != 200) {
    stop("Failed to retrieve data from ICE API")
  }
  
  # Return active chemicals
  result <- resp |> resp_body_json() |> pluck("endPoints")
  
  fields <- c("assay", "casrn", "dtxsid", "substanceName",
              "endpoint", "value")
  
  map(fields, \(x) map_chr(result, x)) |> 
    set_names(fields) |>
    bind_cols() |> 
    filter(endpoint == "Call", value == "Active") |> 
    select(-c(endpoint, value)) |> 
    distinct()
}

assays <- c("APR_HepG2_p53Act_1hr",
            "APR_HepG2_p53Act_24hr",
            "APR_HepG2_p53Act_72hr",
            "ATG_p53_CIS",
            "TOX21_DT40_LUC",
            "TOX21_DT40_100_LUC",
            "TOX21_DT40_657_LUC",
            "TOX21_ELG1_LUC_Agonist",
            "TOX21_H2AX_HTRF_CHO_Agonist_ratio",
            "TOX21_p53_BLA_p1_ratio",
            "TOX21_p53_BLA_p2_ratio",
            "TOX21_p53_BLA_p3_ratio",
            "TOX21_p53_BLA_p4_ratio",
            "TOX21_p53_BLA_p5_ratio")

chemids <- unique(geo_tox_data$exposure$casn)
  
cHTS_hits_API <- get_cHTS_hits(assays = assays, chemids = chemids)
```

## Dose-response data

Use the ICE API to retrieve dose-response data for selected assays and chemicals.

```{r, eval=FALSE}
get_ICE_dose_resp <- function(assays = NULL, chemids = NULL) {
  
  if (is.null(assays) & is.null(chemids)) {
    stop("Must provide at least one of 'assays' or 'chemids'")
  }
  
  # Format query parameters
  req_params <- list()
  
  if (!is.null(assays)) {
    if (!is.list(assays)) assays <- as.list(assays)
    req_params$assays <- assays
  }
  
  if (!is.null(chemids)) {
    if (!is.list(chemids)) chemids <- as.list(chemids)
    req_params$chemids <- chemids
  }
  
  # Query ICE API
  resp <- request("https://ice.ntp.niehs.nih.gov/api/v1/curves") |> 
    req_body_json(req_params) |> 
    req_perform()
  
  if (resp$status_code != 200) {
    stop("Failed to retrieve data from ICE API")
  }
  
  # Return dose-response data
  result <- resp |> resp_body_json() |> pluck("curves")
  
  map(result, function(x) {
    tibble(
      endp = x[["assay"]],
      casn = x[["casrn"]],
      call = x[["dsstoxsid"]],
      chnm = x[["substance"]],
      call = x[["call"]],
      logc = map_dbl(x[["concentrationResponses"]], "concentration") |> log10(),
      resp = map_dbl(x[["concentrationResponses"]], "response")
    )
  }) |> 
    bind_rows()
}

assays <- unique(cHTS_hits_API$assay)

chemids <- intersect(cHTS_hits_API$casrn, geo_tox_data$exposure$casn)

dose_response <- get_ICE_dose_resp(assays = assays, chemids = chemids)

# Only keep active calls for assay/chemical combinations
geo_tox_data$dose_response <- dose_response |> 
  filter(call == "Active") |> 
  select(-call)

# Update dose-response data with CompTox Dashboard data
geo_tox_data$dose_response <- geo_tox_data$dose_response |>
  inner_join(exposure_casrn, by = join_by(casn == CASRN)) |>
  select(endp, casn, chnm = PREFERRED_NAME, logc, resp)
```

# Population data

## Age

Download age data from the <a href="https://www.census.gov/" target="_blank">U.S. Census Bureau</a> by searching for "County Population by Characteristics". A subset of data for North Carolina from <a href="https://www.census.gov/programs-surveys/popest/technical-documentation/research/evaluation-estimates/2020-evaluation-estimates/2010s-county-detail.html" target="_blank">2019</a> is included in the package data as `geo_tox_data$age`.

```{r, eval=FALSE}
# Data for North Carolina
url <- paste0("https://www2.census.gov/programs-surveys/popest/datasets/",
              "2010-2020/counties/asrh/CC-EST2020-ALLDATA-37.csv")
age <- read_csv(url, show_col_types = FALSE)

geo_tox_data$age <- age |>
  # 7/1/2019 population estimate
  filter(YEAR == 12) |>
  # Create FIPS
  mutate(FIPS = str_c(STATE, COUNTY)) |>
  # Keep selected columns
  select(FIPS, AGEGRP, TOT_POP)
```

## Obesity

Follow the "Data Portal" link from <a href="https://www.cdc.gov/places/index.html" target="_blank">CDC PLACES</a> and search for "places county data". Go to the desired dataset webpage, for example <a href="https://data.cdc.gov/500-Cities-Places/PLACES-County-Data-GIS-Friendly-Format-2020-releas/mssc-ksj7/about_data" target="_blank">2020 county data</a>, and download the data by selecting Actions &rarr; API &rarr; Download file. A subset of data for North Carolina is included in the package data as `geo_tox_data$obesity`.

```{r, eval=FALSE}
places <- read_csv("PLACES__County_Data_(GIS_Friendly_Format),_2020_release.csv",
                   show_col_types = FALSE)

# Convert confidence interval to standard deviation
extract_SD <- function(x) {
  range <- as.numeric(str_split_1(str_sub(x, 2, -2), ","))
  diff(range) / 3.92
}

geo_tox_data$obesity <- places |>
  # North Carolina Counties
  filter(StateAbbr == "NC") |>
  # Select obesity data
  select(FIPS = CountyFIPS, OBESITY_CrudePrev, OBESITY_Crude95CI) |>
  # Change confidence interval to standard deviation
  rowwise() |>
  mutate(OBESITY_SD = extract_SD(OBESITY_Crude95CI)) |>
  ungroup() |>
  select(-OBESITY_Crude95CI)
```

## Steady-state plasma concentration (`C_ss`)

Use `httk` to generate `C_ss` values for combinations of age group and weight status for each chemical. The generation of these values is a time-intensive step, so one approach is to generate populations of `C_ss` values initially and then sample them later.

```{r, eval=FALSE}
set.seed(2345)
n_samples <- 500

# Get CASN for which httk simulation is possible. Try using load_dawson2021,
# load_sipes2017, or load_pradeep2020 to increase availability.
load_sipes2017()
casn <- intersect(unique(geo_tox_data$dose_response$casn),
                  get_cheminfo(suppress.messages = TRUE))

# Define population demographics for httk simulation
pop_demo <- cross_join(
  tibble(age_group = list(c(0, 2), c(3, 5), c(6, 10), c(11, 15),
                          c(16, 20), c(21, 30), c(31, 40), c(41, 50),
                          c(51, 60), c(61, 70), c(71, 100))),
  tibble(weight = c("Normal", "Obese"))) |>
  # Create column of lower age_group values
  rowwise() |>
  mutate(age_min = age_group[1]) |>
  ungroup()

# Create wrapper function around httk steps
simulate_css <- function(chem.cas, agelim_years, weight_category,
                         samples, verbose = TRUE) {
  
  if (verbose) {
    cat(chem.cas,
        paste0("(", paste(agelim_years, collapse = ", "), ")"),
        weight_category,
        "\n")
  }
  
  httkpop <- list(method = "vi",
                  gendernum = NULL,
                  agelim_years = agelim_years,
                  agelim_months = NULL,
                  weight_category = weight_category,
                  reths = c(
                    "Mexican American",
                    "Other Hispanic",
                    "Non-Hispanic White",
                    "Non-Hispanic Black",
                    "Other"
                  ))
  
  css <- try(
    suppressWarnings({
      mcs <- create_mc_samples(chem.cas = chem.cas,
                               samples = samples,
                               httkpop.generate.arg.list = httkpop,
                               suppress.messages = TRUE)
      
      calc_analytic_css(chem.cas = chem.cas,
                        parameters = mcs,
                        model = "3compartmentss",
                        suppress.messages = TRUE)
    }),
    silent = TRUE
  )
  
  # Return
  if (is(css, "try-error")) {
    warning(paste0("simulate_css failed to generate data for CASN ", chem.cas))
    list(NA)
  } else {
    list(css)
  }
}

# Simulate `C_ss` values
simulated_css <- map(casn, function(chem.cas) {
  pop_demo |>
    rowwise() |>
    mutate(
      css = simulate_css(.env$chem.cas, age_group, weight, .env$n_samples)
    ) |>
    ungroup()
})
simulated_css <- setNames(simulated_css, casn)

# Remove CASN that failed simulate_css
casn_keep <- map_lgl(simulated_css, function(df) {
  !(length(df$css[[1]]) == 1 && is.na(df$css[[1]]))
})
simulated_css <- simulated_css[casn_keep]

# Get median `C_ss` values for each age_group
simulated_css <- map(
  simulated_css,
  function(cas_df) {
    cas_df |>
      nest(.by = age_group) |>
      mutate(
        age_median_css = map_dbl(data, function(df) median(unlist(df$css),
                                                           na.rm = TRUE))
      ) |>
      unnest(data)
  }
)

# Get median `C_ss` values for each weight
simulated_css <- map(
  simulated_css,
  function(cas_df) {
    cas_df |>
      nest(.by = weight) |>
      mutate(
        weight_median_css = map_dbl(data, function(df) median(unlist(df$css),
                                                              na.rm = TRUE))
      ) |>
      unnest(data) |>
      arrange(age_min, weight)
  }
)

geo_tox_data$simulated_css <- simulated_css
```

# Prune data

Retain only those chemicals found in exposure, dose-response and `C_ss` datasets.

```{r, eval=FALSE}
casn_keep <- names(geo_tox_data$simulated_css)

geo_tox_data$exposure <- geo_tox_data$exposure |> 
  filter(casn %in% casn_keep)

geo_tox_data$dose_response <- geo_tox_data$dose_response |> 
  filter(casn %in% casn_keep)
```

# County/State boundaries

Download cartographic boundary files for counties and states from the <a href="https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.html" target="_blank">U.S. Census Bureau</a>. The geometry data for North Carolina counties and the state are included in the package data as `geo_tox_data$boundaries`.

```{r, eval=FALSE}
county <- st_read("cb_2019_us_county_5m/cb_2019_us_county_5m.shp")
state <- st_read("cb_2019_us_state_5m/cb_2019_us_state_5m.shp")

geo_tox_data$boundaries <- list(
  county = county |>
    filter(STATEFP == 37) |>
    select(FIPS = GEOID, geometry),
  state = state |>
    filter(STATEFP == 37) |>
    select(geometry)
)
```

# sensitivity analysis

    Code
      res
    Output
      $risk_sensitivity_age
      # A tibble: 3 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1    1.94   1.94      9.27     9.27
      2       NA         2    2.06   2.06     10.4     10.4 
      3       NA         3    2.13   2.13     11.3     11.3 
      
      $risk_sensitivity_weight
      # A tibble: 3 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1   0.663  0.663      1.89     1.89
      2       NA         2   0.812  0.812      2.42     2.42
      3       NA         3   2.96   2.96      30.8     30.8 
      
      $risk_sensitivity_css_params
      # A tibble: 3 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1   0.591  0.591      1.64     1.64
      2       NA         2   0.479  0.479      1.29     1.29
      3       NA         3   1.14   1.14       3.82     3.82
      
      $risk_sensitivity_fit_params
      # A tibble: 3 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1   1.47   1.47    320.     320.   
      2       NA         2   0.376  0.376     0.899    0.899
      3       NA         3   0.610  0.610     1.31     1.31 
      
      $risk_sensitivity_C_ext
      # A tibble: 3 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1   1.00   1.00       3.20     3.20
      2       NA         2   0.830  0.830      2.49     2.49
      3       NA         3   2.38   2.38      14.8     14.8 
      

# calc sensitivity

    Code
      sens_tbl
    Output
      # A tibble: 3 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1    1.94   1.94      9.27     9.27
      2       NA         2    2.06   2.06     10.4     10.4 
      3       NA         3    2.13   2.13     11.3     11.3 

# other inputs

    Code
      sens_tbl
    Output
      # A tibble: 3 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1    3.78   3.78    768.     768.  
      2       NA         2    1.97   1.97      9.55     9.55
      3       NA         3    3.82   3.82   4967     4967   

# overwrite

    Code
      risk_tbl
    Output
      # A tibble: 1 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1   0.452  0.452    712700   712700

# normal risk

    Code
      risk_tbl
    Output
      # A tibble: 2 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10  IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>     <dbl>
      1       NA         1   0.497  0.497  15220000  15220000
      2       NA         2   0.601  0.602 135500000 135500000

# risk sensitivity

    Code
      risk_tbl
    Output
      # A tibble: 2 x 6
        assay_id sample_id GCA.Eff IA.Eff GCA.HQ.10 IA.HQ.10
           <dbl>     <dbl>   <dbl>  <dbl>     <dbl>    <dbl>
      1       NA         1   0.590  0.591  15080000 15080000
      2       NA         2   0.596  0.597  89640000 89640000

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_response.R
\name{calc_response}
\alias{calc_response}
\title{Calculate concentrations and risks}
\usage{
calc_response(GT, overwrite = FALSE, ...)
}
\arguments{
\item{GT}{GeoTox object.}

\item{overwrite}{Logical indicating whether to overwrite existing values
(default FALSE).}

\item{...}{Additional arguments passed to \code{\link[=calc_risk]{calc_risk()}}.}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Calculate internal dose, in vitro concentration, and risk estimates.
}
\details{
This is a wrapper around several other functions:
\itemize{
\item \code{\link[=calc_internal_dose]{calc_internal_dose()}} to calculate internal dose (D_int).
\item \code{\link[=calc_invitro_concentration]{calc_invitro_concentration()}} to calculate in vitro concentration
(C_invitro).
\item \code{\link[=calc_risk]{calc_risk()}} to calculate risk estimates.
}

If a \code{risk_name} argument is provided to \code{...} and it is not "risk", then
sensitivity analysis is assumed and the \code{sensitivity} argument in the
internal dose and in vitro concentration calculations is set to TRUE.
}
\examples{
# Setup required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal", 21,
  "00-00-1", 50, 99, "Normal", 22,
  "00-00-1",  0, 49,  "Obese", 61,
  "00-00-1", 50, 99,  "Obese", 62,
  "00-00-2",  0, 49, "Normal", 11,
  "00-00-2", 50, 99, "Normal", 12,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
hill_df <- tibble::tribble(
  ~assay, ~casn, ~logc, ~resp,
  "a1", "00-00-1",    0,  10,
  "a1", "00-00-1",    1,  20,
  "a1", "00-00-1",    2,  80,
  "a1", "00-00-1",    3, 100,
  "a1", "00-00-2", -0.5,   5,
  "a1", "00-00-2",  0.5,  20,
  "a1", "00-00-2",  1.5,  55,
  "a1", "00-00-2",  2.5,  60
)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  set_simulated_css(css_df) |>
  add_exposure_rate_params() |>
  add_hill_params(fit_hill(hill_df, assay = "assay", substance = "casn")) |>
  simulate_population(exposure = exposure_df)

# Calculate concentrations and risk
GT <- GT |> calc_response()

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant table

dplyr::tbl(con, "concentration") |> dplyr::collect()

dplyr::tbl(con, "risk") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=calc_internal_dose]{calc_internal_dose()}}, \code{\link[=calc_invitro_concentration]{calc_invitro_concentration()}},
\code{\link[=calc_risk]{calc_risk()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_sample.R
\name{set_sample}
\alias{set_sample}
\title{Set sample data}
\usage{
set_sample(GT, df, location = "FIPS", overwrite = FALSE)
}
\arguments{
\item{GT}{GeoTox object.}

\item{df}{Data frame containing sample data.}

\item{location}{Column name(s) in \code{df} that contain location identifier(s)
(default "FIPS").}

\item{overwrite}{Logical indicating whether to overwrite existing 'sample'
table and remove existing 'concentration' and 'risk' tables (default
FALSE).}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Create the 'sample' table in the GeoTox database.
}
\details{
The 'sample' table consists of individual characteristics and location
information. At a minimum, it must contain columns with information on age,
weight category, and location identifier(s).

When the \code{df} input contains information on age or weight category, the
column names must be "age" and "weight", respectively. The \code{location} input
(default "FIPS") can be a named vector to specify multiple identifier columns
in \code{df}. For example, \code{location = c(FIPS = "FIPS", state = "ST")} would
indicate that \code{df} contains both FIPS codes and state identifiers for
locations. The location column(s) will be added to the 'location' table and
replaced with a corresponding "location_id" column in the 'sample' table. The
\code{state = "ST"} part in the example above would rename the "ST" column in \code{df}
to "state" in the 'location' table.

There are several other functions that can be used to create or add to the
'sample' table: \code{\link[=simulate_age]{simulate_age()}} to generate age data, \code{\link[=simulate_obesity]{simulate_obesity()}}
to generate weight category data, and \code{\link[=simulate_population]{simulate_population()}}, which is a
wrapper function that can generate both age and weight category data along
with other fields. \code{set_sample()} can be used in combination with these
functions to first set some known sample data, e.g. age, and then simulate
any missing fields, e.g. weight category.

If \code{overwrite = TRUE}, any existing 'concentration' and 'risk' tables will be
dropped before creating the 'sample' table. This is because the existing
concentration and risk data would no longer be valid for the new sample data.
Further downstream tables are not dropped automatically, but can be updated
during subsequent simulation and analysis steps.
}
\examples{
# Example sample data
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)

# Set sample data
GT <- GeoTox() |> set_sample(sample_df)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=simulate_age]{simulate_age()}}, \code{\link[=simulate_obesity]{simulate_obesity()}}, \code{\link[=simulate_population]{simulate_population()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GeoTox-package.R
\docType{package}
\name{GeoTox-package}
\alias{GeoTox-package}
\title{GeoTox: Spatiotemporal Mixture Risk Assessment}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Connecting spatiotemporal exposure to individual and population-level risk via source-to-outcome continuum modeling. The package, methods, and case-studies are described in Messier, Reif, and Marvel (2025) \doi{10.1186/s40246-024-00711-8} and Eccles et al. (2023) \doi{10.1016/j.scitotenv.2022.158905}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://niehs.github.io/GeoTox/}
  \item \url{https://github.com/NIEHS/GeoTox}
  \item Report bugs at \url{https://github.com/NIEHS/GeoTox/issues}
}

}
\author{
\strong{Maintainer}: Skylar Marvel \email{skylar.marvel@nih.gov} (\href{https://orcid.org/0000-0002-2971-9743}{ORCID})

Authors:
\itemize{
  \item Skylar Marvel \email{skylar.marvel@nih.gov} (\href{https://orcid.org/0000-0002-2971-9743}{ORCID})
  \item David Reif \email{david.reif@nih.gov} (\href{https://orcid.org/0000-0001-7815-6767}{ORCID})
  \item Kyle Messier \email{kyle.messier@nih.gov} (\href{https://orcid.org/0000-0001-9508-9623}{ORCID})
}

Other contributors:
\itemize{
  \item Spatiotemporal Exposures and Toxicology Group [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_population.R
\name{simulate_population}
\alias{simulate_population}
\title{Simulate population characteristics and exposures}
\usage{
simulate_population(
  GT,
  age = NULL,
  obesity = NULL,
  exposure = NULL,
  simulate_rate = TRUE,
  sample_css = TRUE,
  ...
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{age}{Data frame with age data.}

\item{obesity}{Data frame with obesity data.}

\item{exposure}{Data frame with exposure data.}

\item{simulate_rate}{Logical indicating whether to simulate exposure rates.
This requires that exposure rate parameters have been added using
\code{\link[=add_exposure_rate_params]{add_exposure_rate_params()}}.}

\item{sample_css}{Logical indicating whether to sample steady-state plasma
concentrations (C\eqn{_{ss}}). This requires that a table of simulated
C\eqn{_{ss}} values has been set using \code{\link[=set_simulated_css]{set_simulated_css()}}. In addition,
\code{\link[=set_fixed_css]{set_fixed_css()}} will be called after sampling to prepare C\eqn{_{ss}}
values for sensitivity analysis.}

\item{...}{Additional arguments passed to wrapped functions (see 'Additional
arguments' section of 'Details').}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Simulate age, weight category, exposure rates, and external exposures. Sample
from pre-simulated steady-state plasma concentrations.
}
\details{
This is a wrapper around several other functions:
\itemize{
\item \code{\link[=add_age]{add_age()}} and \code{\link[=simulate_age]{simulate_age()}} for age simulation.
\item \code{\link[=add_obesity]{add_obesity()}} and \code{\link[=simulate_obesity]{simulate_obesity()}} for weight category simulation.
\item \code{\link[=add_exposure]{add_exposure()}} and \code{\link[=simulate_exposure]{simulate_exposure()}} for external exposure
concentration simulation (C\eqn{_{ext}}).
\item \code{\link[=simulate_exposure_rate]{simulate_exposure_rate()}} for exposure rate simulation.
\item \code{\link[=sample_simulated_css]{sample_simulated_css()}} for sampling steady-state plasma concentrations
(C\eqn{_{ss}}).
\item \code{\link[=set_fixed_css]{set_fixed_css()}} to prepare C\eqn{_{ss}} values for sensitivity analysis.
}

The user can provide data frames for age, obesity, and exposure data; if any
of these are provided, the corresponding add and simulate functions will be
called. The user can also specify whether to simulate exposure rates and
sample C\eqn{_{ss}} values using the \code{simulate_rate} and \code{sample_css}
arguments, respectively. If \code{simulate_rate} is \code{TRUE}, exposure rate
parameters must have been added using \code{\link[=add_exposure_rate_params]{add_exposure_rate_params()}}. If
\code{sample_css} is \code{TRUE}, a table of simulated C\eqn{_{ss}} values must already
exist using \code{\link[=set_simulated_css]{set_simulated_css()}}.
\subsection{Additional arguments:}{

\describe{
\item{n}{Number of samples to simulate (default 1000). Used in
\code{\link[=simulate_age]{simulate_age()}}, \code{\link[=simulate_obesity]{simulate_obesity()}}, and \code{\link[=simulate_exposure]{simulate_exposure()}}. Ignored
if the 'sample' table already exists, in which case the existing sample sizes
are used.}
\item{location}{Column name for location ID (default "FIPS"). Used in
\code{\link[=add_age]{add_age()}}, \code{\link[=add_obesity]{add_obesity()}}, and \code{\link[=add_exposure]{add_exposure()}}.}
\item{overwrite}{Logical indicating whether to overwrite existing values
(default FALSE). Used in \code{\link[=simulate_age]{simulate_age()}}, \code{\link[=simulate_obesity]{simulate_obesity()}},
\code{\link[=simulate_exposure_rate]{simulate_exposure_rate()}}, and \code{\link[=simulate_exposure]{simulate_exposure()}}.}
\item{substance}{Column name for substance ID (default "casn"). Used in
\code{\link[=add_exposure]{add_exposure()}}.}
\item{route}{Column name for exposure route (default "route"). Used in
\code{\link[=add_exposure]{add_exposure()}}.}
\item{rate_extra_cols}{Additional columns to include in exposure_rate table.
Used in \code{\link[=simulate_exposure_rate]{simulate_exposure_rate()}}.}
\item{obes_prev, obes_sd}{Column names for obesity prevalence and standard
deviation (default "OBESITY_CrudePrev" and "OBESITY_SD", respectively). Used
in \code{\link[=simulate_obesity]{simulate_obesity()}}.}
\item{expos_mean, expos_sd}{Column names for exposure concentration mean and
standard deviation (default "mean" and "sd", respectively). Used in
\code{\link[=simulate_exposure]{simulate_exposure()}}.}
\item{css_extra_cols}{Additional columns to include in simulated_css table.
Used in \code{\link[=sample_simulated_css]{sample_simulated_css()}}.}
\item{substance_order}{Named list specifying order of substances. Used in
\code{\link[=sample_simulated_css]{sample_simulated_css()}} and \code{\link[=set_fixed_css]{set_fixed_css()}}.}
}
}
}
\examples{
# Example simulation data

age_df <- data.frame(
  FIPS = rep(c(10000, 20000), each = 19),
  AGEGRP = rep(0:18, times = 2),
  TOT_POP = 0
)
# FIPS 10000, populate age group 40-44
age_df$TOT_POP[c(1, 10)] = 100
# FIPS 20000, populate age groups 50-59
age_df$TOT_POP[c(1, 12, 13) + 19] = c(200, 100, 100)

obesity_df <- data.frame(
  FIPS = c(10000, 20000),
  OBESITY_CrudePrev = c(20, 80),
  OBESITY_SD = 5
)

exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)

# Note: normally the css_df would have many more rows for each combination of
# the non-'css' columns to allow for sampling.
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal",  1,
  "00-00-1", 50, 99, "Normal",  2,
  "00-00-1",  0, 49,  "Obese", 11,
  "00-00-1", 50, 99,  "Obese", 12,
  "00-00-2",  0, 49, "Normal", 21,
  "00-00-2", 50, 99, "Normal", 22,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)

# Simulate population
GT <- GeoTox() |>
  add_exposure_rate_params() |>
  set_simulated_css(css_df) |>
  simulate_population(
    age = age_df,
    obesity = obesity_df,
    exposure = exposure_df,
    n = 3
  )

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "concentration") |> dplyr::collect()

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

dplyr::tbl(con, "substance") |> dplyr::collect()

dplyr::tbl(con, "route") |> dplyr::collect()

# Note: the 'age', 'weight', 'params', and 'other' columns of the
# 'fixed_css' table contain the C_ss values for sensitivity analysis.
# For example, the 'age' column doesn't contain ages, but C_ss values.
dplyr::tbl(con, "fixed_css") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=add_age]{add_age()}}, \code{\link[=simulate_age]{simulate_age()}}, \code{\link[=add_obesity]{add_obesity()}},
\code{\link[=simulate_obesity]{simulate_obesity()}}, \code{\link[=add_exposure]{add_exposure()}}, \code{\link[=simulate_exposure]{simulate_exposure()}},
\code{\link[=simulate_exposure_rate]{simulate_exposure_rate()}}, \code{\link[=sample_simulated_css]{sample_simulated_css()}}, \code{\link[=set_fixed_css]{set_fixed_css()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_exposure.R
\name{simulate_exposure}
\alias{simulate_exposure}
\title{Simulate exposure concentrations}
\usage{
simulate_exposure(
  GT,
  n = 1000,
  overwrite = FALSE,
  expos_mean = NULL,
  expos_sd = NULL,
  sensitivity = FALSE
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{n}{Number of individuals to simulate per location (default 1000).
Ignored if 'sample' table already exists.}

\item{overwrite}{Logical indicating whether to overwrite existing 'C_ext'
values in the 'concentration' table (default FALSE).}

\item{expos_mean}{Column name of exposure concentration mean in the
'exposure' table (default "mean").}

\item{expos_sd}{Column name of exposure concentration standard deviation in
the 'exposure' table (default "sd").}

\item{sensitivity}{Logical indicating whether to simulate exposures for
sensitivity analysis (default FALSE).}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Simulate external exposure (C_ext) values to be stored in the
'concentration' table of a GeoTox database.
}
\details{
An 'external' table containing simulation data must already exist in the
GeoTox database, which is added using \code{\link[=add_exposure]{add_exposure()}}.

The inputs \code{expos_mean} and \code{expos_sd} will be assigned default values of
"mean" and "sd", respectively, if not provided and not already set in the
GeoTox object's parameters, \code{GT$par}. If not \code{NULL}, the provided values will
also be saved to \code{GT$par}.

If \code{sensitivity = TRUE}, exposure concentrations will be simulated for
sensitivity analysis. Typically this shouldn't be used directly by the user,
but rather called by \code{\link[=calc_sensitivity]{calc_sensitivity()}}. In this case, the function will
use the 'concentration_sensitivity' table instead of the 'concentration'
table, and will assume that the 'sample' table already exists.
}
\examples{
# Example exposure simulation data
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)

# Simulate C_ext values
GT <- GeoTox() |>
  add_exposure(exposure_df) |>
  simulate_exposure(n = 3)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "concentration") |> dplyr::collect()

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

dplyr::tbl(con, "substance") |> dplyr::collect()

dplyr::tbl(con, "route") |> dplyr::collect()

# Replace 'exposure' table with different column names
names(exposure_df)[4:5] <- c("mu", "sigma")
DBI::dbRemoveTable(con, "exposure")
GT |> add_exposure(exposure_df)

# Overwrite 'C_ext' values in existing 'concentration' table
# Must specify new 'exposure' column names
# Notice how the column names are added to GT$par
str(GT$par)
GT <- GT |>
  simulate_exposure(expos_mean = "mu", expos_sd = "sigma", overwrite = TRUE)
str(GT$par)

# Look at updated 'concentration' table
dplyr::tbl(con, "concentration") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=add_exposure]{add_exposure()}}, \code{\link[=simulate_population]{simulate_population()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_sensitivity.R
\name{calc_sensitivity}
\alias{calc_sensitivity}
\title{Calculate risk sensitivity to a single variable}
\usage{
calc_sensitivity(
  GT,
  vary = c("age", "weight", "css_params", "fit_params", "C_ext"),
  overwrite = FALSE,
  rate_extra_cols = NULL,
  expos_mean = NULL,
  expos_sd = NULL,
  max_mult = 1.5
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{vary}{Variable to vary. One of "age", "weight", "css_params",
"fit_params", or "C_ext".}

\item{overwrite}{Logical indicating whether to overwrite existing sensitivity
analysis results in the GeoTox database.}

\item{rate_extra_cols}{Additional columns to match from the
'exposure_rate_params' table (default NULL).}

\item{expos_mean}{Column name of exposure concentration mean in the
'exposure' table (default "mean").}

\item{expos_sd}{Column name of exposure concentration standard deviation in
the 'exposure' table (default "sd").}

\item{max_mult}{Upper bound multiplier for max response (default 1.5).}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Compute risk by varying one variable while holding others fixed.
}
\details{
The sensitivity analysis makes use of the C\eqn{_{ss}} values stored in the
'fixed_css' table of the GeoTox database. These values are determined using
the pre-simulated C\eqn{_{ss}} values supplied to \code{\link[=set_simulated_css]{set_simulated_css()}} and
can be set using \code{\link[=set_fixed_css]{set_fixed_css()}} prior to running this function. This step
is automatically done when using \code{\link[=simulate_population]{simulate_population()}} with \code{sample_css = TRUE}.

There are five options for the \code{vary} argument:
\describe{
\item{age}{C\eqn{_{ss}} values from the 'age' column of the 'fixed_css' table
are used. For other cases, exposure rates are re-simulated using the median
age by location in the 'sample' table by calling \code{\link[=simulate_exposure_rate]{simulate_exposure_rate()}}
with \code{sensitivity = TRUE}.}
\item{weight}{C\eqn{_{ss}} values from the 'weight' column of the 'fixed_css'
table are used.}
\item{css_params}{C\eqn{_{ss}} values from the 'params' column of the
'fixed_css' table are used.}
\item{fit_params}{C\eqn{_{ss}} values from the 'other' column of the
'fixed_css' table are used. For other cases, the standard deviation of
dose-response model fit parameters are set to zero by calling \code{\link[=calc_risk]{calc_risk()}}
with \code{fixed = TRUE}.}
\item{C_ext}{C\eqn{_{ss}} values from the 'other' column of the 'fixed_css'
table are used. For other cases, external concentrations are re-simulated
with standard deviations set to zero by calling \code{\link[=simulate_exposure]{simulate_exposure()}} with
\code{sensitivity = TRUE}.}
}
In all cases above, the resulting risk table is named
'risk_sensitivity_~vary~' (e.g., 'risk_sensitivity_age') in the GeoTox
database.

Inputs \code{rate_extra_cols}, \code{expos_mean}, and \code{expos_sd} do not need to be
specified again if they were already provided in a previous call and are set
in the GeoTox parameters (\code{GT$par}).
}
\examples{
# Example setup is shown below in \dontrun().
# Pre-generated results will be loaded instead to avoid long example runtime.

\dontrun{
# Setup required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal", 21,
  "00-00-1", 50, 99, "Normal", 22,
  "00-00-1",  0, 49,  "Obese", 61,
  "00-00-1", 50, 99,  "Obese", 62,
  "00-00-2",  0, 49, "Normal", 11,
  "00-00-2", 50, 99, "Normal", 12,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
hill_df <- tibble::tribble(
  ~assay, ~model, ~casn, ~logc, ~resp,
  "a1", "human", "00-00-1",    0,  10,
  "a1", "human", "00-00-1",    1,  20,
  "a1", "human", "00-00-1",    2,  80,
  "a1", "human", "00-00-1",    3, 100,
  "a1", "human", "00-00-2", -0.5,   5,
  "a1", "human", "00-00-2",  0.5,  20,
  "a1", "human", "00-00-2",  1.5,  55,
  "a1", "human", "00-00-2",  2.5,  60,
  "a2",   "rat", "00-00-1",   -1,   0,
  "a2",   "rat", "00-00-1",    0,  10,
  "a2",   "rat", "00-00-1",    1,  30,
  "a2",   "rat", "00-00-1",    2,  40
)
set.seed(1234)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  set_simulated_css(css_df) |>
  add_exposure_rate_params() |>
  add_hill_params(fit_hill(
    hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
  )) |>
  simulate_population(exposure = exposure_df) |>
  calc_response()

# Calculate sensitivity to age
GT <- GT |> calc_sensitivity("age")

}

# Load results from pre-generated database for this example
temp_dir <- tempdir()
zip::unzip(
  system.file("extdata", "sensitivity.duckdb.zip", package = "GeoTox"),
  junkpaths = TRUE,
  exdir = temp_dir
)
GT <- GeoTox(paste0(temp_dir, "/sensitivity.duckdb"))

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant table

dplyr::tbl(con, "risk_sensitivity_age") |> dplyr::collect()

# Compared to baseline risk table
dplyr::tbl(con, "risk") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=set_simulated_css]{set_simulated_css()}}, \code{\link[=set_fixed_css]{set_fixed_css()}}, \code{\link[=sensitivity_analysis]{sensitivity_analysis()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_risk.R
\name{calc_risk}
\alias{calc_risk}
\title{Calculate risk scores}
\usage{
calc_risk(
  GT,
  max_mult = 1.5,
  fixed = FALSE,
  overwrite = FALSE,
  risk_name = "risk"
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{max_mult}{Upper bound multiplier for max response (default 1.5).}

\item{fixed}{Logical indicating whether to set standard deviation parameters
of Hill fit to zero (default FALSE). Used in sensitivity analysis.}

\item{overwrite}{Logical indicating whether to overwrite existing risk table
(default FALSE).}

\item{risk_name}{Table name to store risk results (default "risk"). Values
other than "risk" are used in sensitivity analysis.}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Calculate generalized concentration addition (GCA) and independent action
(IA) risk scores and hazard quotients (HQ) based on in vitro concentration
data and Hill parameters.
}
\details{
This function requires that the 'concentration', 'sample', and 'hill_params'
tables are present in the GeoTox object. Typically these tables will have
been created by prior calls to \code{\link[=calc_invitro_concentration]{calc_invitro_concentration()}} and
\code{\link[=add_hill_params]{add_hill_params()}}.

The risk scores are calculated for each sample and assay combination and
stored in the 'risk' table in the GeoTox object, unless a different
\code{risk_name} is provided. Supplying a different \code{risk_name} shouldn't be done
directly by the user, but rather by calling \code{\link[=calc_sensitivity]{calc_sensitivity()}}.
}
\examples{
# Setup required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal", 21,
  "00-00-1", 50, 99, "Normal", 22,
  "00-00-1",  0, 49,  "Obese", 61,
  "00-00-1", 50, 99,  "Obese", 62,
  "00-00-2",  0, 49, "Normal", 11,
  "00-00-2", 50, 99, "Normal", 12,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
hill_df <- tibble::tribble(
  ~assay, ~casn, ~logc, ~resp,
  "a1", "00-00-1",    0,  10,
  "a1", "00-00-1",    1,  20,
  "a1", "00-00-1",    2,  80,
  "a1", "00-00-1",    3, 100,
  "a1", "00-00-2", -0.5,   5,
  "a1", "00-00-2",  0.5,  20,
  "a1", "00-00-2",  1.5,  55,
  "a1", "00-00-2",  2.5,  60
)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  set_simulated_css(css_df) |>
  add_exposure_rate_params() |>
  add_hill_params(fit_hill(hill_df, assay = "assay", substance = "casn")) |>
  simulate_population(exposure = exposure_df) |>
  calc_internal_dose() |>
  calc_invitro_concentration()

# Calculate risk
GT <- GT |> calc_risk()

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant table
dplyr::tbl(con, "risk") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=add_hill_params]{add_hill_params()}}, \code{\link[=calc_invitro_concentration]{calc_invitro_concentration()}},
\code{\link[=calc_response]{calc_response()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_age.R
\name{simulate_age}
\alias{simulate_age}
\title{Simulate age values}
\usage{
simulate_age(GT, n = 1000, overwrite = FALSE)
}
\arguments{
\item{GT}{GeoTox object.}

\item{n}{Number of individuals to simulate per location (default 1000).
Ignored if 'sample' table already exists.}

\item{overwrite}{Logical indicating whether to overwrite existing 'age'
values in the 'sample' table (default FALSE).}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Simulate 'age' values to be stored in the 'sample' table of a GeoTox
database.
}
\details{
An 'age' table containing simulation data must already exist in the GeoTox
database, which is added using \code{\link[=add_age]{add_age()}}.
}
\examples{
# Example age simulation data
age_df <- data.frame(
  FIPS = rep(c(10000, 20000), each = 19),
  AGEGRP = rep(0:18, times = 2),
  TOT_POP = 0
)
# FIPS 10000, populate age group 40-44
age_df$TOT_POP[c(1, 10)] = 100
# FIPS 20000, populate age groups 50-59
age_df$TOT_POP[c(1, 12, 13) + 19] = c(200, 100, 100)

# Simulate age values
GT <- GeoTox() |>
  add_age(age_df) |>
  simulate_age(n = 5)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

# Overwrite existing age values
GT <- GT |> simulate_age(overwrite = TRUE)

# Look at updated 'sample' table
dplyr::tbl(con, "sample") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=add_age]{add_age()}}, \code{\link[=simulate_population]{simulate_population()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_fixed_css.R
\name{set_fixed_css}
\alias{set_fixed_css}
\title{Prepare steady-state plasma concentrations for sensitivity analysis}
\usage{
set_fixed_css(GT, substance_order = NULL)
}
\arguments{
\item{GT}{GeoTox object.}

\item{substance_order}{Named list specifying order of substance evaluation
(default NULL).}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Create the 'fixed_css' table in the GeoTox database, which contains values of
steady-state plasma concentrations (C\eqn{_{ss}}) for sensitivity analysis.
}
\details{
Several tables are required in the GeoTox database before the 'fixed_css'
table can be created. Typically, \code{set_fixed_css()} is called after using
\code{\link[=sample_simulated_css]{sample_simulated_css()}}, at which point all required tables will be present.

The resulting 'fixed_css' table is used for sensitivity analysis where one
parameter is allowed to vary at a time while all other parameters are held
constant at fixed values.

The \code{substance_order} argument can be used to specify the order in which
substances are evaluated when sampling C\eqn{_{ss}} values. The default is to
evaluate substances in the order they appear in the 'substance' table.
However, if a different order is needed for some reason (e.g., replication of
results from a previous GeoTox implementation), the user can provide a named
list where the name is the column in the 'substance' table to use for
ordering (e.g., "casn") and the value is a vector of substance identifiers in
the desired order. If provided, the \code{substance_order} will be added to the
GeoTox object parameter list, \code{GT$par}. This is only applicable to the
"params" section described below.
\subsection{Table 'fixed_css' columns:}{

\describe{
\item{age}{Median pre-simulated C\eqn{_{ss}} values are computed by age
group for each substance, then the median C\eqn{_{ss}} values are assigned
to each individual based on their age group.}
\item{weight}{Median pre-simulated C\eqn{_{ss}} values are computed by weight
category for each substance, then the median C\eqn{_{ss}} values are assigned
to each individual based on their weight category.}
\item{params}{Individuals are assigned the median age of their location and
pre-simulated C\eqn{_{ss}} values for the "Normal" weight category are
sampled.}
\item{other}{Median sampled C\eqn{_{ss}} values are computed across all
substances for each location after mean-imputation of missing C\eqn{_{ss}}
values for each substance and location. The median C\eqn{_{ss}} values are
assigned to each individual based on their location.}
}
}
}
\examples{
# Create required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
# Note: normally the css_df would have many more rows for each combination of
# the non-'css' columns to allow for sampling.
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal",  1,
  "00-00-1", 50, 99, "Normal",  2,
  "00-00-1",  0, 49,  "Obese", 11,
  "00-00-1", 50, 99,  "Obese", 12,
  "00-00-2",  0, 49, "Normal", 21,
  "00-00-2", 50, 99, "Normal", 22,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  add_exposure(exposure_df) |>
  simulate_exposure() |>
  set_simulated_css(css_df) |>
  sample_simulated_css()

# Set fixed C_ss values
GT <- GT |> set_fixed_css()

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables
# Note: the 'age', 'weight', 'params', and 'other' columns of the
# 'fixed_css' table contain the C_ss values for sensitivity analysis.
# For example, the 'age' column doesn't contain ages, but C_ss values.

dplyr::tbl(con, "fixed_css") |> dplyr::collect()

dplyr::tbl(con, "concentration") |> dplyr::collect()

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

dplyr::tbl(con, "substance") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=sample_simulated_css]{sample_simulated_css()}}, \code{\link[=simulate_population]{simulate_population()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_simulated_css.R
\name{sample_simulated_css}
\alias{sample_simulated_css}
\title{Sample from pre-simulated steady-state plasma concentrations}
\usage{
sample_simulated_css(GT, css_extra_cols = NULL, substance_order = NULL)
}
\arguments{
\item{GT}{GeoTox object.}

\item{css_extra_cols}{Additional columns to match from the 'simulated_css'
table (default NULL).}

\item{substance_order}{Named list specifying order of substance evaluation
(default NULL).}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Sample steady-state plasma concentrations (C\eqn{_{ss}}) for individuals from
pre-simulated values stored in the 'simulated_css' table in a GeoTox
database.
}
\details{
The C\eqn{_{ss}} values are sampled from the 'simulated_css' table, which
must already exist in the GeoTox database; it can be created using
\code{\link[=set_simulated_css]{set_simulated_css()}}. The minimum characteristics that are used to match
individuals to sets of C\eqn{_{ss}} values are age and weight category
("Normal" or "Obese"). Additional columns (which must exist in both the
'simulated_css' and 'sample' tables) can be specified using the
\code{css_extra_cols} argument. If \code{css_exta_cols} is provided it will be added to
the GeoTox object parameter list, \code{GT$par}.

In addition to the 'simulated_css' table, both 'sample' and 'concentration'
tables must also exist in the GeoTox database. The 'sample' table must
contain the characteristics of individuals (age, weight category, and any
additional columns specified in \code{css_extra_cols}). One way to create this
'sample' table is by using \code{\link[=set_sample]{set_sample()}}. The 'concentration' table will
typically be created using \code{\link[=simulate_exposure]{simulate_exposure()}} and will be populated with
rows for each individual and substance combination where exposure data is
available. The sampled C\eqn{_{ss}} values will be stored in a new "C_ss"
column in the 'concentration' table.

The \code{substance_order} argument can be used to specify the order in which
substances are evaluated when sampling C\eqn{_{ss}} values. The default is to
evaluate substances in the order they appear in the 'substance' table.
However, if a different order is needed for some reason (e.g., replication of
results from a previous GeoTox implementation), the user can provide a named
list where the name is the column in the 'substance' table to use for
ordering (e.g., "casn") and the value is a vector of substance identifiers in
the desired order. If provided, the \code{substance_order} will be added to the
GeoTox object parameter list, \code{GT$par}.
}
\examples{
# Create required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
# Note: normally the css_df would have many more rows for each combination of
# the non-'css' columns to allow for sampling.
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal",  1,
  "00-00-1", 50, 99, "Normal",  2,
  "00-00-1",  0, 49,  "Obese", 11,
  "00-00-1", 50, 99,  "Obese", 12,
  "00-00-2",  0, 49, "Normal", 21,
  "00-00-2", 50, 99, "Normal", 22,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  add_exposure(exposure_df) |>
  simulate_exposure() |>
  set_simulated_css(css_df)

# Sample simulated C_ss values
GT <- GT |> sample_simulated_css()

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables. sample_simulated_css() generated the 'C_ss' column
# of the 'concentration' table.

dplyr::tbl(con, "concentration") |> dplyr::collect()

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

dplyr::tbl(con, "substance") |> dplyr::collect()

dplyr::tbl(con, "route") |> dplyr::collect()

# Replace sample and css tables with new data including an extra column
# Limit to a single substance for simplicity
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight, ~sign,
  10000, 25, "Normal", "+",
  10000, 35,  "Obese", "-",
  20000, 50, "Normal", "+"
)
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css, ~sign,
  "00-00-1",  0, 49, "Normal",   1, "+",
  "00-00-1", 50, 99, "Normal",   2, "+",
  "00-00-1",  0, 49,  "Obese",  11, "+",
  "00-00-1", 50, 99,  "Obese",  12, "+",
  "00-00-1",  0, 49, "Normal",  -1, "-",
  "00-00-1", 50, 99, "Normal",  -2, "-",
  "00-00-1",  0, 49,  "Obese", -11, "-",
  "00-00-1", 50, 99,  "Obese", -12, "-"
)
GT <- GT |>
 set_sample(sample_df, overwrite = TRUE) |>
 simulate_exposure() |>
 set_simulated_css(css_df, overwrite = TRUE)

# Sample simulated C_ss values with extra column
# Notice how the extra column name is added to GT$par
str(GT$par)
GT <- GT |> sample_simulated_css(css_extra_cols = "sign")
str(GT$par)

# Look at new 'concentration' table. Values will be missing for substance 2
# since it is not in the new css_df.
dplyr::tbl(con, "concentration") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=set_simulated_css]{set_simulated_css()}}, \code{\link[=simulate_population]{simulate_population()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sensitivity_analysis.R
\name{sensitivity_analysis}
\alias{sensitivity_analysis}
\title{Perform sensitivity analysis}
\usage{
sensitivity_analysis(GT, max_mult = c(1.5, 1.5, 1.5, 1.5, 1.5), ...)
}
\arguments{
\item{GT}{GeoTox object.}

\item{max_mult}{Vector of length 5 containing upper bound multipliers for max
response (default 1.5).}

\item{...}{Additional arguments passed to each call of \code{\link[=calc_sensitivity]{calc_sensitivity()}}.}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Calculate risk sensitivity to all available \code{vary} parameters in
\code{\link[=calc_sensitivity]{calc_sensitivity()}}.
}
\details{
Sensitivity is calculated in the order: age, weight, css_params, fit_params,
C_ext. The \code{max_mult} vector allows specifying different upper bound
multipliers for each parameter.
}
\examples{
# Example setup is shown below in \dontrun().
# Pre-generated results will be loaded instead to avoid long example runtime.

\dontrun{
# Setup required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal", 21,
  "00-00-1", 50, 99, "Normal", 22,
  "00-00-1",  0, 49,  "Obese", 61,
  "00-00-1", 50, 99,  "Obese", 62,
  "00-00-2",  0, 49, "Normal", 11,
  "00-00-2", 50, 99, "Normal", 12,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
hill_df <- tibble::tribble(
  ~assay, ~model, ~casn, ~logc, ~resp,
  "a1", "human", "00-00-1",    0,  10,
  "a1", "human", "00-00-1",    1,  20,
  "a1", "human", "00-00-1",    2,  80,
  "a1", "human", "00-00-1",    3, 100,
  "a1", "human", "00-00-2", -0.5,   5,
  "a1", "human", "00-00-2",  0.5,  20,
  "a1", "human", "00-00-2",  1.5,  55,
  "a1", "human", "00-00-2",  2.5,  60,
  "a2",   "rat", "00-00-1",   -1,   0,
  "a2",   "rat", "00-00-1",    0,  10,
  "a2",   "rat", "00-00-1",    1,  30,
  "a2",   "rat", "00-00-1",    2,  40
)
set.seed(1234)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  set_simulated_css(css_df) |>
  add_exposure_rate_params() |>
  add_hill_params(fit_hill(
    hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
  )) |>
  simulate_population(exposure = exposure_df) |>
  calc_response()

# Perform sensitivity analysis
GT <- GT |> sensitivity_analysis()
}

# Load results from pre-generated database for this example
temp_dir <- tempdir()
zip::unzip(
  system.file("extdata", "sensitivity.duckdb.zip", package = "GeoTox"),
  junkpaths = TRUE,
  exdir = temp_dir
)
GT <- GeoTox(paste0(temp_dir, "/sensitivity.duckdb"))

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant table

dplyr::tbl(con, "risk_sensitivity_age") |> dplyr::collect()

dplyr::tbl(con, "risk_sensitivity_weight") |> dplyr::collect()

dplyr::tbl(con, "risk_sensitivity_css_params") |> dplyr::collect()

dplyr::tbl(con, "risk_sensitivity_fit_params") |> dplyr::collect()

dplyr::tbl(con, "risk_sensitivity_C_ext") |> dplyr::collect()

# Compared to baseline risk table
dplyr::tbl(con, "risk") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=calc_sensitivity]{calc_sensitivity()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_tables.R
\name{add_age}
\alias{add_age}
\title{Add age simulation data}
\usage{
add_age(GT, df, location = "FIPS")
}
\arguments{
\item{GT}{GeoTox object.}

\item{df}{Data frame with age simulation data.}

\item{location}{Column name(s) in \code{df} that contain location identifier(s)
(default "FIPS").}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Create or add to the 'age' table in a GeoTox database.
}
\details{
The simulation data must have columns "AGEGRP" and "TOT_POP" and at least one
column containing location information (default "FIPS"). Each location must
have 19 rows for AGEGRP 0-18, where 1-18 are age groups in increments of 5,
e.g. AGEGRP = 5 would be ages 20 to 24, and AGEGRP = 0 is the combination of
all age groups. The age data is used by \code{\link[=simulate_age]{simulate_age()}} to generate age
samples for each location.

The \code{location} input can be a named vector to specify multiple identifier
columns in \code{df}. For example, \code{location = c(FIPS = "FIPS", state = "ST")}
would indicate that \code{df} contains both FIPS codes and state identifiers for
locations. The \code{state = "ST"} part would rename the "ST" column in \code{df} to
"state" in the 'location' table.
}
\examples{
# Example age simulation data
age_df <- data.frame(
  FIPS = rep(c(10000, 20000), each = 19),
  AGEGRP = rep(0:18, times = 2),
  TOT_POP = 0
)
# FIPS 10000, populate age group 40-44
age_df$TOT_POP[c(1, 10)] = 100
# FIPS 20000, populate age groups 50-59
age_df$TOT_POP[c(1, 12, 13) + 19] = c(200, 100, 100)

# Add age simulation data to GeoTox database
GT <- GeoTox() |> add_age(age_df)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "age") |> dplyr::filter(TOT_POP > 0) |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=simulate_age]{simulate_age()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_obesity.R
\name{simulate_obesity}
\alias{simulate_obesity}
\title{Simulate obesity values}
\usage{
simulate_obesity(
  GT,
  n = 1000,
  overwrite = FALSE,
  obes_prev = NULL,
  obes_sd = NULL
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{n}{Number of individuals to simulate per location (default 1000).
Ignored if 'sample' table already exists.}

\item{overwrite}{Logical indicating whether to overwrite existing 'weight'
values in the 'sample' table (default FALSE).}

\item{obes_prev}{Column name of obesity prevalence in the 'obesity' table
(default "OBESITY_CrudePrev").}

\item{obes_sd}{Column name of obesity standard deviation in the 'obesity'
table (default "OBESITY_SD").}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Simulate 'weight' category values to be stored in the 'sample' table of a
GeoTox database.
}
\details{
An 'obesity' table containing simulation data must already exist in the
GeoTox database, which is added using \code{\link[=add_obesity]{add_obesity()}}.

The inputs \code{obes_prev} and \code{obes_sd} will be assigned default values of
"OBESITY_CrudePrev" and "OBESITY_SD", respectively, if not provided and not
already set in the GeoTox object's parameters, \code{GT$par}. If not \code{NULL}, the
provided values will also be saved to \code{GT$par}.
}
\examples{
# Example obesity simulation data
obesity_df <- data.frame(
  FIPS = c(10000, 20000),
  OBESITY_CrudePrev = c(20, 80),
  OBESITY_SD = 5
)

# Simulate weight category values
GT <- GeoTox() |>
  add_obesity(obesity_df) |>
  simulate_obesity(n = 5)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

# Replace 'obesity' table with different column names
names(obesity_df)[2:3] <- c("prev", "sd")
DBI::dbRemoveTable(con, "obesity")
GT |> add_obesity(obesity_df)

# Overwrite 'weight' category values in existing 'sample' table
# Must specify new 'obesity' column names
# Notice how the column names are added to GT$par
str(GT$par)
GT <- GT |>
  simulate_obesity(obes_prev = "prev", obes_sd = "sd", overwrite = TRUE)
str(GT$par)

# Look at updated 'sample' table
dplyr::tbl(con, "sample") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=add_obesity]{add_obesity()}}, \code{\link[=simulate_population]{simulate_population()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{geo_tox_data}
\alias{geo_tox_data}
\title{GeoTox Data}
\format{
A list with items:
\describe{
\item{exposure}{2020 AirToxScreen exposure concentrations for a subset of
chemicals in North Carolina counties.}
\item{dose_response}{Subset of chemicals curated by ICE cHTS as active
within a set of assays.}
\item{age}{County population estimates for 7/1/2020 in North Carolina.}
\item{obesity}{CDC PLACES obesity data for North Carolina counties in
2020.}
\item{simulated_css}{Simulated steady-state plasma concentrations for
various age groups and obesity status combinations.}
\item{boundaries}{County and state boundaries for North Carolina in 2020.}
}
}
\usage{
geo_tox_data
}
\description{
Sample data for use in vignettes and function examples. See the Package
Data vignette, \code{vignette("package_data", package = "GeoTox")}, for
details on how this data was gathered.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_hill.R
\name{fit_hill}
\alias{fit_hill}
\title{Fit 2- or 3-parameter Hill model}
\usage{
fit_hill(
  x,
  conc = "logc",
  resp = "resp",
  fixed_slope = TRUE,
  assay = NULL,
  substance = NULL
)
}
\arguments{
\item{x}{Data frame of dose response data.}

\item{conc}{Column name of base-10 log scaled concentration (default "logc").}

\item{resp}{Column name of response (default "resp").}

\item{fixed_slope}{Logical indicating whether to fit a 2-parameter (TRUE) or
3-parameter (FALSE) Hill model (default TRUE).}

\item{assay}{Column name of assay identifier(s) (optional, default NULL).}

\item{substance}{Column name of substance identifier(s) (optional, default
NULL).}
}
\value{
A list with elements 'fit', 'assay', 'substance'. The 'fit' element
is a data frame of fit parameters, while the 'assay' and 'substance'
elements indicate the column names used for assay and substance
identifiers, respectively.
}
\description{
Fit a 2-parameter (fixed slope) or 3-parameter (variable slope) Hill model to
concentration-response data.
}
\details{
The input \code{x} data frame must contain columns specified by \code{conc} and \code{resp}
arguments representing the log10-transformed concentration and response,
respectively.

Optional \code{assay} and \code{substance} identifiers can be named vectors that are
used to fit multiple substances and/or assays. For example, \code{assay = c(name = "assay", "model")} would indicate that \code{x} contains both and assay name and
model. The \code{name = "assay"} part would rename the "assay" column in \code{x} to
"name" in the 'assay' table when aded with \code{\link[=add_hill_params]{add_hill_params()}}.

Returned column 'tp' is the top asymptote and 'logAC50' is the 50\% response
concentration. If the computation of the standard deviations of these two
parameters fails, then the standard deviation is set equal to the parameter
estimate and is indicated by the respective imputed flag being TRUE.
}
\examples{
hill_df <- tibble::tribble(
  ~assay, ~model, ~casn, ~logc, ~resp,
  "a1", "human", "00-00-1",    0,  10,
  "a1", "human", "00-00-1",    1,  20,
  "a1", "human", "00-00-1",    2,  80,
  "a1", "human", "00-00-1",    3, 100,
  "a1", "human", "00-00-2", -0.5,   5,
  "a1", "human", "00-00-2",  0.5,  20,
  "a1", "human", "00-00-2",  1.5,  55,
  "a1", "human", "00-00-2",  2.5,  60,
  "a2",   "rat", "00-00-1",   -1,   0,
  "a2",   "rat", "00-00-1",    0,  10,
  "a2",   "rat", "00-00-1",    1,  30,
  "a2",   "rat", "00-00-1",    2,  40
)

# Fit 2-parameter Hill model
fit_hill(
  hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
)

# Fit 3-parameter Hill model
fit_hill(hill_df, assay = "assay", substance = "casn", fixed_slope = FALSE)
}
\seealso{
\code{\link[=add_hill_params]{add_hill_params()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_internal_dose.R
\name{calc_internal_dose}
\alias{calc_internal_dose}
\title{Calculate internal dose}
\usage{
calc_internal_dose(GT, overwrite = FALSE, sensitivity = FALSE)
}
\arguments{
\item{GT}{GeoTox object.}

\item{overwrite}{Logical indicating whether to overwrite existing 'D_int'
values in the 'concentration' table (default FALSE).}

\item{sensitivity}{Logical indicating whether to simulate internal dose for
sensitivity analysis (default FALSE).}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Calculates internal dose (D_int) as the product of external concentration
('C_ext' in the 'concentration' table) and exposure rate ('rate' in the
'exposure_rate' table), and stores the results in the 'D_int' column of the
'concentration' table in the GeoTox database.
}
\details{
If \code{sensitivity = TRUE}, 'D_int' will be calculated for sensitivity
analysis. Typically this shouldn't be used directly by the user, but rather
called by \code{\link[=calc_sensitivity]{calc_sensitivity()}}. In this case, the function will use the
'concentration_sensitivity' and 'exposure_rate_sensitivity' tables instead of
the 'concentration' and 'exposure_rate' tables.
}
\examples{
# Setup required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  add_exposure_rate_params() |>
  simulate_population(exposure = exposure_df, sample_css = FALSE)

# Calculate internal dose
GT <- GT |> calc_internal_dose()

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant tables

dplyr::tbl(con, "concentration") |> dplyr::collect()

dplyr::tbl(con, "exposure_rate") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=calc_response]{calc_response()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_tables.R
\name{add_hill_params}
\alias{add_hill_params}
\title{Add Hill model parameters}
\usage{
add_hill_params(GT, hill_params)
}
\arguments{
\item{GT}{GeoTox object.}

\item{hill_params}{List output from \code{\link[=fit_hill]{fit_hill()}}.}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Create or add to the 'hill_params' table in a GeoTox database.
}
\examples{
# Example Hill model data
hill_df <- tibble::tribble(
  ~assay, ~model, ~casn, ~logc, ~resp,
  "a1", "human", "00-00-1",    0,  10,
  "a1", "human", "00-00-1",    1,  20,
  "a1", "human", "00-00-1",    2,  80,
  "a1", "human", "00-00-1",    3, 100,
  "a1", "human", "00-00-2", -0.5,   5,
  "a1", "human", "00-00-2",  0.5,  20,
  "a1", "human", "00-00-2",  1.5,  55,
  "a1", "human", "00-00-2",  2.5,  60,
  "a2",   "rat", "00-00-1",   -1,   0,
  "a2",   "rat", "00-00-1",    0,  10,
  "a2",   "rat", "00-00-1",    1,  30,
  "a2",   "rat", "00-00-1",    2,  40
)
hill_params <- fit_hill(
  hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
)

# Add Hill model parameters to GeoTox database
GT <- GeoTox() |> add_hill_params(hill_params)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "hill_params") |> dplyr::collect()

dplyr::tbl(con, "assay") |> dplyr::collect()

dplyr::tbl(con, "substance") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=fit_hill]{fit_hill()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_simulated_css.R
\name{set_simulated_css}
\alias{set_simulated_css}
\title{Set pre-simulated steady-state plasma concentrations}
\usage{
set_simulated_css(GT, df, substance = "casn", overwrite = FALSE)
}
\arguments{
\item{GT}{GeoTox object.}

\item{df}{Data frame containing simulated C\eqn{_{ss}} values for groups of
population characteristics.}

\item{substance}{Column name(s) in \code{df} that contain substance identifier(s)
(default "casn").}

\item{overwrite}{Logical indicating whether to overwrite existing
'simulated_css' table (default FALSE).}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Create the 'simulated_css' table in a GeoTox database, which contains
pre-simulated steady-state plasma concentrations (C\eqn{_{ss}}).
}
\details{
The 'simulated_css' table is used by \code{\link[=sample_simulated_css]{sample_simulated_css()}} to assign
C\eqn{_{ss}} values to individuals. The minimum required columns in the \code{df}
data frame are "age_lb", "age_ub", "weight", "css", and at least one column
with substance information (default "casn").

The values for "age_lb" and "age_ub" should be non-overlapping integers
representing age ranges (in years). For example, two subsequent age groups
might be \code{c(0, 4)} and \code{c(5, 9)}. The "weight" column should contain the
weight category and contain values of either "Normal" or "Obese". The "css"
column should contain the pre-simulated C\eqn{_{ss}} values.

The \code{substance} input can be a named vector to specify multiple substance
identifier columns in \code{df}. For example, \code{c(casn = "casn", name = "chnm")}
would indicate that \code{df} contains both CAS numbers and chemical names for
substances. The \code{name = "chnm"} part would rename the "chnm" column in \code{df}
to "name" in the 'substance' table.
}
\examples{
# Example pre-simulated C_ss data
# Note: normally the css_df would have many more rows for each combination of
# the non-'css' columns to allow for sampling.
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal",  1,
  "00-00-1", 50, 99, "Normal",  2,
  "00-00-1",  0, 49,  "Obese", 11,
  "00-00-1", 50, 99,  "Obese", 12,
  "00-00-2",  0, 49, "Normal", 21,
  "00-00-2", 50, 99, "Normal", 22,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)

# Set simulated C_ss values
GT <- GeoTox() |> set_simulated_css(css_df)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant tables

dplyr::tbl(con, "simulated_css") |> dplyr::collect()

dplyr::tbl(con, "substance") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=sample_simulated_css]{sample_simulated_css()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/boundary_table.R
\name{set_boundary}
\alias{set_boundary}
\alias{get_boundary}
\title{Store and retrieve boundary geometries}
\usage{
set_boundary(GT, df_list, location = "county", overwrite = FALSE)

get_boundary(GT)
}
\arguments{
\item{GT}{GeoTox object.}

\item{df_list}{Named list of data frames containing boundary geometries as sf
objects.}

\item{location}{Name of element in \code{df_list} that contains location boundary
information (default "county").}

\item{overwrite}{Logical indicating whether to overwrite existing 'boundary'
table.}
}
\value{
For \code{set_boundary()}, the same GeoTox object, invisibly. For
\code{get_boundary()}, a data frame with columns 'id' (boundary name) and 'data'
(sf object).
}
\description{
Boundary information is stored as serialized 'BLOB' objects in the 'boundary'
table.
}
\details{
NOTE: This function requires the \code{sf} package to be installed.

This function takes a named list of sf objects and stores them in the
database. If a location boundary (default "county") is provided, then the
non-geometry fields will be added to the 'location' table and replaced with a
'location_id' value so they can be linked to data in the 'sample' table.
}
\examples{
# Setup sf objects
county <- sf::st_sf(
  FIPS = c(10000, 20000),
  geometry = sf::st_sfc(sf::st_point(1:2), sf::st_point(3:4))
)
state <- sf::st_sf(
  STATE = "XYZ",
  geometry = sf::st_sfc(sf::st_point(5:6))
)
df_list <- list(county = county, state = state)

# Add boundary to GeoTox database
GT <- GeoTox() |> set_boundary(df_list)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "boundary") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

# Retrieve boundary from GeoTox database
boundary <- get_boundary(GT)
boundary
boundary |> tibble::deframe()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_tables.R
\name{add_exposure_rate_params}
\alias{add_exposure_rate_params}
\title{Add exposure rate simulation data}
\usage{
add_exposure_rate_params(
  GT,
  route = "inhalation",
  params = NULL,
  overwrite = FALSE
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{route}{Exposure route (default "inhalation").}

\item{params}{Data frame with exposure rate parameters. If \code{NULL}, default
parameters for the specified \code{route} will be used (default \code{NULL}).}

\item{overwrite}{Logical indicating whether to overwrite existing exposure
rate parameters for the specified \code{route} (default \code{FALSE}).}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Create or add to the 'exposure_rate_params' table in a GeoTox database.
}
\details{
There are two routes with built-in default exposure rate parameters:
"inhalation" and "drinking". If \code{params} is not provided (\code{NULL}), the
default parameters for the specified \code{route} will be used. If \code{params} is
provided, it must contain columns "age_lb", "age_ub", "mean", and "sd". The
exposure rate parameters are used by \code{\link[=simulate_exposure_rate]{simulate_exposure_rate()}} to generate
exposure rate samples for each individual in the population.

The build-in parameters come from the following sources:
\itemize{
\item Inhalation: \href{https://www.epa.gov/sites/default/files/2015-09/documents/efh-chapter06.pdf}{EPA Exposure Factors Handbook (2015), Table 6.7}
in m\eqn{^3}/kg-day. The "mean" and "sd" values are the average of male and
female values.
\item Drinking: \href{https://www.epa.gov/sites/default/files/2019-02/documents/efh_-_chapter_3_update.pdf}{EPA Exposure Factors Handbook (2019), Table 3-30}
in mL/kg-day.
}
}
\examples{
# Add both default params to GeoTox database
GT <- GeoTox() |>
  add_exposure_rate_params() |>
  add_exposure_rate_params(route = "drinking")

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant tables
params_tbl <- dplyr::tbl(con, "exposure_rate_params") |> dplyr::collect()

params_tbl |> dplyr::filter(route_id == 1)

params_tbl |> dplyr::filter(route_id == 2)

dplyr::tbl(con, "route") |> dplyr::collect()

# Add custom params with new column (gender), assign to route "custom"
params_df <- tibble::tribble(
  ~age_lb, ~age_ub, ~gender, ~mean, ~sd,
   0, 49, "male",   10, 1,
  50, 99, "male",   20, 1,
   0, 49, "female", 30, 1,
  50, 99, "female", 40, 1
)
GT |> add_exposure_rate_params(params_df, route = "custom")

# Look at updated tables
params_tbl <- dplyr::tbl(con, "exposure_rate_params") |> dplyr::collect()

params_tbl |> dplyr::filter(route_id == 1)

params_tbl |> dplyr::filter(route_id == 2)

params_tbl |> dplyr::filter(route_id == 3)

dplyr::tbl(con, "route") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=simulate_exposure_rate]{simulate_exposure_rate()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_invitro_concentration.R
\name{calc_invitro_concentration}
\alias{calc_invitro_concentration}
\title{Calculate in vitro concentration}
\usage{
calc_invitro_concentration(GT, overwrite = FALSE, sensitivity = FALSE)
}
\arguments{
\item{GT}{GeoTox object.}

\item{overwrite}{Logical indicating whether to overwrite existing 'C_invitro'
values in the 'concentration' table (default FALSE).}

\item{sensitivity}{Logical indicating whether to simulate in vitro
concentration for sensitivity analysis (default FALSE).}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Calculates the in vitro concentration (C_invitro) as the product of steady
state plasma concentration (C_ss) and internal dose (D_int), and stores the
results in the 'C_invitro' column of the 'concentration' table in the GeoTox
database.
}
\details{
If \code{sensitivity = TRUE}, 'C_invitro' will be calculated for sensitivity
analysis. Typically this shouldn't be used directly by the user, but rather
called by \code{\link[=calc_sensitivity]{calc_sensitivity()}}. In this case, the function will use the
'concentration_sensitivity' table instead of the 'concentration' table.
}
\examples{
# Setup required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal", 21,
  "00-00-1", 50, 99, "Normal", 22,
  "00-00-1",  0, 49,  "Obese", 61,
  "00-00-1", 50, 99,  "Obese", 62,
  "00-00-2",  0, 49, "Normal", 11,
  "00-00-2", 50, 99, "Normal", 12,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  set_simulated_css(css_df) |>
  add_exposure_rate_params() |>
  simulate_population(exposure = exposure_df) |>
  calc_internal_dose()

# Calculate in vitro concentration
GT <- GT |> calc_invitro_concentration()

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at relevant tables
dplyr::tbl(con, "concentration") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=calc_internal_dose]{calc_internal_dose()}}, \code{\link[=calc_response]{calc_response()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_assay_table.R, R/get_risk_quantiles.R,
%   R/get_risk_sensitivity.R, R/get_risk_values.R
\name{get_assay_table}
\alias{get_assay_table}
\alias{get_risk_quantiles}
\alias{get_risk_sensitivity}
\alias{get_risk_values}
\title{Get risk metric results}
\usage{
get_assay_table(GT)

get_risk_quantiles(
  GT,
  metric = c("GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"),
  quantiles = c(0.1, 0.25, 0.5, 0.75, 0.9),
  table_name = "risk"
)

get_risk_sensitivity(
  GT,
  metric = c("GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"),
  assay = NULL
)

get_risk_values(
  GT,
  metric = c("GCA.Eff", "IA.Eff", "GCA.HQ.10", "IA.HQ.10"),
  assay = NULL,
  table_name = "risk"
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{metric}{Risk metric to retrieve. One of "GCA.Eff", "IA.Eff",
"GCA.HQ.10", or "IA.HQ.10".}

\item{quantiles}{Numeric vector of quantiles to calculate.}

\item{table_name}{Name of the risk table to query (default "risk").}

\item{assay}{Named vector specifying an assay filter (default NULL).}
}
\value{
A data frame or vector.
}
\description{
Several functions used to fetch risk metric results from risk tables in a
GeoTox database. The outputs of these functions are useful for plotting or
further analysis.
}
\details{
Normally an 'assay' table is created when adding the Hill model
concentration-response fit parameters with \code{\link[=add_hill_params]{add_hill_params()}}. If no \code{assay}
input is specified in \code{\link[=fit_hill]{fit_hill()}}, then an 'assay' table will not be
created. For \code{\link[=get_risk_values]{get_risk_values()}}, the \code{assay} parameter can be used to filter
results based on assay details stored in the 'assay' table. This is useful
when multiple assays are available in the database. The \code{assay} input should
be a named vector specifying the column name and value to filter by, e.g.
\code{assay = c(endp = "mortality")}. For \code{\link[=get_risk_quantiles]{get_risk_quantiles()}}, if there is no
assay data in the GeoTox database, then the output "assay_id" column will be
filled with NA values. If there is assay data, use \code{\link[=get_assay_table]{get_assay_table()}} to
retrieve assay information and link the "assay_id" to assay details.

Use the \code{table_name} parameter to specify which risk table to query. There
can be several risk tables in a GeoTox database. The default table is named
"risk" and is created by either \code{\link[=calc_risk]{calc_risk()}} or the wrapper function
\code{\link[=calc_response]{calc_response()}}. Sensitivity analysis results created by either
\code{\link[=calc_sensitivity]{calc_sensitivity()}} or the wrapper function \code{\link[=sensitivity_analysis]{sensitivity_analysis()}} are
stored in other tables with names like "risk_sensitivity_age", etc. Refer to
\code{\link[=calc_sensitivity]{calc_sensitivity()}} to see which tables may be available.

\code{\link[=get_risk_sensitivity]{get_risk_sensitivity()}} is a wrapper function for \code{\link[=get_risk_values]{get_risk_values()}} around
all risk tables created by the original risk computation and subsequent
sensitivity analysis. The column names are "baseline" for the original risk
table, while the other column names correspond to the parameter varied in the
sensitivity analysis.
}
\examples{
# Example setup is shown below in \dontrun().
# Pre-generated results will be loaded instead to avoid long example runtime.

\dontrun{
# Setup required tables
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight,
  10000, 25, "Normal",
  10000, 35,  "Obese",
  20000, 50, "Normal"
)
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
css_df <- tibble::tribble(
  ~casn, ~age_lb, ~age_ub, ~weight, ~css,
  "00-00-1",  0, 49, "Normal", 21,
  "00-00-1", 50, 99, "Normal", 22,
  "00-00-1",  0, 49,  "Obese", 61,
  "00-00-1", 50, 99,  "Obese", 62,
  "00-00-2",  0, 49, "Normal", 11,
  "00-00-2", 50, 99, "Normal", 12,
  "00-00-2",  0, 49,  "Obese", 31,
  "00-00-2", 50, 99,  "Obese", 32
)
hill_df <- tibble::tribble(
  ~assay, ~model, ~casn, ~logc, ~resp,
  "a1", "human", "00-00-1",    0,  10,
  "a1", "human", "00-00-1",    1,  20,
  "a1", "human", "00-00-1",    2,  80,
  "a1", "human", "00-00-1",    3, 100,
  "a1", "human", "00-00-2", -0.5,   5,
  "a1", "human", "00-00-2",  0.5,  20,
  "a1", "human", "00-00-2",  1.5,  55,
  "a1", "human", "00-00-2",  2.5,  60,
  "a2",   "rat", "00-00-1",   -1,   0,
  "a2",   "rat", "00-00-1",    0,  10,
  "a2",   "rat", "00-00-1",    1,  30,
  "a2",   "rat", "00-00-1",    2,  40
)
set.seed(1234)
GT <- GeoTox() |>
  set_sample(sample_df) |>
  set_simulated_css(css_df) |>
  add_exposure_rate_params() |>
  add_hill_params(fit_hill(
    hill_df, assay = c(name = "assay", model = "model"), substance = "casn"
  )) |>
  simulate_population(exposure = exposure_df) |>
  calc_response() |>
  sensitivity_analysis()

}

# Load results from pre-generated database for this example
temp_dir <- tempdir()
zip::unzip(
  system.file("extdata", "sensitivity.duckdb.zip", package = "GeoTox"),
  junkpaths = TRUE,
  exdir = temp_dir
)
GT <- GeoTox(paste0(temp_dir, "/sensitivity.duckdb"))

# Look at 'assay' table contents
get_assay_table(GT)

# Get "GCA.HQ.10" values from 'risk' table for the "a1" assay
get_risk_values(GT, metric = "GCA.HQ.10", assay = c(name = "a1"))

# Get "IA.Eff" values from all risk tables for the "a2" assay
get_risk_sensitivity(GT, metric = "IA.Eff", assay = c(name = "a2"))

# Get "GCA.Eff" quantiles from 'risk' table
get_risk_quantiles(GT, metric = "GCA.Eff", quantiles = c(0.25, 0.5, 0.75))

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at the 'risk' table contents
dplyr::tbl(con, "risk") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=calc_risk]{calc_risk()}}, \code{\link[=calc_response]{calc_response()}}, \code{\link[=calc_sensitivity]{calc_sensitivity()}},
\code{\link[=sensitivity_analysis]{sensitivity_analysis()}}, \code{\link[=fit_hill]{fit_hill()}}, \code{\link[=add_hill_params]{add_hill_params()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_tables.R
\name{add_exposure}
\alias{add_exposure}
\title{Add exposure simulation data}
\usage{
add_exposure(GT, df, location = "FIPS", substance = "casn", route = "route")
}
\arguments{
\item{GT}{GeoTox object.}

\item{df}{Data frame with exposure simulation data.}

\item{location}{Column name(s) in \code{df} that contain location identifier(s)
(default "FIPS").}

\item{substance}{Column name(s) in \code{df} that contain substance identifier(s)
(default "casn").}

\item{route}{Column name in \code{df} that contains route identifier (default
"route").}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Create or add to the 'exposure' table in a GeoTox database.
}
\details{
The simulation data must contain columns for exposure mean and standard
deviation (default "mean" and "sd", respectfully), at least one column
containing location information (default "FIPS"), at least one column
containing substance information (default "casn"), and one column containing
route information (default "route"). The exposure data is used by
\code{\link[=simulate_exposure]{simulate_exposure()}} to generate external exposure concentration samples for
each location.

The \code{location} and \code{substance} inputs can be named vectors to specify
multiple identifier columns in \code{df}. For example, \code{substance = c(casn = "casn", name = "chnm")} would indicate that \code{df} contains both CAS numbers
and chemical names for substances. The \code{name = "chnm"} part would rename the
"chnm" column in \code{df} to "name" in the 'substance' table.
}
\examples{
# Example exposure simulation data
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)

# Add exposure simulation data to GeoTox database
GT <- GeoTox() |> add_exposure(exposure_df)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables
dplyr::tbl(con, "exposure") |> dplyr::collect()
dplyr::tbl(con, "location") |> dplyr::collect()
dplyr::tbl(con, "substance") |> dplyr::collect()
dplyr::tbl(con, "route") |> dplyr::collect()

# Add another substance with a new field and route
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~chnm, ~route, ~mean, ~sd,
  10000, "00-00-3", "chem3", "drinking", 100, 1,
  20000, "00-00-3", "chem3", "drinking", 200, 1
)
GT |> add_exposure(exposure_df, substance = c(casn = "casn", name = "chnm"))

# Look at updated tables

dplyr::tbl(con, "exposure") |> dplyr::collect()

dplyr::tbl(con, "substance") |> dplyr::collect()

dplyr::tbl(con, "route") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=simulate_exposure]{simulate_exposure()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_tables.R
\name{add_obesity}
\alias{add_obesity}
\title{Add obesity simulation data}
\usage{
add_obesity(GT, df, location = "FIPS")
}
\arguments{
\item{GT}{GeoTox object.}

\item{df}{Data frame with obesity simulation data.}

\item{location}{Column name(s) in \code{df} that contain location identifier(s)
(default "FIPS").}
}
\value{
The same GeoTox object, invisibly.
}
\description{
Create or add to the 'obesity' table in a GeoTox database.
}
\details{
The simulation data must contain columns for obesity prevalence and standard
deviation (default "OBESITY_CrudePrev" and "OBESITY_SD", respectfully) and at
least one column containing location information (default "FIPS"). The
obesity data is used by \code{\link[=simulate_obesity]{simulate_obesity()}} to generate weight category
samples for each location.

The \code{location} input can be a named vector to specify multiple identifier
columns in \code{df}. For example, \code{location = c(FIPS = "FIPS", state = "ST")}
would indicate that \code{df} contains both FIPS codes and state identifiers for
locations. The \code{state = "ST"} part would rename the "ST" column in \code{df} to
"state" in the 'location' table.
}
\examples{
# Example obesity simulation data
obesity_df <- data.frame(
  FIPS = c(10000, 20000),
  OBESITY_CrudePrev = c(20, 80),
  OBESITY_SD = 5
)

# Add obesity simulation data to GeoTox database
GT <- GeoTox() |> add_obesity(obesity_df)

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "obesity") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

# Add another location with additional information
obesity_df <- data.frame(
 FIPS = 30000,
 ST = "State3",
 OBESITY_CrudePrev = 50,
 OBESITY_SD = 10
)
GT |> add_obesity(obesity_df, location = c(FIPS = "FIPS", state = "ST"))

# Look at updated tables

dplyr::tbl(con, "obesity") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=simulate_obesity]{simulate_obesity()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_exposure_rate.R
\name{simulate_exposure_rate}
\alias{simulate_exposure_rate}
\title{Simulate exposure rates}
\usage{
simulate_exposure_rate(
  GT,
  rate_extra_cols = NULL,
  overwrite = FALSE,
  sensitivity = FALSE
)
}
\arguments{
\item{GT}{GeoTox object.}

\item{rate_extra_cols}{Additional columns to match from the
'exposure_rate_params' table (default NULL).}

\item{overwrite}{Logical indicating whether to overwrite existing
'exposure_rate' table (default FALSE).}

\item{sensitivity}{Logical indicating whether to simulate exposure rates for
sensitivity analysis (default FALSE).}
}
\value{
The updated GeoTox object, invisibly.
}
\description{
Simulate exposure rate values for each sample and exposure route based on
parameters in the \code{exposure_rate_params} table.
}
\details{
An 'exposure_rate_params' table must already exist in the GeoTox database,
which is added using \code{\link[=add_exposure_rate_params]{add_exposure_rate_params()}}. There must also be a
'sample' table with an 'age' column, which can be created using
\code{\link[=simulate_age]{simulate_age()}} or \code{\link[=set_sample]{set_sample()}}. 'location' and 'route' tables must also
exist, but they are created when adding the rate parameters and samples.

If \code{rate_exta_cols} is provided it will be added to the GeoTox object
parameter list, \code{GT$par}. These columns must exist in the
'exposure_rate_params' table and will be used to match between the 'sample'
and 'exposure_rate_params' tables when simulating exposure rates.

Typically this function will be called with \code{sensitivity} set to \code{FALSE}. In
this case, the function will create an 'exposure_rate' table with simulated
exposure rates for each sample and exposure route.

If \code{sensitivity} is \code{TRUE}, exposure rates will be simulated for sensitivity
analysis. Typically this shouldn't be used directly by the user, but rather
called by \code{\link[=calc_sensitivity]{calc_sensitivity()}}. In this case, the existing 'exposure_rate'
table will be copied to the 'exposure_rate_sensitivity' table where the rate
values will be overwritten.
}
\examples{
# Setup required tables
# Note: 'gender' is ignored when using the default rate params
sample_df <- tibble::tribble(
  ~FIPS, ~age, ~weight, ~gender,
  10000, 25, "Normal",   "male",
  10000, 35,  "Obese",   "male",
  20000, 50, "Normal", "female"
)
GT <- GeoTox() |>
  add_exposure_rate_params() |>
  set_sample(sample_df)

# Simulate exposure rates
GT |> simulate_exposure_rate()

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at created tables

dplyr::tbl(con, "exposure_rate") |> dplyr::collect()

dplyr::tbl(con, "sample") |> dplyr::collect()

dplyr::tbl(con, "location") |> dplyr::collect()

dplyr::tbl(con, "route") |> dplyr::collect()

# Replace exposure rate params with a new table that includes gender
params_df <- tibble::tribble(
  ~age_lb, ~age_ub, ~gender, ~mean, ~sd,
   0, 49,   "male", 10, 1,
  50, 99,   "male", 20, 1,
   0, 49, "female", 30, 1,
  50, 99, "female", 40, 1
)
DBI::dbRemoveTable(con, "exposure_rate_params")
GT |> add_exposure_rate_params(params = params_df)

# Overwrite 'rate' values in existing 'exposure_rate' table
# Must specify additional column names
# Notice how the column names are added to GT$par
str(GT$par)
GT <- GT |>
  simulate_exposure_rate(rate_extra_cols = c("gender"), overwrite = TRUE)
str(GT$par)

# Look at updated 'exposure_rate' table
dplyr::tbl(con, "exposure_rate") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
\seealso{
\code{\link[=add_exposure_rate_params]{add_exposure_rate_params()}}, \code{\link[=simulate_population]{simulate_population()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_concentration_mean.R
\name{get_concentration_mean}
\alias{get_concentration_mean}
\title{Get concentration mean values}
\usage{
get_concentration_mean(GT, col)
}
\arguments{
\item{GT}{GeoTox object.}

\item{col}{Column name in the 'concentration' table for which to calculate
the mean, grouped by substance, route, and location.}
}
\value{
A data frame.
}
\description{
Calculates the concentration mean values for each substance and route at each
location based on the data in the 'concentration' and 'sample' tables of the
GeoTox database. The output of this function is useful for plotting or
further analysis.
}
\details{
The \code{col} parameter specifies which column in the 'concentration' table to
use for the mean calculation. The available choices will depend on what data
has been stored in the 'concentration' table during the simulation process.
}
\examples{
# Setup required tables
exposure_df <- tibble::tribble(
  ~FIPS, ~casn, ~route, ~mean, ~sd,
  10000, "00-00-1", "inhalation", 10, 1,
  10000, "00-00-2", "inhalation", 20, 1,
  20000, "00-00-1", "inhalation", 30, 1,
  20000, "00-00-2", "inhalation", 40, 1
)
GT <- GeoTox() |>
  add_exposure(exposure_df) |>
  simulate_exposure(n = 100)

# Calculate mean external concentration by substance and location
get_concentration_mean(GT, "C_ext")

# Open a connection to GeoTox database
con <- get_con(GT)

# Look at column names in the 'concentration' table
dplyr::tbl(con, "concentration") |> dplyr::collect()

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GeoTox.R
\name{GeoTox}
\alias{GeoTox}
\alias{get_con}
\title{GeoTox S3 object}
\usage{
GeoTox(dbname = tempfile(fileext = ".duckdb"), reset_seed = FALSE, ...)

get_con(GT)
}
\arguments{
\item{dbname}{Database file name. Default is a temporary file.}

\item{reset_seed}{Logical indicating whether to reset the user's global
\code{.Random.seed} after certain database operations (default \code{FALSE}).}

\item{...}{Additional arguments passed to \code{\link[duckdb:duckdb]{duckdb::duckdb()}}.}

\item{GT}{GeoTox object.}
}
\value{
For \code{GeoTox()}, a GeoTox S3 object. For \code{get_con()}, a database
connection.
}
\description{
Create a GeoTox object and connect to the underlying database.
}
\details{
The \code{dbname} will point to a DuckDB database file. If the file does not
already exist, a new database will be created. Additional arguments passed
via \code{...} will be forwarded to \code{\link[duckdb:duckdb]{duckdb::duckdb()}}.

The \code{reset_seed} parameter is necessary for replicating results from the
previous GeoTox implementation. Some database functions create temporary
tables where a random string is appended to the table name. This advances the
\code{.Random.seed} state and causes functions that use randomness to produce
different results between the previous and current implementations. Setting
\code{reset_seed = TRUE} will reset the user's \code{.Random.seed} to the value it had
before certain database operations.

Various parameters will be stored in the 'par' table. These parameters will
also be loaded into the GeoTox object as a list in \code{GT$par}. The value for
\code{reset_seed} can only be set on initial GeoTox object creation and will be
loaded from the 'par' table for existing databases.
}
\examples{
# Create a GeoTox object
GT <- GeoTox()

# Open a connection to GeoTox database
con <- get_con(GT)

# List database tables
DBI::dbListTables(con)

# Look at the GT parameters
dplyr::tbl(con, "par") |> dplyr::collect()
str(GT$par)

# Clean up example
DBI::dbDisconnect(con)
file.remove(GT$db_info$dbdir)
}
